#ifndef __GUI_DATA_FRAME_HPP_
#define __GUI_DATA_FRAME_HPP_

/*  
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrei Shkeda
 *
 * File Description:
 *
 */

#include <util/bitset/bm.h>
#include <util/bitset/bmsparsevec.h>
#include <util/bitset/bmstrsparsevec.h>
#include <util/bitset/bmsparsevec_serial.h>
#include <util/bitset/bmsparsevec_compr.h>
#include <util/bitset/bmintervals.h>
#include <vector>
#include <ostream>
#include <memory>
#include <algorithm>

using namespace std;

typedef enum {
    eDF_String,
    eDF_Uint,
    eDF_Bit
} EDF_ColumnType;

 
 

#ifndef __USE_BM_XOR_COMPRESSION__
#define __USE_BM_XOR_COMPRESSION__
#endif


/*!
    CDataFrame implements succinct columnar data frame 
    It supports three data types : String, Uint and Bit 
    The data types correspond to BitMagic's vector types:
        str_sparse_vector, sparse_vector<unsigned>, bvector
    
    Template parameter MAX_SIZE defines the max size for string columns

    Constructor's initializer_list describes the number and the types of columns 

    GetColumn(size_t idx)
    GetColumns() 
    provide access to the columns

    Column is a union structure with pointer to the correspondind data types
    The pointer provides access to underlying vectors for further manipulations
    The caller is responsible for invoking correct pointer to the corresponding data type

    Serialize implements serialization to ostream 
    Deserialize implements deserialization from a string buffer

 */

template<unsigned MAX_SIZE>
class CDataFrame 
{

public:
    typedef bm::sparse_vector<unsigned, bm::bvector<> > svector_u32;
    typedef bm::rsc_sparse_vector<unsigned, svector_u32 >  rsc_svector_u32;

    typedef typename bm::str_sparse_vector<char,  bm::bvector<>, MAX_SIZE> TStrType;
    typedef typename bm::str_sparse_vector<char,  bm::bvector<>, MAX_SIZE>::const_iterator TStrIterator;
    typedef svector_u32 TUintType;
    
    typedef bm::bvector<> TBitType;

//    typedef bm::rsc_sparse_vector<unsigned, bm::sparse_vector<unsigned, bm::bvector<>> >  rsc_svector_u32;



    /**
     * Column provides access to the vector data via union of pointers 
     * to the specific data type
     * Caller is responisble for accesion the correct pointer 
     * Usage:
     * df.GetColumn(0).Str->set(0, "1"); // access to str_sparse_vector
     * df.GetColumn(1).Uint->set(0, 1);  // access to sparse_vector<unsigned>
     * df.GetColumn(2).Bit->set(0);      // access to bit vector 
     * 
    */
    struct Column {
        EDF_ColumnType type;
        union {
            unique_ptr<TStrType> Str;
            unique_ptr<TUintType> Uint;
            unique_ptr<TBitType> Bit;
        };

        Column(unique_ptr<TStrType>& p)  {
            new (&Str) unique_ptr<TStrType>{move(p)};
            type = eDF_String;
        }
        Column(unique_ptr<TUintType>& p)  {
            new (&Uint) unique_ptr<TUintType>{move(p)};
            type = eDF_Uint;
        }
        Column(unique_ptr<TBitType>& p)  {
            new (&Bit) unique_ptr<TBitType>{move(p)};
            type = eDF_Bit;
        }
        ~Column() {
            switch (type)
            {
                case eDF_String: Str.~unique_ptr<TStrType>(); 
                    break;
                case eDF_Uint: Uint.~unique_ptr<TUintType>(); 
                    break;
                case eDF_Bit: Bit.~unique_ptr<TBitType>(); 
                    break;
            }
        }
        
        bool Equal(const Column& c) const 
        {
            if (type != c.type)
                return false;
            switch (type)
            {
                case eDF_String: 
                    return Str->equal(*c.Str);
                case eDF_Uint: 
                    return Uint->equal(*c.Uint);
                case eDF_Bit: 
                    return Bit->equal(*c.Bit);
            }            
            return false;
        }
    };


    typedef vector<unique_ptr<Column>> TColumns;

    /** Default constructor */
    CDataFrame() {
    }
    
    /** Constructor with column initiazlization
     * @param[in] columns initializer list of column types
     *   Usage:
     *   CDataFrame<255> df({eDF_String, eDF_Uint, eDF_Bit});
    */
    CDataFrame(initializer_list<EDF_ColumnType> columns) 
    {
        CreateColumns(columns);
    }

    virtual ~CDataFrame() {}

    void CreateColumns(initializer_list<EDF_ColumnType> columns) 
    {
        for (const auto& t : columns) {
            unique_ptr<Column> c;
            switch (t) {
            case eDF_String: {
                    unique_ptr<TStrType> v(new TStrType);
                    c.reset(new Column(v));
                break;
            }
            case eDF_Uint: {
                unique_ptr<TUintType> v(new TUintType);
                c.reset(new Column(v));
                break;
            }
            case eDF_Bit: {
                unique_ptr<TBitType> v(new TBitType);
                c.reset(new Column(v));
                break;
            }
            }
            m_Columns.emplace_back(move(c));
        }
    }


    /** 
     * Access to a column by type and index with no boundary checks
    */

    void SetString(size_t col, size_t row, const string& str) noexcept
    {
        m_Columns[col]->Str->set(row, str.c_str());
    }

    TStrType& String(size_t idx) noexcept
    {
        return *m_Columns[idx]->Str;
    }

    TUintType& Uint(size_t idx) noexcept
    {
        return *m_Columns[idx]->Uint;
    }

    TBitType& Bit(size_t idx) noexcept
    {
        return *m_Columns[idx]->Bit;
    }

    const TStrType& String(size_t idx) const noexcept
    {
        return *m_Columns[idx]->Str;
    }

    const TUintType& Uint(size_t idx) const noexcept
    {
        return *m_Columns[idx]->Uint;
    }

    const TBitType& Bit(size_t idx) const noexcept
    {
        return *m_Columns[idx]->Bit;
    }

    /** 
     * Access to all columns 
    */
    const TColumns& GetColumns() const noexcept
    {
        return m_Columns;
    }

    /** 
     * Optimize internal data after adding or updating new rows 
     * @param[in] remap when true will remap string vector to save the space
    */
    void Optimize(bool remap = false) {
        BM_DECLARE_TEMP_BLOCK(TB)
        for (auto& c : m_Columns) {
            switch (c->type) {
            case eDF_String:
                // if (remap)
                //     c->Str->remap();
                c->Str->optimize(TB);
                break;
            case eDF_Uint:
                c->Uint->optimize(TB);
                break;
            case eDF_Bit: 
                c->Bit->optimize(TB);
                break;
            };
        }
    }

    /** 
     * Serialization to ostream 
     * @return true on success
    */
    bool Serialize(ostream& os) 
    {

        BM_DECLARE_TEMP_BLOCK(TB)

        bm::serializer<TBitType> bit_serializer;
        bit_serializer.byte_order_serialization(false);
        bit_serializer.gap_length_serialization(false);
        TBitType::statistics bit_st;
        
        rsc_svector_u32 offset_table(bm::use_null);

        bm::sparse_vector_serial_layout<TStrType> str_lay;
        bm::sparse_vector_serializer<TStrType> str_serializer;


        bm::sparse_vector_serial_layout<TUintType> uint_lay;
        bm::sparse_vector_serializer<TUintType> uint_serializer;

#ifdef __USE_BM_XOR_COMPRESSION__        
        str_serializer.enable_xor_compression();
        uint_serializer.enable_xor_compression();
#endif        

        vector<unsigned char> buffer;
        uint64_t offset = 0;
        for (auto& c : m_Columns) {
            offset_table.set(offset, c->type);
            switch (c->type) {
            case eDF_String:
            {
                c->Str->optimize(TB);
                str_serializer.serialize(*c->Str, str_lay);
                //bm::sparse_vector_serialize(*c->Str, str_lay, TB);
                const unsigned char* buf = str_lay.data();
                uint32_t sz = str_lay.size();
                offset += sz;
                if (!os.write((char*)&buf[0], sz))
                    return false;

                break;
            }
            case eDF_Uint:
            {
                c->Uint->optimize(TB);
                uint_serializer.serialize(*c->Uint, uint_lay);
                //bm::sparse_vector_serialize(*c->Uint, uint_lay, TB);
                const unsigned char* buf = uint_lay.data();
                uint32_t sz = uint_lay.size();                
                offset += sz;
                if (!os.write((char*)&buf[0], sz))
                    return false;
                break;
            }
            case eDF_Bit: {
                c->Bit->optimize(TB);
                c->Bit->calc_stat(&bit_st);
                buffer.resize(bit_st.max_serialize_mem);
                bit_serializer.serialize(*c->Bit, &buffer[0], bit_st.max_serialize_mem); 
                uint32_t sz = buffer.size();
                offset += sz;
                if (!os.write((char*)&buffer[0], sz))
                    return false;
                break;
            }
            };
        }
        // now write offset table
        {
            bm::sparse_vector_serializer<rsc_svector_u32> rsc_serializer;
            bm::sparse_vector_serial_layout<rsc_svector_u32> rsc_lay;
            rsc_serializer.serialize(offset_table, rsc_lay);
            const unsigned char* buf = rsc_lay.data();
            if (!os.write((char*)&buf[0], rsc_lay.size()))
                return false;
        }
        // offset points to the begining of offset table
        // write it as last 8 bytes
        if (!os.write(reinterpret_cast<const char*>(&offset), sizeof(offset))) 
            return false;

        return true;
    }

    /** 
     * Deserialization from string starting from pos
     * returns true on success
    */
    bool Deserialize(const char* data, size_t data_size) {

        m_Columns.clear();

        BM_DECLARE_TEMP_BLOCK(TB)
        bm::sparse_vector_deserializer<TStrType> str_deserializer;
        bm::sparse_vector_deserializer<TUintType> uint_deserializer;
        bm::deserializer<TBitType, bm::decoder> bit_deserializer;

    
        // last 8 bytes contain the beginning of the offset table 
        uint64_t offset_table_start = *(uint64_t*)&data[data_size - sizeof(offset_table_start)]; 
        rsc_svector_u32 offset_table(bm::use_null);
        sparse_vector_deserialize(offset_table, (const unsigned char*)&data[offset_table_start], TB);

        // iterate through offset table to get blob types and offsets 
        bm::interval_enumerator<bm::bvector<>> ien(*offset_table.get_null_bvector());
        auto offset_it = offset_table.get_const_iterator(0);
        do {
            size_t offset = ien.start();
            offset_it.go_to(offset);
            auto type = offset_it.value();
            unique_ptr<Column> c;
            switch ((EDF_ColumnType)type) {
                case eDF_String: {
                    unique_ptr<TStrType> v(new TStrType);
                    c.reset(new Column(v));
                    str_deserializer.deserialize(*c->Str, (const unsigned char*)&data[offset]);
                    break;
                }
                case eDF_Uint:
                {
                    unique_ptr<TUintType> v(new TUintType);
                    c.reset(new Column(v));
                    uint_deserializer.deserialize(*c->Uint, (const unsigned char*)&data[offset]);
                    break;
                }
                case eDF_Bit: {
                    unique_ptr<TBitType> v(new TBitType);
                    c.reset(new Column(v));
                    bit_deserializer.deserialize(*c->Bit, (const unsigned char*)&data[offset]);
                    break;
                }
            }
            m_Columns.emplace_back(move(c));

        } while (ien.advance());
        return true;
    }

    /** 
     * Dataframe comparsion for testing
    */
      
    bool Equal(const CDataFrame& df) const 
    {
        const auto& cols = df.GetColumns();
        if (m_Columns.size() != cols.size()) 
            return false;
        for (size_t i = 0; i < m_Columns.size(); ++i) {
            if (!m_Columns[i]->Equal(*cols[i]))
                return false;
        }
        return true;
    }

protected:
    TColumns m_Columns;
};

#endif /* __GUI_DATA_FRAME_HPP_ */

