#ifndef __GUI_SEQ_DICTIONARY_HPP_
#define __GUI_SEQ_DICTIONARY_HPP_

/*  
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrei Shkeda
 *
 * File Description:
 *
 */

#include <util/bitset/bm.h>
#include <util/bitset/bmsparsevec.h>
#include <util/bitset/bmstrsparsevec.h>
#include <util/bitset/bmsparsevec_serial.h>
#include <util/bitset/bmsparsevec_compr.h>
#include <util/bitset/bmsparsevec_algo.h>
#include <util/bitset/bmintervals.h>

#include <vector>
#include <ostream>
#include <memory>
#include <algorithm>
#include <util/checksum.hpp>

using namespace std;

 

#ifndef __USE_BM_XOR_COMPRESSION__
#define __USE_BM_XOR_COMPRESSION__
#endif


template<bool T> 
struct TSeqAlphabet
{
    static const char chars[29];
};

// Default dictionary alphabet
// with DNA letters (ACGTN) are placed first 
// As in 'in the beginning there was DNA'"
template<bool T> 
const char TSeqAlphabet<T>::chars[29] = {
    'A', 'C', 'G', 'T', 'N', 'B', 'D', 'E', 'F', 'H', 'I', 'J', 'K', 
    'L', 'M', 'O', 'P', 'Q', 'R', 'S', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '*', '-', '?' };


template<typename>
class CSeqDictionaryBuilder;

/*!
    Succinct Sequence Dictionary
    supports compressed list of terms 
    A term's set of characters is specified by Alphabet parameter
    Aplhabet is a static structure that is expected to have 'chars' member which is an array of chars.
    The index of a character is used as (unisgned)code used by the class in storing the dictionary terms. 
    For example, when default TSeqAlphabet is used, 'A' will be stored with a code '0', 'C' - code '1', etc.

    CSeqDictionaryBuilder should be used to create a CSeqDictionary
    CSeqDictionaryBuilder returns term indexes (TIndex) which are expected to be stored by the user separately

    TIndex is 64-bit unsigned number: 
    left 32-bit value is an offset in the term's list 
    right 32-bit values is the length of the term 

    In addition, there are two special TIndex values 
    that don't reference an entry in the term list:

    'Char' TIndex represents a single letter terms 
    left 32-bit value is a TAlphabet code of the character 
    right 32-bit values is set to zero 

    'Region' TIndex represents a region of a specific len (e.g. MSA Unaligned regions)
    left 32-bit value is set to zero 
    right 32-bit values is the length of the region 


*/
template<typename Alphabet = TSeqAlphabet<true>>
class CSeqDictionary 
{
public:
    typedef unsigned long long TIndex;
    typedef bm::sparse_vector<unsigned, bm::bvector<> > svector_u32; 

    CSeqDictionary() 
    {
    }

    /// returns Sequence by Index
    string& GetSequence(TIndex index, string& seq) const
    {
        seq.clear();
        if (IsRegion(index) == false) {
            uint32_t len = x_GetLen(index); 
            uint32_t offset = x_GetOffset(index);
            if (len == 0) {
                seq.append(1, Alphabet::chars[(unsigned)offset]);
            } else {
                auto it = m_Terms.get_const_iterator(offset);
                for (size_t i = 0; i < len && it.valid(); ++i) {
                    seq.append(1, Alphabet::chars[it.value()]);
                    it.advance();
                }
            }
        }
        return seq;
    }

    /// Checks if index references an actual sequence or it just contains a length of unaligned region
    static bool IsRegion(TIndex index) 
    {
        return x_GetOffset(index) == 0 && x_GetLen(index) > 0;
    }

    /// Returns the sequence length encoded in the index 
    static size_t GetSeqLength(TIndex index) 
    {
        uint32_t len = x_GetLen(index);
        return len == 0 ? 1 : len;
    }

    /** 
     * Serialization to ostream 
     * @return true on success
    */

    bool Serialize(ostream& os) 
    {
        BM_DECLARE_TEMP_BLOCK(TB)
        bm::sparse_vector_serial_layout<svector_u32> sv_lay;
        bm::sparse_vector_serializer<svector_u32> sv_serializer;

#ifdef __USE_BM_XOR_COMPRESSION__
        sv_serializer.enable_xor_compression();
#endif        

        m_Terms.optimize(TB);
        sv_serializer.serialize(m_Terms, sv_lay);
        const unsigned char* buf = sv_lay.data();
        uint32_t sz = sv_lay.size();
        if (!os.write(reinterpret_cast<const char*>(&sz), sizeof(sz))) 
            return false;        
        if (sz && !os.write((char*)&buf[0], sz))
            return false;
        return true;
    }

    /** 
     * Deserialization from string starting from pos
     * returns number of deserilized bytes 
    */
    size_t Deserialize(const char* data) 
    {
        size_t pos = 0;
        uint32_t sz = *(uint32_t*)data; 
        pos += sizeof(sz); 
        if (sz) {
            bm::sparse_vector_deserializer<svector_u32> sv_deserializer;
            sv_deserializer.deserialize(m_Terms, (const unsigned char*)&data[pos]);
            pos += sz;
        }
        return pos;
    }

protected:
    /// Index mask functions
    static 
    TIndex x_SetIndex(uint32_t offset, uint32_t len) 
    {
        TIndex idx = offset;
        idx <<= 32;
        idx |= len;
        return idx;
    }

    static 
    uint32_t x_GetLen(TIndex index) 
    {
        return  uint32_t(index); // index & 0x00000000FFFFFFFF
    }

    static 
    uint32_t x_GetOffset(TIndex index) 
    {
        index >>= 32;
        return (uint32_t)index;
    }

    static
    TIndex x_SetIndexAsChar(uint32_t char_code) 
    {
        return x_SetIndex(char_code, 0);
    }

    static 
    TIndex x_SetIndexAsRegion(uint32_t len)
    {
        return x_SetIndex(0, len);
    }

private:
    svector_u32 m_Terms;     /// the list of terms
    template<typename> friend class CSeqDictionaryBuilder;
};

static const unsigned kBAD_CODE = 0xFF;

/*!
    Sequence Dictionary Builder is used to build CSeqDictionary
    ...
    CSeqDictionaryBuilder<> builder;
    auto acgt_idx = builder.AddSequence("ACGT");
    auto r_index = builder.AddRegion(55);
    auto sd = builder.Build();
    ...

    BOOST_CHECK_EQUAL(sd->GetSequence(acgt_idx, s), "ACGT");
    BOOST_CHECK_EQUAL(sd->GetSeqLength(acgt_idx), 4);
    BOOST_CHECK(sd->IsRegion(r_index));
    BOOST_CHECK_EQUAL(sd->GetSeqLength(r_index), 55);

*/

template<typename Alphabet = TSeqAlphabet<true>>
class CSeqDictionaryBuilder
{
public:
    typedef CSeqDictionary<Alphabet> TSeqDict;
    typedef typename TSeqDict::TIndex TIndex;
    typedef typename TSeqDict::svector_u32 svector_u32;

    CSeqDictionaryBuilder() 
    {
        m_SeqPos.set_new_blocks_strat(bm::BM_GAP);
        m_CurrPos = 1;
        m_Coder.resize(256, kBAD_CODE);
        for (size_t i = 0;  i < sizeof(Alphabet::chars); ++i) {
            m_Coder[(size_t)Alphabet::chars[i]] = i;
        }
    }

    /// Add sequence and returns its index
    TIndex AddSequence(const string& seq) {
        TIndex index = 0;
        if (seq.length() == 1) {
            index = x_AddChar(seq[0]);
        } else {
            // calculate crc32 
            m_CheckSum.Reset(ncbi::CChecksum::eCRC32);
            m_CheckSum.AddLine(seq);
            auto crc32 = m_CheckSum.GetChecksum();
            if (m_SeqPos.test(crc32)) { // we've seen it already
                unsigned pos;
                // find that string and create its index 
                // if not found then it's CRC32 collision
                // and we have to register string 
                if (x_FindSequence(seq, pos)) {
                    index = TSeqDict::x_SetIndex(pos + 1, seq.length());
                } else {
                    index = x_AddSequence(seq);
                }
            } else {
                m_SeqPos.set(crc32); // register that we've seen that string already 
                index = x_AddSequence(seq);
            }
        }
        return index;
    }

    /// Create Region index
    TIndex AddRegion(uint32_t len) 
    {
        return TSeqDict::x_SetIndexAsRegion(len);
    }

    /// Builds the sequence dictionary from previously added sequences
    unique_ptr<TSeqDict> Build() 
    {
        unique_ptr<TSeqDict> seqdict(new TSeqDict());
        auto sz = m_SearchStr.size();
        size_t curr_pos = 0;
        seqdict->m_Terms[0] = 0;
        while (curr_pos < sz) {
            vector<unsigned> buffer(10 * 1024 * 1024 );
            size_t i = 0;
            for (;i < buffer.size() && curr_pos < sz; ++i, ++curr_pos) {
                buffer[i] = m_Coder[(unsigned)m_SearchStr[curr_pos]];
            }
            seqdict->m_Terms.import_back(&buffer[0], i);
        }
        return seqdict;

    }

private:
    bool x_FindSequence(const string& seq, unsigned& pos) 
    {
        auto seq_sz = seq.size();
        if (seq_sz <= 1)
            return false;
        m_SearchStr.push_back('\0');
        auto ptr = strstr(&m_SearchStr[0], seq.c_str());
        m_SearchStr.pop_back();
        if (ptr) {
            pos = ptr - &m_SearchStr[0];
            return true;
        }
        return false;
        
    }

    TIndex x_AddChar(char C) 
    {
        auto code = m_Coder[(unsigned)C];
        if (code == kBAD_CODE)
            throw runtime_error("The sequence contains bad symbol '" + string(1, C) + "'");
        return TSeqDict::x_SetIndexAsChar(code);
    }

    /// if necessary adds a ductionary sequence and returns its index 
    TIndex x_AddSequence(const string& seq) {
        TIndex idx = 0;
        auto seq_size = seq.size();
        _ASSERT(seq_size > 1);
        if (seq_size == 0)
            return idx;
        vector<unsigned> seq_v(seq_size);
        for (size_t i = 0; i < seq_size; ++i) {
            auto code = m_Coder[(unsigned)seq[i]];
            if (code == kBAD_CODE) throw runtime_error("The sequence contains bad symbol '" + string(1, seq[i]) + "'");
            seq_v[i] = code;
        }
        idx = TSeqDict::x_SetIndex(m_CurrPos, seq_size);
        m_CurrPos += seq_size; 
        copy(seq.begin(), seq.end(), back_inserter(m_SearchStr));
        return idx;
    }

    vector<unsigned> m_Coder;
    ncbi::CChecksum m_CheckSum;
    // where index = CRC32 of a dictionary string 
    // if string X CRC32 is 12345 then m_SeqPos[12345] == 1.
    bm::bvector<> m_SeqPos;
    size_t m_CurrPos = 0;
    vector<char> m_SearchStr;
};


#endif /* __GUI_SEQ_DICTIONARY_HPP_ */

