#ifndef GUI_WIDGETS___LOADERS___COLUMNAR_VCF_READER__HPP_
#define GUI_WIDGETS___LOADERS___COLUMNAR_VCF_READER__HPP_
/*  $Id: columnar_vcf_reader.hpp 45881 2021-01-05 03:10:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrea Asztalos, Anatoliy Kuznetsov
 *
 * File Description:
 *   
 *
 */


#include <corelib/ncbistd.hpp>
#include <corelib/ncbiobj.hpp>
#include <corelib/ncbimisc.hpp>
#include <gui/gui_export.h>
#include <util/icanceled.hpp>
#include <gui/widgets/loaders/columnar_vcf_variants.hpp>
#include <unordered_map>
#include <functional>
#include <mutex>
#include <utility>


BEGIN_NCBI_SCOPE

class CVCFVariantList;
BEGIN_SCOPE(objects)
    class CSeq_id;
    class ILineErrorListener;
    class CObjReaderLineException;
END_SCOPE(objects);

struct SVcfFieldData;

/// Class responsible to load VCF files 
class NCBI_GUIWIDGETS_LOADERS_EXPORT CColumnarVCFReader
{
public:
    using TFieldVector = vector<CConstRef<SVcfFieldData>>;
    using TSeqIdVarsListPair = pair<CConstRef<objects::CSeq_id>, CRef<CVCFVariantList>>;
    /// Defines a callable object, used when a variants list is processed by the reader
    /// @param[in] Reference to the variants lst.
    using TOnVCFVariantListReady = std::function<void(CVCFVariantList&)>;
    using TReportProgress = function<void(const string&)>;

    CColumnarVCFReader() {}
    virtual ~CColumnarVCFReader() {}

    // all reading methods operate on the same input stream
    // reads only the header section of the file
    bool ReadHeader(CNcbiIstream& in, ICanceled* canceled = nullptr, objects::ILineErrorListener* listener = nullptr);
    
    void LoadAllInfo(bool value) { m_LoadAllInfo = value; }
    set<string>& LoadSelectedInfoFields() { return m_LoadInfoFields;  }

    void LoadAllSamples(bool value) { m_LoadAllSamples = value; }
    map<unsigned, string>& LoadSelectedSamples() { return m_LoadSamples; }

    /// Reads only the data section of the file
    /// @param[in] in
    ///     Reference to an input stream.
    /// @param[in] canceled
    ///     Optional pointer to check for cancelation requests.
    /// @param[in] listener
    ///     Optional pointer to report VCF syntax errors.
    /// @param[in] on_variants_list_ready
    ///     Optional function, called when a VCF variants list becomes available.
    /// @return
    ///     Returns false if the reading was canceled, true otherwise.
    ///
    /// The reader has two modes of operation:
    /// - Events based - when on_variants_list_ready is set to a callable object, the reader calls it for every variants list read from the input and then releases the list;
    /// - DOM mode - all variants lists are loaded in memory and can be obtained by calling @ref GetVariantsForChr(const string&).
    bool ReadData(CNcbiIstream& in, ICanceled* canceled = nullptr, objects::ILineErrorListener* listener = nullptr, TReportProgress prog_func = TReportProgress(), TOnVCFVariantListReady on_variants_list_ready = TOnVCFVariantListReady());
    
    /// Reads a list of variants
    /// @param[in] in
    ///     Reference to an input stream.
    /// @param[in] chr_list
    ///     Map of seq-id to a list of synonyms.
    /// @param[in] canceled
    ///     Optional pointer to check for cancelation requests.
    /// @param[in] listener
    ///     Optional pointer to report VCF syntax errors.
    /// @param[in] on_variants_list_ready
    ///     Optional function, called when a VCF variants list becomes available.
    /// @return
    ///     Returns pairs of seq-id - variants list
    ///
    /// The reader has two modes of operation:
    /// - Events based - when on_variants_list_ready is set to a callable object, the reader calls it for every variants list read from the input and then releases the list;
    /// - DOM mode - all variants lists are loaded in memory and returned. They can be obtained by calling @ref GetVariantsForChr(const string&). 
    vector<TSeqIdVarsListPair>
        ReadVariantsForChrs(CNcbiIstream& in, 
            const vector<pair<CConstRef<objects::CSeq_id>, vector<string>>>& chr_list,
            ICanceled* canceled = nullptr, 
            objects::ILineErrorListener* listener = nullptr,
            TReportProgress prog_func = TReportProgress(),
            TOnVCFVariantListReady on_variants_list_ready = TOnVCFVariantListReady());

    /// Returns a vector, holding the chrs/contigs identifiers, read from the file
    /// @return
    ///     Returns false if the reading was canceled, true otherwise.
    /// @note
    ///     Use these function when the entire file has been read.
    vector<string> GetChromosomeNames() const;
    /// Retrieves the variants list for a given chr/contig
    /// @param[in] in
    ///     String reference specifying the chr/contig
    /// @return
    ///     Returns a pointer to the variants list for a given chr/contig.
    /// @note
    ///     Use these function when the entire file has been read.
    CRef<CVCFVariantList> GetVariantsForChr(const string& chr_name) const;

    // use these functions after ReadHeader() was called
    const string& GetVCFversion() const { return m_VCFversion; }
    const string& GetAssembly() const { return m_Assembly; }
    const set<CConstRef<SVcfFieldData>>& GetInfoFields() const { return m_InfoFields; }
    const map<unsigned, string>& GetSamples() const { return m_SampleCols;  }

protected:
    void x_InterruptReading();
    void x_ResetInfo();

    unsigned x_ProcessHeaderLine(const string& header_line, unsigned line_nr, objects::ILineErrorListener* listener);
    void x_GatherSampleColNames(const string& header_line, objects::ILineErrorListener* listener, unsigned line_nr);
    void x_GetSamplesToLoad(const string& header_line, objects::ILineErrorListener* listener, unsigned line_nr);

    void x_ProcessWarning(objects::CObjReaderLineException& err, objects::ILineErrorListener* error_cont);
    void x_ProcessError(objects::CObjReaderLineException& err, objects::ILineErrorListener* error_cont);
    void x_ProcessCriticalError(objects::CObjReaderLineException& err, objects::ILineErrorListener* error_cont);

    // used when reading the header
    string m_VCFversion;
    string m_Assembly;
    string m_ReferenceLine;
    ///List of INFO fields parsed from the Header of the file
    set<CConstRef<SVcfFieldData>> m_InfoFields;
    map<unsigned, string> m_SampleCols;  // order is important!
    
    // used when reading the data section
    bool m_LoadAllInfo{ true };  ///< Flag to load every INFO field
    set<string> m_LoadInfoFields;  ///< List of INFO fields required to be loaded

    bool m_LoadAllSamples{ true }; ///< Flag to load every SAMPLE column
    map<unsigned, string> m_LoadSamples;  ///< List of SAMPLES required to be loaded
    unordered_map<string, CRef<CVCFVariantList>> m_ChromosomeMap;
    std::mutex  m_ChromosomeMapMutex;
};

struct NCBI_GUIWIDGETS_LOADERS_EXPORT SVcfFieldData : public CObject
{
    SVcfFieldData(const string& line);
    SVcfFieldData(const string& name, const string& descr, const string& nr)
        : m_Name(name), m_Description(descr), m_Number(nr) {}

    string m_Name;
    string m_Description;
    string m_Number;
};

// Testing the reader
class NCBI_GUIWIDGETS_LOADERS_EXPORT CColumnarVCFReaderTest : public CColumnarVCFReader
{
public:
    // These two functions also output memory footprint and timings
    void SerializeToDisk(const string& prefix, CNcbiOstream* out = nullptr);
    void Deserialize(const string& prefix, CNcbiOstream* out = nullptr);

    void ListColumns(CNcbiOstream& out, bool only_sv_cols = false);
    void ListIndexVectors(CNcbiOstream& out);
    void GetStatistics(CNcbiOstream& out);

};

END_NCBI_SCOPE


#endif //GUI_WIDGETS___LOADERS___COLUMNAR_VCF_READER__HPP_
