#ifndef GUI_WIDGETS___LOADERS___COLUMNAR_VCF_VARIANTS__HPP_
#define GUI_WIDGETS___LOADERS___COLUMNAR_VCF_VARIANTS__HPP_
/*  $Id: columnar_vcf_variants.hpp 45881 2021-01-05 03:10:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrea Asztalos, Anatoliy Kuznetsov
 *
 * File Description:
 *   
 *
 */


#include <corelib/ncbistd.hpp>
#include <corelib/ncbiobj.hpp>
#include <corelib/ncbimisc.hpp>
#include <gui/gui_export.h>
#include <util/icanceled.hpp>
#include <util/range.hpp>
#include <gui/widgets/loaders/columnar_vcf_variants_base.hpp>
#include <unordered_map>


BEGIN_NCBI_SCOPE

// Contains all information about every variant associated with a
// specific chromosome name, read from a VCF file
class NCBI_GUIWIDGETS_LOADERS_EXPORT CVCFVariantList : public CVCFVariantsBase
{
public:
    CVCFVariantList(const string& chr_name, 
        bool load_all_info = true, 
        const set<string>& info_fields = set<string>(),
        const map<unsigned, string>& sample_cols = map<unsigned, string>());

    // constructor to initialize from a buffer and deserializes all data
    CVCFVariantList(const string& chr_name, const vector<char>& data);
    CVCFVariantList(const string& chr_name, const string& filename);

    ~CVCFVariantList();

    void ParseLine(const string& line);
    void FinalizeReading();

    virtual bool GetPositionsForVariant(const string& variant_id, vector<unsigned>& positions);
    virtual void GetPositionsForMissingVarID(vector<unsigned>& positions);

    const string& GetChrName() const { return m_ChrName; }

    const vector<char>& GetSerializedData() const;
    void WriteSerializedData(const string& filename);

    // returns true if all variant ids except the missing variant id (".") are unique
    bool AreVariantIdsUnique(vector<pair<string, unsigned>>& copies);

    void List(CNcbiOstream& out, bool only_sv_cols = false) const;
    void ListSamples(CNcbiOstream& out) const;

    // duplicates are included in the list
    vector<string> GetAllVariantIDS() const;

    bool operator==(const CVCFVariantList& other) const;

    CVCFVariantList(const CVCFVariantList&) = delete;
    CVCFVariantList& operator=(const CVCFVariantList&) = delete;

public:
    // Functions used for TESTING
    // serializes/deserializes column data into separate files
    bool SerializeVariantData(const string& prefix, CNcbiOstream* out = nullptr);
    bool DeserializeAndCheck(const string& prefix, CNcbiOstream* out = nullptr);
    bool RemoveSerializedOutput(const string& prefix);

    void GetStatistics(CNcbiOstream& out);

private:
    string x_GetFilePrefix(const string& prefix) const;

    string m_ChrName;
    size_t m_Index{ 0 };  // aids in reading the variants line by line, after that it's set to 0
};


class NCBI_GUIWIDGETS_LOADERS_EXPORT CVCFSlicedVariants : public CVCFVariantsBase
{
public: 
    CVCFSlicedVariants(const vector<char>& data, 
        const TSeqRange* range = nullptr, 
        const set<string>& cols_to_decode = set<string>());

    ~CVCFSlicedVariants() {}

    void SetRange(const TSeqRange& range);
    void SetRange(const TSeqPos from, const TSeqPos to);

    virtual bool GetPositionsForVariant(const string& variant_id, vector<unsigned>& positions);
    virtual void GetPositionsForMissingVarID(vector<unsigned>& positions);

    CVCFSlicedVariants(const CVCFSlicedVariants&) = delete;
    CVCFSlicedVariants& operator=(const CVCFSlicedVariants&) = delete;
private:
    const unsigned char* m_BufferPtr{ nullptr };
    size_t m_NrCols{ 0 };  ///< number of data columns to be deserialized
    set<string> m_ColsDecode;    ///< the name of data columns to be deserialized
};

END_NCBI_SCOPE


#endif //GUI_WIDGETS___LOADERS___COLUMNAR_VCF_VARIANTS__HPP_
