#ifndef GUI_WIDGETS___LOADERS___COLUMNAR_VCF_VARIANTS_BASE__HPP_
#define GUI_WIDGETS___LOADERS___COLUMNAR_VCF_VARIANTS_BASE__HPP_
/*  $Id: columnar_vcf_variants_base.hpp 46153 2021-01-26 00:00:32Z evgeniev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrea Asztalos, Anatoliy Kuznetsov
 *
 * File Description:
 *   
 *
 */


#include <corelib/ncbistd.hpp>
#include <corelib/ncbiobj.hpp>
#include <corelib/ncbimisc.hpp>
#include <gui/gui_export.h>
#include <objects/gbproj/AbstractProjectItem.hpp>
#include <util/icanceled.hpp>
#include <util/range.hpp>
#include <util/bitset/ncbi_bitset.hpp>
#include <util/bitset/bmstrsparsevec.h>
#include <util/bitset/bmsparsevec_compr.h>
#include <util/bitset/bmsparsevec_algo.h>
#include <util/bitset/bmsparsevec_serial.h>
#include <unordered_map>


BEGIN_NCBI_SCOPE

class NCBI_GUIWIDGETS_LOADERS_EXPORT CVariantDescriptors
{
public:
    using bvector_type = bm::bvector<>;
    using TSparseStrVector = bm::str_sparse_vector<char, bvector_type, 390>;
   
    // as in Optional: INFO fields, SAMPLE columns
    using TSparseOptVector = bm::str_sparse_vector<char, bvector_type, 200>;

    CVariantDescriptors(bool load_all_info = true, 
        const set<string>& info_fields = set<string>(), 
        const map<unsigned, string>& sample_cols = map<unsigned, string>());
    ~CVariantDescriptors() {}

    void InitInserters();
    bool AreInsertersReady() const { return (!m_BackInserters.empty()); }

    /// Push back 'value' into the vector identified by 'label'
    /// The 'value' is not actually stored in the vector until the vector's insert iterator is flushed
    void PushBack(const string& label, const string& value);

    void FinalizeReading();
    size_t& SetIndex() { return m_Index; }

    TSparseStrVector& operator[] (const string& label);
    const TSparseStrVector& operator[] (const string& label) const;

    vector<string> GetInfoFieldNames() const;
    const TSparseOptVector& GetInfoField(const string& field_name) const;
    TSparseOptVector& SetInfoField(const string& field_name);
    void SetNumberofInfoFields(size_t nr_info);

    vector<string> GetSampleNames() const;
    const TSparseOptVector& GetSample(const string& name) const;
    TSparseOptVector& SetSample(const string& name);
    void SetNumberofSampleCols(size_t nr_samples);

    bool GetIndicesForVariant(const string& variant_id, TSparseStrVector::bvector_type& indices) const;
    void GetIndicesForMissingVarID(TSparseStrVector::bvector_type& indices) const;

    bool operator==(const CVariantDescriptors& other) const;
    void Clear();

    CVariantDescriptors(const CVariantDescriptors&) = delete;
    CVariantDescriptors& operator=(const CVariantDescriptors&) = delete;

private:
    vector<unique_ptr<TSparseStrVector>> m_DescrMatrix;
    unordered_map<string, size_t> m_DescrLabels;  //TODO - combine the two

    using TInserter = TSparseStrVector::back_insert_iterator;
    vector<unique_ptr<TInserter>> m_BackInserters;

    using TOptInserter = TSparseOptVector::back_insert_iterator;
    unordered_map<string, unique_ptr<TSparseOptVector>> m_InfoFields;
    unordered_map<string, unique_ptr<TOptInserter>> m_InfoInserters;

    vector<pair<string, unique_ptr<TSparseOptVector>>> m_SampleFields;
    vector<unique_ptr<TOptInserter>> m_SampleInserters;
    unordered_map<string, size_t> m_SampleIndices;

    // used when loading data
    size_t m_Index{ 0 };
    bool m_LoadAllInfo{ true };
    set<string> m_LoadInfoFields;
    map<unsigned, string> m_LoadSamples;

    mutable bm::sparse_vector_scanner<TSparseStrVector> m_Scanner;
};


class NCBI_GUIWIDGETS_LOADERS_EXPORT CPosToIndex
{
public:
    using svector = bm::sparse_vector<unsigned, CVariantDescriptors::bvector_type>;
    using TIndexVector = bm::rsc_sparse_vector<unsigned, svector>;
    using TSparseStrVector = CVariantDescriptors::TSparseStrVector;

    enum EType {
        eMin,
        eMax
    };

    CPosToIndex();
    ~CPosToIndex() {}

    bool Add(const unsigned& index, const unsigned& value);
    void FinalizeReading();

    void Lookup(const TSparseStrVector::bvector_type& values, vector<unsigned>& indices) const;
    
    // mapping functions between genomic coordinates and VCF line coordinates (vector indices)
    size_t GetIndexForPosition(const unsigned& pos, CPosToIndex::EType type) const;
    unsigned GetPositionForIndex(const size_t& index) const;

    unsigned GetNumberOfIndexVecs() const { return m_Indices.size(); }
    bool operator==(const CPosToIndex& other) const;

public:
    // Functions used for TESTING
    void SerializeVectors(const string& prefix, CNcbiOstream* out, unsigned& cum_memory_used, unsigned& cum_layout_size);
    void DeserializeVectors(const string& prefix, CNcbiOstream* out);
    bool RemoveSerializedOutput(const string& prefix);

    void ListVectors(CNcbiOstream& out) const;
    void GetStatistics(bm::bv_statistics& stat_sum, bool Jaccard_index, CNcbiOstream& out);
    const size_t& GetMaxIndex() const { return m_MaxStoredIndex; }

    const vector<unique_ptr<TIndexVector>>& GetIndexVectors() const { return m_Indices; }

    CPosToIndex(const CPosToIndex&) = delete;
    CPosToIndex& operator=(const CPosToIndex&) = delete;

private:
    string x_GetFilePrefix(const string& prefix) const;
    void x_SaveMaximumIndex();

    vector<unique_ptr<TIndexVector>> m_Indices;

    using TInserter = TIndexVector::back_insert_iterator;
    vector<unique_ptr<TInserter>> m_BackInserters;

    // vector used for reading in positions
    vector<unsigned> m_LastIndices;

    size_t m_MaxStoredIndex{ 0 };

    mutable bm::sparse_vector_scanner<TIndexVector> m_Scanner;

    friend class CVCFVariantsBase;
};

class NCBI_GUIWIDGETS_LOADERS_EXPORT CVCFVariantsBase : public CObject
{
public:
    using TSparseStrVector = CVariantDescriptors::TSparseStrVector;
    using TSparseOptVector = CVariantDescriptors::TSparseOptVector;
    using TIndexVector = CPosToIndex::TIndexVector;
    using svector = CPosToIndex::svector;

    static const string sm_MissingValue;
    static const string sm_ID;
    static const string sm_REF;
    static const string sm_ALT;
    static const string sm_QUAL;
    static const string sm_FILTER;
    static const string sm_INFO;
    static const string sm_FORMAT;
    static const string sm_SAMPLES;
    static const vector<string>& s_GetColNames();     // does not contain sm_INFO and sm_SAMPLES
    static const vector<string>& s_GetAllColNames();  // contains sm_INFO, sm_SAMPLES

    virtual ~CVCFVariantsBase();

    const CVariantDescriptors& GetDescriptors() const { return m_Descriptors; }

    unsigned GetPositionForIndex(const size_t& index) const;
    size_t GetIndexForPosition(const unsigned& pos, CPosToIndex::EType type) const;

    virtual bool GetPositionsForVariant(const string& variant_id, vector<unsigned>& positions) = 0;
    virtual void GetPositionsForMissingVarID(vector<unsigned>& positions) = 0;

    // returns the position of the first and the last variant in genomic coordinates
    TSeqPos GetStart() const { return m_StartPos; }
    TSeqPos GetStop() const { return m_StopPos; }

    // values that appear multiple times are also counted for
    unsigned GetNumberOfVariants() const;
    unsigned Count(const TSeqRange& range);

    // faster version
    void GetHistogram(svector& hist, const svector::size_type& bin_width) const;
    // slower version
    void GetHistogram(TIndexVector& hist, const svector::size_type& bin_width);

    bool operator==(const CVCFVariantsBase& other) const;

    void ListPositionVectors(CNcbiOstream& out) const;

    static string s_GetPreviousVersion();
    static string s_GetCurrentVersion(bool format_only = false);
    static bool s_ConvertFromPreviousVersion(objects::CAbstractProjectItem& prj_item);

    CVCFVariantsBase(const CVCFVariantsBase&) = delete;
    CVCFVariantsBase& operator=(const CVCFVariantsBase&) = delete;

public:
    // Functions used for TESTING
    unsigned GetNumberOfIndexVecs() const { return m_Posindexmap.GetNumberOfIndexVecs(); }
    const size_t& GetMaxIndex() const { return m_Posindexmap.GetMaxIndex(); }

protected:
    CVCFVariantsBase() {}
    CVCFVariantsBase(bool load_info, const set<string>& info_fields, const map<unsigned, string>& sample_cols);
    /* Data is serialized as:
       number of index vectors
       number of columns
       size of index_vec_0  <-- the index vectors are stored first
       index_vec_0 data
       ...
       size of index_vec_n
       index_vec_n data
       size of col_0        <-- followed by the data columns
       col_0 data
       ...
       size of col_n
       col_n data
       number of info columns  <--- followed by the info columns
       size of info_col_0_name
       name of info_col_0
       size of info_col_0
       info_col_0 data
       ...
       size of info_col_n_name
       name of info_col_n
       size of info_col_n
       info_col_n data    
       number of samples    <---- followed by the samples
       size of sample_0_name
       name of sample_0_name
       size of sample_0
       sample_0 data
       ...
       size of sample_n_name
       name of sample_n_name
       size of sample_n
       sample_n data

    */
    void x_SerializeData() const;
    void x_DeserializeAllData();

    void x_DeserializeIndexVectors(const unsigned char* &buf_ptr, size_t& nr_cols);
    // deserializes all data into succint vectors
    void x_DeserializeDescr_All(const unsigned char* buf_ptr, const size_t& nr_cols);
    void x_DeserializeDescr_Range(const unsigned char* buf_ptr, const size_t& nr_cols, 
        const TSeqRange* range = nullptr, const set<string>& cols_to_decode = set<string>());
    // this function does not handle the INFO and SAMPLES columns
    void x_DeserializeColumn(const string& col_name, const unsigned char* buf_ptr, const size_t& nr_cols);

    CVariantDescriptors m_Descriptors;
    CPosToIndex m_Posindexmap;

    TSeqPos m_StartPos{ 0 }; ///< in genomic coordinates (1-based)
    TSeqPos m_StopPos{ 0 };  ///< in genomic coordinates (1-based)

    mutable vector<char>* m_Buffer{ nullptr };

    bm::sparse_vector_deserializer<TSparseStrVector> m_Deserializer;
    bm::sparse_vector_deserializer<TSparseOptVector> m_OptDeserializer;
};

END_NCBI_SCOPE


#endif 
    //GUI_WIDGETS___LOADERS___COLUMNAR_VCF_VARIANTS_BASE__HPP_
