#ifndef GUI_WIDGETS___LOADERS___DATA_CRAWLER_CACHE__HPP_
#define GUI_WIDGETS___LOADERS___DATA_CRAWLER_CACHE__HPP_
/*  $Id: data_crawler_cache.hpp 46098 2021-01-22 21:44:41Z evgeniev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladislav Evgeniev
 *
 * File Description:
 *   
 *
 */

#include <connect/services/neticache_client.hpp>
#include <memory>
#include <string>
#include <atomic>
#include <thread>
#include <gui/gui_export.h>

BEGIN_NCBI_SCOPE

/////////////////////////////////////////////////////////////////////////////
/// Base class for data crawler cache objects
class NCBI_GUIWIDGETS_LOADERS_EXPORT CDataCrawlerCache
{
public:
    /// Cache object states
    enum ECacheState {
        eState_None,            /// No cache data
        eState_InProgress,      /// The remote file is being processed
        eState_NotAccessible,   /// The remote file is not accessible
        eState_Created          /// The remote file was processed and its data is in the cache
    };
    /// Default constructor
    CDataCrawlerCache() = default;

    /// Gets the state of the cache object.
    /// @return
    ///     The current state
    ECacheState GetState() const;
    /// Changes the state of the cache object.
    /// @param[in] desired_state
    ///     The desired state to switch to.
    /// @param[out] actual_state
    ///     The actual state of the cache object.
    /// @return
    ///     True if the state was changed successfully, false otherwise.
    bool SetState(ECacheState desired_state, ECacheState &actual_state);

    /// Returns stream interface to read BLOB data.
    /// @return
    ///     Smart pointer to read stream.
    std::unique_ptr<CNcbiIstream> GetReadStream();
    /// Returns stream interface to write BLOB data.
    /// @return
    ///     Smart pointer to write stream.
    std::unique_ptr<CNcbiOstream> GetWriteStream();
    /// Removes the cache entry.
    void Remove();

    /// Returns the NetCache key of the BLOB.
    /// @return
    ///     String, holding the NetCache key.
    const std::string& GetKey() const { return m_BlobKey; }

    /// Destructor
    virtual ~CDataCrawlerCache();
protected:
    /// Types of cache objects
    enum ECacheType {
        eCache_Blob,    /// BLOB object
        eCache_Sync     /// Synchronization object
    };

    /// Initializes the cache instances.
    /// @note Must be invoked by the constructors of the derived classes.
    void x_Initialize();

    /// Generates the cache key for a given type
    /// @param[in] type
    ///     The desired cache type.
    /// @return
    ///     String, holding the cache key.
    std::string x_GetCacheKey(ECacheType type);

    /// Creates the BLOB NetCache instance
    /// @return
    ///     The NetCache instance.
    virtual CNetICacheClient x_InitBlobCache() = 0;
    /// Creates the sync NetCache instance
    /// @return
    ///     The NetCache instance.
    virtual CNetICacheClient x_InitSyncCache() = 0;

    /// Generates the cache key
    /// @return
    ///     String, holding the cache key.
    virtual std::string x_GetCacheKey() = 0;


    /// Starts the heartbeat (thread to keep the sybc object alive)
    void x_StartHeartbeat();
    /// Stops the heartbeat (thread to keep the sybc object alive)
    void x_StopHeartbeat();
    /// The heartbeat thread function
    void x_Heartbeat();

    /// Creates a NetCache instance
    /// @param[in] cache
    ///     Name of the registry section to look for the configuration parameters in.
    /// @param[in] service
    ///     The service name.
    /// @param[in] db
    ///     The cache name.
    /// @return
    ///     The NetCache instance.
    CNetICacheClient x_InitICache(const std::string &cache, const std::string &service, const std::string &db);

    /// Reads the state of the sync object
    /// @return
    ///     The cache state, read from the sync object.
    ECacheState x_ReadState() const;

    /// Writes a new state to the sync object
    /// @param[in] state
    ///     The state to write.
    /// @param[in] ttl
    ///     Time to live in seconds. 60 seconds by default.
    void x_WriteState(ECacheState state, unsigned int ttl = 60);

private:
    /// The NetCache key of the BLOB object
    std::string m_BlobKey;
    /// The NetCache key of the sync object
    std::string m_SyncKey;
    /// Blobs NetCache instance
    mutable CNetICacheClient    m_BlobCache;
    /// Sync objects NetCache instance
    mutable CNetICacheClient    m_SyncCache;
    /// Boolean flag to stop the heartbeat thread
    std::atomic_bool    m_StopHeartbeat { false };
    /// The heartbeat thread
    std::unique_ptr<std::thread>    m_Heartbeat;
};

END_NCBI_SCOPE

#endif // GUI_WIDGETS___LOADERS___DATA_CRAWLER_CACHE__HPP_
