#ifndef GUI_WIDGETS_SEQ_GRAPHIC___VCF_DS__HPP
#define GUI_WIDGETS_SEQ_GRAPHIC___VCF_DS__HPP

/*  $Id: vcf_ds.hpp 45883 2021-01-05 17:57:04Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 * File Description:
 *
 */

#include <mutex>
#include <corelib/ncbimtx.hpp>
#include <gui/widgets/seq_graphic/feature_ds.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_job_result.hpp>
#include <gui/widgets/seq_graphic/feature_enums.hpp>
#include <gui/objutils/nonasn_track_data.hpp>

BEGIN_NCBI_SCOPE

class NCBI_GUIWIDGETS_SEQGRAPHIC_EXPORT CVcfTrackData :
    public CObject,
    public INonAsnTrackData
{
public:
    CVcfTrackData(const string &title, const string &annot_name, const TDataBlob &vcf_blob)
        : m_Title{ title }
        , m_Annots{ annot_name }
        , m_Blob{ vcf_blob }
    {}

    const string& GetKey() const { return m_Key; }

    const string& GetSubkey() const { return m_Subkey; }

    const string& GetDisplayName() const { return m_Title; }

    bool GetShown() const { return true;  }
    
    const string& GetCategory() const { return m_Categoty; }
    
    const string& GetSubcategory() const { return m_Subcategoty; }
    
    const TAnnots& GetAnnots() const { return m_Annots; }
    
    // Returns the actual data
    const TDataBlob& GetBlob() const { return m_Blob; }

    const string GetDataSourceTypeName() const;
	// performs some sort of a search in data and place results into a list of objects
	virtual void DoSearch(const objects::CSeq_id* seq_id, const string& sSearchTerm, list<CRef<CObject>>& results) const;

private:
    const string    m_Key{ "vcf_track" };
    const string    m_Subkey{ "VcfVariation" };
    const string    m_Title;
    const TAnnots   m_Annots;
    const string    m_Categoty{"Variations"};
    const string    m_Subcategoty{ "External Data" };
    const TDataBlob &m_Blob;
};

class NCBI_GUIWIDGETS_SEQGRAPHIC_EXPORT CVcfTrackDataFactory :
    public CObject,
    public INonAsnTrackDataFactory,
    public IExtension
{
public:
    /// @name INonAsnTrackDataFactory interface implementation
    /// @{
    INonAsnTrackData* CreateTrackData(const objects::CSeq_id &seq_id, const objects::CProjectItem &prj_item) const;
    /// @}

    /// @name IExtension interface implementation
    /// @{
    string GetExtensionIdentifier() const;
    string GetExtensionLabel() const;
    /// @}
};

///////////////////////////////////////////////////////////////////////////////
/// CVcfFeatureDS
///
class NCBI_GUIWIDGETS_SEQGRAPHIC_EXPORT CVcfFeatureDS : 
    public CSGFeatureDS,
    public INonAsnDataSource
{
public:
    using TDataBlob = INonAsnTrackData::TDataBlob;
    using TBlobsMap = map<string, const TDataBlob*>;

    CVcfFeatureDS(objects::CScope& scope, const objects::CSeq_id& id);

    virtual void LoadFeatures(objects::SAnnotSelector& sel,
                      const TSeqRange& range, TModelUnit window,
                      TJobToken token = -1, int max_feat = -1,
                      ELinkedFeatDisplay LinkedFeatDisplay = ELinkedFeatDisplay::eLFD_Default,
		              const string& = "");

    virtual void LoadFeatures(const TSeqRange& range, TModelUnit window,
        TJobToken token, const TFeatBatchJobRequests& requests) { NCBI_THROW(CException, eUnknown, "Batch loading not implemented for VCF datasource"); }

    virtual void LoadMainFeatures(
        objects::SAnnotSelector& lm_feat_sel,
        objects::SAnnotSelector& m_feat_sel,
        TJobToken lm_token, TJobToken main_token,
        int merge_style, bool landmark_feat,
        int min_overview_feat, const TSeqRange& range,
        TModelUnit window, int max_feat,
        bool show_histogram, int highlight_mode) { NCBI_THROW(CException, eUnknown, "Loading main features not implemented for VCF datasource"); }
    
    /// @name INonAsnDataSource implementation
    /// @{
    void SetNonAsnDataRegistry(const INonAsnDataRegistry *registry);
    bool HasNonAsnData(const string &annot_name) const;
    const TDataBlob& GetNonAsnDataBlob(const string &annot_name) const;
	virtual void DoSearch(const string &annot_name, const string& sSearchTerm, list<CRef<CObject>>& results) const;
	/// @}

private:
    CConstIRef<INonAsnDataRegistry> m_DataRegistry;
};


///////////////////////////////////////////////////////////////////////////////
/// CSGFeatureDSType
///
class NCBI_GUIWIDGETS_SEQGRAPHIC_EXPORT CVcfFeatureDSType :
    public CObject,
    public ISGDataSourceType,
    public IExtension
{
public:
    /// create an instance of the layout track type using default settings.
    virtual ISGDataSource* CreateDS(SConstScopedObject& object) const;

    /// @name IExtension interface implementation
    /// @{
    virtual string GetExtensionIdentifier() const;
    virtual string GetExtensionLabel() const;
    /// @}

    /// check if the data source can be shared.
    virtual bool IsSharable() const;
};


END_NCBI_SCOPE

#endif  /* GUI_WIDGETS_SEQ_GRAPHIC___VCF_DS__HPP */
