#ifndef GUI_WIDGETS_SEQ_GRAPHIC___VCF_GLYPH__HPP
#define GUI_WIDGETS_SEQ_GRAPHIC___VCF_GLYPH__HPP

/*  $Id: vcf_glyph.hpp 45881 2021-01-05 03:10:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio, Liangshou Wu
 *
 * File Description:
 *    CVcfGlyph -- utility class to arrange CSeq_feat objects in hierarchical
 *                (tree) order.
 */

#include <gui/gui.hpp>
#include <gui/objutils/label.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/snp_gui.hpp>
#include <gui/widgets/seq_graphic/seq_glyph.hpp>
#include <gui/widgets/seq_graphic/feature_conf.hpp>
#include <gui/widgets/seq_graphic/feature_enums.hpp>
#include <gui/widgets/seq_graphic/object_based_glyph.hpp>

#include <objects/seqfeat/Variation_inst.hpp>

#include <corelib/ncbiobj.hpp>

BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
/// Class for storing a set of nested features.
/// A feature describes the genetic location in true gene coordinates of any
/// sort of genetic object (gene / protein / RNA / etc).  Each feature has an
/// overall extent as well as a series of subregions
///
BEGIN_objects_SCOPE
class CUser_object;
END_objects_SCOPE


class NCBI_GUIWIDGETS_SEQGRAPHIC_EXPORT CVcfGlyph :
    public CSeqGlyph, public IObjectBasedGlyph
{
public:
    /// @name ctors.
    /// @{
    CVcfGlyph(objects::CSeq_id& seq_id,
		        const string& sID,
                size_t pos, ///< in VCF (i.e. 1-based coordinates)
                const string& ref_allele, 
                const string& alt_alleles);    ///< alternative alleles separated by /
    /// @}

    /// @name CSeqGlyph virtual method implementation.
    /// @{
    virtual bool NeedTooltip(const TModelPoint& /*p*/, ITooltipFormatter& /*tt*/, string& /*t_title*/) const;
    virtual void GetTooltip(const TModelPoint& p, ITooltipFormatter& tt, string& t_title) const;
    virtual void GetHTMLActiveAreas(TAreaVector* p_areas) const;
    virtual TSeqRange GetRange(void) const;
    virtual bool SetSelected(bool f);
    virtual void SetHideLabel(bool b);
    virtual bool HasSideLabel() const;
    virtual bool IsClickable() const;
    virtual bool LessBySeqPos(const CSeqGlyph& obj) const;
    virtual bool HitTestHor(TSeqPos x, const CObject *obj);
    /// @}
    /// @name IObjectBasedGlyph interface implementation (needed for object inspector and search selection to work)
    /// @{
    virtual const objects::CSeq_loc& GetLocation(void) const;
    virtual CConstRef<CObject> GetObject(TSeqPos pos) const;
    virtual void  GetObjects(vector<CConstRef<CObject> >& objs) const;
    virtual bool HasObject(CConstRef<CObject> obj) const;
    virtual string GetSignature() const;
    virtual const TIntervals& GetIntervals(void) const;
    /// @}
    /// Get the customized color for a given feature.
    /// Special color code.
    /// For features with Var_type and identity qualifier set,
    /// we render them using special color code:
    ///	Color dbVar features based on qual::Var_type
    ///     - Deletions/Loss=red
    ///     - Indels=red
    ///     - Gains/Insertion=green
    ///     - Inversions=gray
    ///	Color segmental duplication feature based on qual::identity
    ///     - Gray: 90-98%
    ///     - Yellow: 98-99%
    ///     - Orange:>99%
    /// @param color the returned customized color
    /// @return true if custom color is needed
    bool GetCustomColor(CRgbaColor& color) const;
    void GetCustomColor(size_t idx, CRgbaColor& color) const;

    /// determines whether the glyph should be drawn or not at all (currently children of unselected parents when m_LinkedFeat
    /// is set to "Expandable")
    bool isDrawn() const { return x_isDrawn(); }

    /// update the layout for everything encompassing this glyph
    void LayoutChanged();
    CConstRef<CFeatureParams> GetConfig() const;
    void SetConfig(CConstRef<CFeatureParams> conf);

    /// variation additional info on the right (alleles)
    void GetAdditionalInfo(string& sAdditionalInfo) const;
    /// retrieve label for a given type
    void GetLabel(string& label, CLabel::ELabelType type) const;
protected:
    virtual void x_Draw() const;
    virtual void x_UpdateBoundingBox();

	virtual void x_DrawHighlight() const;
	
	void x_MaybeDrawLabelAbove(TModelUnit& base) const;
    void x_DrawVariationBar(TModelUnit base) const;
    void x_DrawAdditionalInfo(TModelUnit base) const;
    void x_MaybeDrawLabelOthers(TModelUnit base) const;

    bool x_RedundantLabelCheck(const string& label) const;
    void x_DrawInsertion(TModelUnit SeqPosTriangleMidPointX,
                            TModelUnit BoundaryYLow,
                            TModelUnit BoundaryYHigh,
                            TModelUnit YCenterLine) const;
    void x_DrawDeletion(TModelUnit SeqPosTriangleMidPointX,
                            TModelUnit BoundaryYLow,
                            TModelUnit BoundaryYHigh) const;
    void x_DrawLabelWithXPinned(TModelUnit& base) const;
    void x_DrawLabelWithYPinned(TModelUnit base) const;
    void x_DrawInnerLabels(TModelUnit base, vector<TModelRange> *labels_range = nullptr, TSeqRange* interval = nullptr) const;

    /// determines whether the glyph should be drawn or not at all (currently children of unselected parents when m_LinkedFeat
    /// is set to "Expandable")
    bool x_isDrawn() const;

private:
    TModelUnit x_GetBarHeight() const;
    bool x_HasOuterOnly() const;
    bool x_ShowLabel() const;
    /// check to see if label needs to be drawn before feature bar.
    /// for side-placed labels, returns true on positive strand if left side of the glyph is before start of the whole sequence
    /// or on negative strand beyond the end (has no relation
    /// to the part of sequence being drawn) i.e. most of the time it is "false"
    bool x_LabelFirst() const;
    /// true if a label should be placed to the left of the feature bar (provided that it is ePos_Side)
    bool x_LabelOnLeft() const;

private:
    CConstRef<CFeatureParams>    m_Config;
    bool m_HideLabel{false};

	// object that we want to be associated with this glyph
	CConstRef<objects::CVcfVariant> m_Object;

	TIntervals m_Intervals;
};


///////////////////////////////////////////////////////////////////////////////
/// CVcfGlyph inline implementations

inline
CConstRef<CFeatureParams> CVcfGlyph::GetConfig() const
{
    return m_Config;
}

inline
void CVcfGlyph::SetConfig(CConstRef<CFeatureParams> conf)
{
    m_Config = conf;
}

inline
void CVcfGlyph::GetCustomColor(size_t idx, CRgbaColor& color) const
{
    const CCustomFeatureColor::TColorCode& colors =
        m_Config->m_CustomColors->GetColorCode();
    if (idx == 0) {
        color = m_Config->m_fgColor;
    } else if (idx < colors.size())  {
        color = colors[idx];
    }
}

END_NCBI_SCOPE

/* @} */

#endif  // GUI_WIDGETS_SEQ_GRAPHIC___VCF_GLYPH__HPP
