#ifndef GUI_WIDGETS_SEQ_GRAPHIC___VCF_HISTOGRAM__HPP
#define GUI_WIDGETS_SEQ_GRAPHIC___VCF_HISTOGRAM__HPP

/*
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladislav Evgeniev
 *
 * File Description:
 *
 */
#include <gui/widgets/seq_graphic/graph_cache.hpp>
#include <gui/widgets/seq_graphic/histogram_glyph.hpp>
#include <gui/widgets/loaders/columnar_vcf_variants_base.hpp>
#include <mutex>
#include <map>
#include <vector>

BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
///   CVcfHistogram
class NCBI_GUIWIDGETS_SEQGRAPHIC_EXPORT CVcfHistogram : public CObject
{
public:
    using THistogramVector = CVCFVariantsBase::TIndexVector;

    static string GetStorageName() {
        return "VcfHistogramCache";
    }

    ///< Creates a cache key for the given remote file and seq_id
    ///< the key value is CCompoundID
    ///< that parsed by constructor when the class is instantiated
    static string GetCacheKey(const string& annot_name, const size_t bin_width);

    CVcfHistogram(const string& cache_key);

    ///< Copy constructor is invoked by CGraphCache
    ///< which clones the data for aynchroneous saving operation
    CVcfHistogram(const CVcfHistogram& data);

    virtual ~CVcfHistogram() { }

    ///<  Init is invoked by CGraphCache
    ///<  when a new instance of CWigGraph is initialized
    void Init(TGraphCachePool& icache_pool);

    ///<  Save is invoked by CGraphCache
    ///<  when cached data needs to be permanently saved into NetCache
    void Save(TGraphCachePool& icache_pool);
   
    ///< Initializes the histogram map for a given range
    void GetHistogram(const TSeqRange& range, CHistogramGlyph::TMap& dmap) const;

	///< Sets the histogram data
    void SetData(const THistogramVector& data)
    {
        lock_guard<mutex> guard(m_DataMutex);
        m_Data = data;
    }

    static void InitHistogramGlyph(size_t bin_width, const TSeqRange& range, const THistogramVector& histogram, CHistogramGlyph::TMap& dmap)
    {
        size_t bin_start = range.GetFrom();
        while (bin_start < range.GetTo())
        {
            TSeqRange BinRange = TSeqRange(bin_start, bin_start + bin_width - 1);
            size_t bin_pos = (bin_start / bin_width) * bin_width;
            dmap.AddRange(BinRange, bin_pos < histogram.size() ? histogram[bin_pos] : 0);
            bin_start += bin_width;
        }
    }

    static void EraseCachePerAnnotation(const string& annot_name);

protected:
    string m_AnnotName;
    size_t m_BinWidth;
    
    mutable mutex m_DataMutex;      ///< data access mutex

    string m_DataKey;
    THistogramVector m_Data;

    static mutex m_KeysMapMutex;                    ///< Mutex for the cache keys
    static map<string, vector<string>>   m_KeysMap; ///< Map of cache keys per annotation
};

END_NCBI_SCOPE

#endif  // GUI_WIDGETS_SEQ_GRAPHIC___VCF_HISTOGRAM__HPP
