/*  $Id: columnar_vcf_load_manager.cpp 46175 2021-01-28 15:47:23Z evgeniev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladislav Evgeniev
 *
 */

#include <ncbi_pch.hpp>

#include "columnar_vcf_load_manager.hpp"

#include <gui/core/loading_app_job.hpp>
#include <gui/core/project_service.hpp>

#include <util/line_reader.hpp>
#include <serial/iterator.hpp>
#include <objects/seqloc/Seq_id.hpp>
#include <gui/objutils/label.hpp>
#include <gui/objutils/gencoll_svc.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/project_item_extra.hpp>

#include <gui/widgets/loaders/map_assembly_params.hpp>
#include <gui/widgets/loaders/assembly_sel_panel.hpp>
#include <gui/widgets/loaders/report_loader_errors.hpp>
#include <gui/widgets/loaders/columnar_vcf_variants.hpp>
#include <gui/widgets/loaders/vcf_assembly_sel_panel.hpp>
#include <gui/widgets/loaders/vcf_ref_sequences_panel.hpp>
#include <gui/widgets/loaders/vcf_column_sel_panel.hpp>
#include <gui/widgets/loaders/vcf_sample_sel_panel.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/file_extensions.hpp>
#include <gui/widgets/wx/compressed_file.hpp>
#include <gui/widgets/wx/async_call.hpp>

#include <gui/framework/workbench.hpp>

#include <objects/gbproj/LoaderDescriptor.hpp>
#include <objects/gbproj/AbstractProjectItem.hpp>
#include <objects/seq/Annotdesc.hpp>

#include <util/md5.hpp>
#include <wx/filename.h>
#include <chrono>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CVcfLoadingJob
class  CVcfLoadingJob : 
    public CDataLoadingAppJob,
    public CReportLoaderErrors
{
public:
    using TSeqIdsVector = vector<CConstRef<CSeq_id>>;

    CVcfLoadingJob();

    /// @name CDataLoadingAppJob implementation
    /// @{
    void GetHTMLResults(wxString& title, wxString& html);
    /// @}

    void SetVcfFiles(const vector<wxString>& files) { m_VcfFiles = files; }
    TSeqIdsVector& SetChromosomes() { return m_ChromIds; }
    TSeqIdsVector& SetSeqIdsToLoad() { return m_SeqIdsToLoad; }
    void SetReadAllChrms(bool value) { m_ReadAllChrms = value; }
    void SetGenCollAssembly(const CRef<CGC_Assembly>& assm ) { m_GenCollAssm = assm; }

    set<string>& SetInfoFields() { return m_LoadInfoFields; }
    void SetReadAllInfoFields(bool value) { m_ReadAllInfoFields = value; }

    map<unsigned, string>& SetSamples() { return m_Samples; }
    void SetReadAllSamples(bool value) { m_ReadAllSamples = value; }

protected:
    /// @name CDataLoadingAppJob implementation
    /// @{
    virtual void x_CreateProjectItems();
    /// @}

private:
    void x_OnVCFVariantListReady(CVCFVariantList &vars);
    void x_CreateColumnarVCFProjectItem(const CSeq_id &seq_id, CVCFVariantList &vars);
    void x_CreateSeqIdProjectItem(const CSeq_id &seq_id);
    vector<string> x_FindSynonyms(const CSeq_id& seq_id, const CGC_Assembly::TSequenceList& chromosomes);
    
    vector<wxString>    m_VcfFiles;
    TSeqIdsVector       m_ChromIds;
    TSeqIdsVector       m_SeqIdsToLoad;
    bool                m_ReadAllChrms{ true };
    CRef<CGC_Assembly>  m_GenCollAssm;
    unique_ptr<IIdMapper> m_IdMapper;
    wxFileName          m_CurrentFile;

    set<string>  m_LoadInfoFields;
    bool         m_ReadAllInfoFields;

    map<unsigned, string> m_Samples;
    bool                  m_ReadAllSamples;
};

///////////////////////////////////////////////////////////////////////////////
/// CColumnarVCFLoaderManager
CColumnarVCFLoaderManager::CColumnarVCFLoaderManager()
     : m_Descr("VCF (Variant Call Format) files", "")
     , m_SrvLocator(nullptr)
     , m_ParentWindow(nullptr)
     , m_State(eInvalid)
     , m_RefSequencesPanel(nullptr)
     , m_ColumnsPanel(nullptr)
     , m_SamplesPanel(nullptr)
     , m_AssemblyPanel(nullptr)
{
    m_Descr.SetLogEvent("loaders");
}

//
// IFileLoadPanelClient implementation
//

string CColumnarVCFLoaderManager::GetLabel() const
{
    return m_Descr.GetLabel();
}

wxString CColumnarVCFLoaderManager::GetFormatWildcard() const
{
    return CFileExtensions::GetDialogFilter(CFileExtensions::kVCF) + wxT("|") +
           CFileExtensions::GetDialogFilter(CFileExtensions::kAllFiles);
}

//
// IUILoadManager implementation
//

void CColumnarVCFLoaderManager::SetServiceLocator(IServiceLocator* srv_locator)
{
    m_SrvLocator = srv_locator;
}


void CColumnarVCFLoaderManager::SetParentWindow(wxWindow* parent)
{
    m_ParentWindow = parent;
}


const IUIObject& CColumnarVCFLoaderManager::GetDescriptor() const
{
    return m_Descr;
}


void CColumnarVCFLoaderManager::InitUI()
{
    m_State = eAssembly;
}


void CColumnarVCFLoaderManager::CleanUI()
{
    m_State = eInvalid;
    m_RefSequencesPanel = nullptr; // window is destroyed by the system
    m_AssemblyPanel     = nullptr; // window is destroyed by the system
    m_ColumnsPanel = nullptr;      // window is destroyed by the system
    m_SamplesPanel = nullptr;
}

wxPanel* CColumnarVCFLoaderManager::GetCurrentPanel()
{
    switch (m_State) {
    case eAssembly: {
        m_RefSeqIds.clear();
        if (m_AssemblyPanel == nullptr) {
            m_AssemblyPanel = new CVcfAssemblySelectionPanel(m_ParentWindow, &m_VCFreader, m_FileNames.front());
        }
        else {
            m_AssemblyPanel->SetNewFile(m_FileNames.front());
        }
        m_AssemblyPanel->TransferDataToWindow();
        return m_AssemblyPanel;
    }
    case eInfoSelection:
        if (m_ColumnsPanel == nullptr) {
            m_ColumnsPanel = new CVcfColumnSelectionPanel(m_ParentWindow);
        }
        m_ColumnsPanel->SetInfoFields(m_VCFreader.GetInfoFields());
        return m_ColumnsPanel;
    case eSampleSelection:
        if (m_SamplesPanel == nullptr) {
            m_SamplesPanel = new CVcfSampleSelectionPanel(m_ParentWindow);
        }
        m_SamplesPanel->SetSamples(m_VCFreader.GetSamples());
        return m_SamplesPanel;
    case eRefSequences:
    {
        if (m_RefSequencesPanel == nullptr) {
            m_RefSequencesPanel = new CVcfRefSequencesPanel(m_ParentWindow);
        }
        m_RefSequencesPanel->SetReferenceSeqIds(m_RefSeqIds);
        m_RefSequencesPanel->SetGenCollAssembly(m_GenCollAssm);
        if (m_TopLevelSeqCount != 0)
            m_RefSequencesPanel->SetTopLevelSeqCount(m_TopLevelSeqCount);
        return m_RefSequencesPanel;
    }
    default:
        return nullptr;
    }
    return nullptr;
}


bool CColumnarVCFLoaderManager::CanDo(EAction action)
{
    switch (m_State) {
    case eAssembly:
        return action == eNext;
    case eInfoSelection:
    case eSampleSelection:
    case eRefSequences:
        return action == eBack || action == eNext;
    case eCompleted:
        return action == eBack;
    default:
        _ASSERT(false);
        return false;
    }
}


bool CColumnarVCFLoaderManager::IsFinalState()
{
    return m_State == eCompleted;
}


bool CColumnarVCFLoaderManager::IsCompletedState()
{
    return m_State == eCompleted;
}


bool CColumnarVCFLoaderManager::DoTransition(EAction action)
{
    if (m_State == eAssembly && action == eNext) {
        if (m_AssemblyPanel->IsInputValid()) {
            const CMapAssemblyParams& map_params = m_AssemblyPanel->m_AssemblyPanel->GetData();
            if (map_params.GetUseMapping()) 
                m_AssmAcc = map_params.GetAssemblyAcc();
            else
                m_AssmAcc.clear();

            m_RefSeqIds.clear();
            m_GenCollAssm.Reset();
            m_TopLevelSeqCount = 0;
            wxBusyCursor wait;
            x_GetRefSeqs();
            
            m_State = eInfoSelection;
            return true;
        }
        return false;
    }
    else if (m_State == eInfoSelection) {
        if (action == eBack) {
            m_State = eAssembly;
            return true;
        }
        else if (action == eNext) {
            if (!m_VCFreader.GetSamples().empty()) {
                m_State = eSampleSelection;
            }
            else {
                m_State = eRefSequences;
            }
            return true;
        }
    } 
    else if (m_State == eSampleSelection) {
        if (action == eBack) {
            m_State = eInfoSelection;
            return true;
        }
        else if (action == eNext) {
            m_State = eRefSequences;
            return true;
        }
    }
    else if (m_State == eRefSequences) {
        if (action == eBack) {
            if (!m_VCFreader.GetSamples().empty()) {
                m_State = eSampleSelection;
            }
            else {
                m_State = eInfoSelection;
            }
            return true;
        }
        else if (action == eNext) {
            m_State = eCompleted;
            return true;
        }
    }
    else if (m_State == eCompleted) {
        if (action == eBack) {
            m_State = eRefSequences;
            return true;
        }
    }
    _ASSERT(false);
    return false;
}

void CColumnarVCFLoaderManager::x_GetRefSeqs()
{
    if (!m_AssmAcc.empty()) {
        // this is used only by Gbench, so using seqconfig as assembly service should be safe
        m_GenCollAssm = CGencollSvc::GetInstance()->GetGCAssembly(m_AssmAcc, true, "Gbench", nullptr, true);
    }

    m_RefSeqIds.clear();
    m_MappedSeqIds.clear();
    m_TopLevelSeqCount = 0;
    
    // retrieve all chromosomes
    if (!m_GenCollAssm) {
        LOG_POST(Error << "Chromosomes could not be retrieved as no assembly was found");
        return;
    }

    CGC_TypedSeqId::E_Choice seqid_type{ m_GenCollAssm->IsGenBank() ? CGC_TypedSeqId::e_Genbank : CGC_TypedSeqId::e_Refseq };

    CGC_Assembly::TSequenceList chromosomes;

    GUI_AsyncExec([this, &chromosomes](ICanceled& canceled)
    {
        m_GenCollAssm->GetMolecules(chromosomes, CGC_Assembly::eChromosome);

        auto start = chrono::steady_clock::now();

        CGC_Assembly::TSequenceList topLevel;
        m_GenCollAssm->GetMolecules(topLevel, CGC_Assembly::eTopLevel);
        m_TopLevelSeqCount = topLevel.size();

        auto diff = chrono::steady_clock::now() - start;
        LOG_POST(Info << "Got " << topLevel.size() << " top level sequences in " << chrono::duration_cast<chrono::milliseconds>(diff).count() << " ms ");

    }, wxT("Querying GenColl..."));
  
    set<string> added_ids;
    for (const auto& it : chromosomes) {
        CConstRef<CSeq_id> best_id = it->GetSynonymSeq_id(seqid_type, CGC_SeqIdAlias::e_Public);
        if (best_id) {
            const auto id_str = best_id->AsFastaString();
            if (added_ids.insert(id_str).second) {
                m_RefSeqIds.push_back(best_id);
            }
        }
    }
}

IAppTask* CColumnarVCFLoaderManager::GetTask()
{
    CIRef<CProjectService> srv = m_SrvLocator->GetServiceByType<CProjectService>();
    CIRef<CVcfLoadingJob> job{ new CVcfLoadingJob() };
    // create a wrapper task

    CSelectProjectOptions options;

    unique_ptr<CDataLoadingAppTask> task(new CDataLoadingAppTask(srv, options, *job));
    task->SetReportErrors(false); // we handle errors ourselves

    job->SetVcfFiles(m_FileNames);
    job->SetGenCollAssembly(m_GenCollAssm);
    job->SetChromosomes() = m_RefSeqIds;

    if (m_RefSequencesPanel) {
        m_RefSequencesPanel->GetSelectedReferenceSeqIds(job->SetSeqIdsToLoad());
        job->SetReadAllChrms(m_RefSequencesPanel->ImportAll());
    }
    if (m_ColumnsPanel) {
        m_ColumnsPanel->GetSelectedInfoFields(job->SetInfoFields());
        job->SetReadAllInfoFields(m_ColumnsPanel->AllAreChecked());
    }
    if (m_SamplesPanel) {
        m_SamplesPanel->GetSelectedSamples(job->SetSamples());
        job->SetReadAllSamples(m_SamplesPanel->AllAreChecked());
    }

    return task.release();
}

IExecuteUnit* CColumnarVCFLoaderManager::GetExecuteUnit()
{
    return nullptr;
}

wxString CColumnarVCFLoaderManager::GetFormatWildcard()
{
    return CFileExtensions::GetDialogFilter(CFileExtensions::kVCF) + wxT("|") +
           CFileExtensions::GetDialogFilter(CFileExtensions::kAllFiles);
}


bool CColumnarVCFLoaderManager::ValidateFilenames(const vector<wxString>&)
{
    // not implemented
    return true;
}


void CColumnarVCFLoaderManager::SetFilenames(const vector<wxString>& filenames)
{
    m_FileNames = filenames;
}

void CColumnarVCFLoaderManager::GetFilenames(vector<wxString>& filenames) const
{
    filenames = m_FileNames;
}

bool CColumnarVCFLoaderManager::IsInitialState()
{
    return m_State == eAssembly;
}


bool CColumnarVCFLoaderManager::RecognizeFormat(const wxString& filename)
{
    wxString ext;
    wxFileName::SplitPath(filename, 0, 0, &ext);
    return CFileExtensions::RecognizeExtension(CFileExtensions::kVCF, ext);
}

bool CColumnarVCFLoaderManager::RecognizeFormat(CFormatGuess::EFormat fmt)
{
    return fmt == CFormatGuess::eVcf;
}

string CColumnarVCFLoaderManager::GetExtensionIdentifier() const
{
    static string sid("columnar_vcf_format_load_manager");
    return sid;
}


string CColumnarVCFLoaderManager::GetExtensionLabel() const
{
    static string slabel("VCF (Variant Call Format) Format Load Manager");
    return slabel;
}

void CColumnarVCFLoaderManager::SetRegistryPath(const string& path)
{
    m_RegPath = path; // store for later use
}

static const char* kProjectParamsTag = "ProjectParams";

void CColumnarVCFLoaderManager::SaveSettings() const
{
    if (!m_RegPath.empty()) {
        if (m_AssemblyPanel)
            m_AssemblyPanel->SaveSettings();
    }
}

void CColumnarVCFLoaderManager::LoadSettings()
{
    if (!m_RegPath.empty()) {
        if (m_AssemblyPanel)
            m_AssemblyPanel->LoadSettings();
    }
}


//CVcfLoadingJob
CVcfLoadingJob::CVcfLoadingJob()
{
    m_Descr = "Loading VCF file";
}

void CVcfLoadingJob::GetHTMLResults(wxString& title, wxString& html)
{
    title = "VCF import errors";
    html = x_GetHTMLResults();
}

void CVcfLoadingJob::x_CreateProjectItems()
{
    _ASSERT(m_GenCollAssm); 
    m_IdMapper.reset(CGencollIdMapperAdapter::GetIdMapper(m_GenCollAssm));

    // Add the serialized VCF to the project
    for (const auto& fn : m_VcfFiles) {
        wxFileName wx_file(fn);
        m_CurrentFile = wx_file;

        CCompressedFile file(fn);
        CColumnarVCFReader reader;
        reader.LoadAllInfo(m_ReadAllInfoFields);
        reader.LoadSelectedInfoFields() = m_LoadInfoFields;
        reader.LoadAllSamples(m_ReadAllSamples);
        reader.LoadSelectedSamples() = m_Samples;
        CRef<CErrorContainer> err_cont{ new CErrorContainer(1000) };
        if (m_ReadAllChrms) {
            try {
                CColumnarVCFReader::TOnVCFVariantListReady callback = std::bind(&CVcfLoadingJob::x_OnVCFVariantListReady, this, std::placeholders::_1);
                CColumnarVCFReader::TReportProgress progress = bind(&CVcfLoadingJob::x_SetStatusText, this, placeholders::_1);
                reader.ReadData(file.GetIstream(false), x_GetICanceled(), err_cont, progress, callback);
                x_UpdateHTMLResults(fn, err_cont);
            }
            catch (const CException& e) {
                LOG_POST(Error << "Loading VCF file " << wx_file.GetFullName().utf8_str() << " failed: " << e.GetMsg());
                x_UpdateHTMLResults(fn, err_cont);
            }
            catch (const exception& e) {
                LOG_POST(Error << "Loading VCF file " << wx_file.GetFullName().utf8_str() << " failed: " << e.what());
                x_UpdateHTMLResults(fn, err_cont);
            }
        }
        else {
            // read only the requested ids
            CGC_Assembly::TSequenceList top_level;
            m_GenCollAssm->GetMolecules(top_level, CGC_Assembly::eTopLevel);

            vector<pair<CConstRef<CSeq_id>, vector<string>>> chrsyn_list;
            for (const auto& seqid_it : m_SeqIdsToLoad) {
                vector<string> synonyms = x_FindSynonyms(*seqid_it, top_level);
                chrsyn_list.emplace_back(seqid_it, synonyms);
            }

            try {
                CColumnarVCFReader::TOnVCFVariantListReady callback = std::bind(&CVcfLoadingJob::x_OnVCFVariantListReady, this, std::placeholders::_1);
                CColumnarVCFReader::TReportProgress progress = bind(&CVcfLoadingJob::x_SetStatusText, this, placeholders::_1);
                auto variants_list = reader.ReadVariantsForChrs(file.GetIstream(false), chrsyn_list, x_GetICanceled(), err_cont, progress, callback);
                x_UpdateHTMLResults(fn, err_cont);
            }
            catch (const CException& e) {
                LOG_POST(Error << "Loading chromosomes from VCF file " << wx_file.GetFullName().utf8_str() << " failed: " << e.GetMsg());
                x_UpdateHTMLResults(fn, err_cont);
            }
            catch (const exception& e) {
                LOG_POST(Error << "Loading chromosomes from VCF file " << wx_file.GetFullName().utf8_str() << " failed: " << e.what());
                x_UpdateHTMLResults(fn, err_cont);
            }
        }
    }
}

void CVcfLoadingJob::x_OnVCFVariantListReady(CVCFVariantList &vars)
{
    const auto &chr = vars.GetChrName();
    CRef<CSeq_id> mapped_id = CSeqUtils::MapStringId(chr, m_IdMapper.get());
    if (!mapped_id) {
        LOG_POST(Warning << "VCF ID " << chr << " cannot be mapped.");
        try {
            mapped_id.Reset(new CSeq_id(chr));
        }
        catch (const CException &) {
            LOG_POST(Warning << "VCF ID " << chr << " cannot be used as local ID.");
            return;
        }
    }
    x_SetStatusText("Creating items for " + chr + " (" + mapped_id->GetSeqIdString(true) + ")...");
    x_CreateColumnarVCFProjectItem(*mapped_id, vars);
    x_CreateSeqIdProjectItem(*mapped_id);
    x_SetTextResult("");
}

inline void CVcfLoadingJob::x_CreateColumnarVCFProjectItem(const CSeq_id &seq_id, CVCFVariantList &vars)
{   
    CRef<CProjectItem> item(new CProjectItem());
    CProjectItem::TItem& sub_item = item->SetItem();

    CAbstractProjectItem &other = sub_item.SetOther();

    other.SetType("ColumnarVCF: " + vars.s_GetCurrentVersion());
    CAbstractProjectItem::TData &data = other.SetData();

    // Time consuming part
    const auto &vars_data = vars.GetSerializedData();
    data.resize(vars_data.size());
    std::copy(vars_data.begin(), vars_data.end(), data.begin());

    string label(NStr::NumericToString(vars.GetNumberOfVariants(), NStr::fWithCommas));
    label += " variants on ";
    label += seq_id.GetSeqIdString(true);
    label += " (";
    label += m_CurrentFile.GetFullName().utf8_str();
    label += ')';
    item->SetLabel(label);
    item->AddTag(seq_id.AsFastaString());
    // Add the item to a list
    AddProjectItem(*item);

    CProjectItemExtra::SetStr(*item, "OriginalPath", string(m_CurrentFile.GetPath(true).utf8_str()));
    CProjectItemExtra::SetStr(*item, "OriginalFile", string(m_CurrentFile.GetFullName().utf8_str()));
    {
        string filename{ m_CurrentFile.GetFullPath().utf8_str() };
        CMD5 md5;
        md5.Update(filename.c_str(), filename.length());
        auto fasta_id = seq_id.AsFastaString();
        md5.Update(fasta_id.c_str(), fasta_id.length());
        CProjectItemExtra::SetStr(*item, "AnnotName", string("ColumnarVCF_") + md5.GetHexSum());
    }
    // Add the Seq-id to the project item
    CRef<CAnnotdesc> annot(new CAnnotdesc());
    annot->SetSrc().Assign(seq_id);
    item->SetDescr().push_back(annot);
}

inline void CVcfLoadingJob::x_CreateSeqIdProjectItem(const CSeq_id &seq_id)
{
    CRef<CProjectItem>   item(new CProjectItem());
    CRef<CSeq_id> seqId(new CSeq_id());
    seqId->Assign(seq_id);
    item->SetObject(*seqId);
    CLabel::SetLabelByData(*item, m_Scope.GetPointer());
    // add item to a list
    AddProjectItem(*item);
}

vector<string> CVcfLoadingJob::x_FindSynonyms(const CSeq_id& seq_id, const CGC_Assembly::TSequenceList& chromosomes)
{
    CGC_TypedSeqId::E_Choice seqid_type{ (m_GenCollAssm && m_GenCollAssm->IsGenBank()) ? CGC_TypedSeqId::e_Genbank : CGC_TypedSeqId::e_Refseq };
    vector<string> syn;
    for (const auto& chr_it : chromosomes) {
        CConstRef<CSeq_id> best_id = chr_it->GetSynonymSeq_id(seqid_type, CGC_SeqIdAlias::e_Public);
        if (best_id && best_id->Equals(seq_id)) {
            if (!chr_it->IsSetSeq_id_synonyms()) {
                break;
            }
   
            ITERATE(CGC_Sequence::TSeq_id_synonyms, syn_iter, chr_it->GetSeq_id_synonyms()) {
                CTypeConstIterator<CSeq_id> id_iter(**syn_iter);
                for (; id_iter; ++id_iter) {
                    if (id_iter->IsGi())
                        continue;

                    string label;
                    id_iter->GetLabel(&label, CSeq_id::eContent);
                    syn.push_back(label);
                }
            }
            break;
        }
    }
    sort(syn.begin(), syn.end());
    syn.erase(unique(syn.begin(), syn.end()), syn.end());
    return syn;
}

END_NCBI_SCOPE
