/*  $Id: table_data_vcf.cpp 45881 2021-01-05 03:10:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Dmitry Rudnev
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/nonasn_track_data.hpp>
#include <gui/utils/app_job.hpp>
#include <gui/utils/app_job_impl.hpp>
#include <gui/packages/pkg_sequence/table_data_vcf.hpp>

#include <objmgr/util/create_defline.hpp>
#include <objmgr/util/sequence.hpp>

#include <gui/widgets/loaders/columnar_vcf_variants.hpp>
#include <gui/widgets/loaders/columnar_vcf_iterator.hpp>

#include <objects/seq/Seq_annot.hpp>
#include <objects/seq/Seq_literal.hpp>
#include <objects/seq/Seq_data.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/general/User_object.hpp>
#include <objects/general/User_field.hpp>
#include <objects/seqfeat/Variation_inst.hpp>
#include <objects/seqfeat/Delta_item.hpp>

#include <common/ncbi_source_ver.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


class CTableDataVcfJob : public CJobCancelable
{
public:
    CTableDataVcfJob(CBioseq_Handle handle, 
		const TSeqRange& range,
		int max_vcf,
		const INonAsnTrackData::TDataBlob& vcf_blob);
    virtual ~CTableDataVcfJob() {}

    /// @name IAppJob implementation
    /// @{
    virtual EJobState                   Run();
    virtual CConstIRef<IAppJobProgress> GetProgress();
    virtual CRef<CObject>               GetResult();
    virtual CConstIRef<IAppJobError>    GetError();
    virtual string                      GetDescr() const;
    /// @}

private:
	objects::CBioseq_Handle m_Handle;           ///< target sequence
	TSeqRange               m_Range;            ///< target range
	CRef<CObject> m_listVcfs;

	/// -1: theoretically unlimited, practically there is one (kMaxObjNum)
	/// >0: any limit set from a user
	int     m_VcfLimit;
	size_t  m_Total;
    size_t  m_Completed;
    CRef<CAppJobError>      m_Error;
	const INonAsnTrackData::TDataBlob& m_VcfBlob;
};


CTableDataVcfJob::CTableDataVcfJob(CBioseq_Handle handle, 
	const TSeqRange& range,
	int max_vcf,
	const INonAsnTrackData::TDataBlob& vcf_blob)
	:
	m_Handle(handle),
	m_Range(range),
	m_VcfLimit(max_vcf),
	m_Total(100),
    m_Completed(0),
	m_VcfBlob(vcf_blob)
{
    m_listVcfs.Reset(new CObjectFor<CTableDataVcf::TVcfDataList>());
}

CConstIRef<IAppJobProgress> CTableDataVcfJob::GetProgress()
{
    CRef<CAppJobProgress> p( new CAppJobProgress() );
    p->SetNormDone( float(m_Completed) / float(m_Total) );
    return CConstIRef<IAppJobProgress>( p.GetPointer() );
}

CRef<CObject> CTableDataVcfJob::GetResult()
{
    return m_listVcfs;
}


CConstIRef<IAppJobError> CTableDataVcfJob::GetError()
{
    return CConstIRef<IAppJobError>( m_Error.GetPointer() );
}


string CTableDataVcfJob::GetDescr() const
{
    return "VCF table data loader job";
}


IAppJob::EJobState CTableDataVcfJob::Run()
{
	try {
		CTableDataVcf::TVcfDataList& vcf_list(dynamic_cast<CObjectFor<CTableDataVcf::TVcfDataList>*>(m_listVcfs.GetPointer())->GetData());
		const vector<string>& all_col_names(CVCFVariantsBase::s_GetAllColNames());
		set<string> data_cols(all_col_names.begin(), all_col_names.end());
		CVCFSlicedVariants variants(m_VcfBlob, &m_Range, data_cols);

		const CVariantDescriptors& descriptors(variants.GetDescriptors());

		vcf_list.m_InfoFields = descriptors.GetInfoFieldNames();
		vcf_list.m_SampleNames = descriptors.GetSampleNames();

		int size = (int)variants.Count(m_Range);
		if (size > 0) {
			if (m_VcfLimit == -1) {
				// unlimited, but still set a upper limit to guard against problem.
				m_VcfLimit = CTableDataVcf::c_MaxRows;
			}
			m_Total = m_VcfLimit;

			CVCFVariant_CI variant_CI(variants, m_Range);
			CRef<CSeq_id> seq_id(new CSeq_id);
			seq_id->Assign(*m_Handle.GetSeqId());
			int nCurrent(0);

			while(variant_CI.IsValid() && nCurrent++ < m_VcfLimit && !IsCanceled()) {
				CVcfVariant* vcf_variant = new CVcfVariant(*seq_id, variant_CI.GetVariantID(), variant_CI.GetPosition(),
					variant_CI.GetRef(), variant_CI.GetAlt());
				if (vcf_variant) {
					if(variant_CI.IsSetInfo()) {
						CVcfVariant::TColumns& info_columns(vcf_variant->SetInfoColumns());
						for(auto info_column_name: vcf_list.m_InfoFields) {
							if(variant_CI.IsSetInfoField(info_column_name)) {
								info_columns[info_column_name] = variant_CI.GetInfoField(info_column_name);
							}
						}
					}
					CVcfVariant::TColumns& sample_columns(vcf_variant->SetSampleColumns());
					for(auto sample_column_name: vcf_list.m_SampleNames) {
						if(variant_CI.IsSetSample(sample_column_name)) {
							sample_columns[sample_column_name] = variant_CI.GetSample(sample_column_name);
						}
					}
					vcf_list.m_VcfVariants.push_back(CRef<CVcfVariant>(vcf_variant));
				}
				++variant_CI;
			}
		}
		m_Completed = size;
	} catch (CException& ex) {
		m_Error.Reset(new CAppJobError(ex.GetMsg()));
		return eFailed;
	} catch (std::exception& ex) {
		m_Error.Reset(new CAppJobError(ex.what()));
		return eFailed;
	}
	return IsCanceled() ? eCanceled : eCompleted;
}


BEGIN_EVENT_MAP( CTableDataVcf, CEventHandler )
    ON_EVENT(CAppJobNotification, CAppJobNotification::eStateChanged, &CTableDataVcf::OnJobNotification)
END_EVENT_MAP()

// maximal amount of VCF variants we should realistically see in a VCF table
const size_t CTableDataVcf::c_MaxRows(100000);

CTableDataVcf::CTableDataVcf(CScope& scope,
	CBioseq_Handle handle, 
	const TSeqRange& range,
	const INonAsnTrackData::TDataBlob& vcf_blob)
:
	m_Scope(&scope),
	m_Handle(handle),
	m_Range(range),
	m_VcfBlob(vcf_blob),
	m_DataLoadStatus(eNotReady)
{
}

CTableDataVcf::~CTableDataVcf()
{
    if(m_DataLoadStatus == eNotReady) {
        CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
        disp.DeleteJob(m_hJob);
    }
}



void CTableDataVcf::LoadData()
{
    CRef<CTableDataVcfJob> job(new CTableDataVcfJob(m_Handle, m_Range, c_MaxRows, m_VcfBlob));
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    try {
        m_hJob = disp.StartJob(*job, "ObjManagerEngine", *this);
    } catch( CAppJobException& e ){
        m_DataLoadStatus = eFailed;
        LOG_POST( Error
            << "CTableDataVcf failed to start job: "
            << e.GetMsg()
        );
        LOG_POST(e.ReportAll());
    }
}


void CTableDataVcf::OnJobNotification(ncbi::CEvent *evt)
{
    CAppJobNotification* notif = dynamic_cast<CAppJobNotification*>( evt );

    _ASSERT(notif);

    int                 job_id  = notif->GetJobID();
    IAppJob::EJobState  state   = notif->GetState();

    if (job_id != m_hJob) return;

    switch (state) {
        case IAppJob::eCompleted:
        {
            CAppJobDispatcher& disp(CAppJobDispatcher::GetInstance());
			m_listVcfs.Reset(dynamic_cast<CObjectFor<CTableDataVcf::TVcfDataList>*>(disp.GetJobResult(job_id).GetPointer()));
			m_ColNames.clear();
			// order must correspond with CTableDataVcf::EColContents
			m_ColNames.push_back("POS");
			m_ColNames.push_back("ID");
			m_ColNames.push_back("Type");
			m_ColNames.push_back("Alleles");
			m_FirstInfoCol = m_ColNames.size();
			m_ColNames.insert(m_ColNames.end(), m_listVcfs->GetData().m_InfoFields.begin(), m_listVcfs->GetData().m_InfoFields.end());
			m_FirstSampleCol = m_ColNames.size();
			m_ColNames.insert(m_ColNames.end(), m_listVcfs->GetData().m_SampleNames.begin(), m_listVcfs->GetData().m_SampleNames.end());
			m_DataLoadStatus = eReady;

            // send a message to the view that data has changed (i.e. ready)
            CVcfTableEvent evt(*this);
            Send(&evt, ePool_Parent);
            break;
        }
        case IAppJob::eCanceled:
        case IAppJob::eFailed:
            m_DataLoadStatus = eFailed;
            break;
        default:
            break;
    }
}


ITableData::ColumnType CTableDataVcf::GetColumnType(size_t col) const
{
    if(m_DataLoadStatus != eReady) {
        return ITableData::kString;
    }
    switch(col){
        case ePosition:
            return ITableData::kInt;
        default:
            return ITableData::kString;
    }
    return ITableData::kNone;
}


bool CTableDataVcf::IsHtmlColumn(size_t col) const
{
    if(m_DataLoadStatus != eReady) {
        return false;
    }
    return false;
}

string CTableDataVcf::GetColumnLabel(size_t col) const
{
    if(m_DataLoadStatus != eReady && col == 0) {
        return "Status";
    }
    if (col < m_ColNames.size()) {
        return m_ColNames[col];
	}
    return kEmptyStr;
}


size_t CTableDataVcf::GetRowsCount() const
{
    return m_DataLoadStatus == eReady ? m_listVcfs->GetData().m_VcfVariants.size() : 1;
}


size_t CTableDataVcf::GetColsCount() const
{
	return m_DataLoadStatus == eReady ? m_ColNames.size() : 1;
}

void CTableDataVcf::GetStringValue(size_t row, size_t col, string& value) const
{
	value.clear();
    if(m_DataLoadStatus == eNotReady) {
        value = (col == 0 ? "Loading..." : "");
        return;
    }
    if(m_DataLoadStatus == eFailed) {
        value = (col == 0 ? "Load failed!" : "");
        return;
    }
	//!! remove this if we implement dynamic loading
    if(row+1 == c_MaxRows) {
        value = "Warning: Only the first " + NStr::NumericToString(c_MaxRows) + " rows can be shown";
        return;
    }
	string sColLabel(GetColumnLabel(col));
	if(!sColLabel.empty() && row < GetRowsCount()) {
		CConstRef<CVcfVariant> vcf_variant(x_GetRowData(row));
		switch(col) {
		case ePosition:
			value = NStr::NumericToString(vcf_variant->GetPos()+1);
			break;
		case eId:
			value = vcf_variant->GetID();
			break;
		case eVariationType:
			value = vcf_variant->GetVariationTypeAsString();
			break;
		case eAlleles:
			value = vcf_variant->GetAllAlleles_display();
			break;
		default:
			if(m_FirstInfoCol <= col && col < m_FirstSampleCol) {
				if(vcf_variant->GetInfoColumns().find(sColLabel) != vcf_variant->GetInfoColumns().end()) {
					value = vcf_variant->GetInfoColumns().at(sColLabel);
					// this provides for Type=Flag INFO values to be shown
					if(value.empty()) {
						value = sColLabel;
					}
				}
			}
			if(m_FirstSampleCol <= col && col < m_ColNames.size()) {
				if(vcf_variant->GetSampleColumns().find(sColLabel) != vcf_variant->GetSampleColumns().end()) {
					value = vcf_variant->GetSampleColumns().at(sColLabel);
				}
			}
		}
	}
//	LOG_POST(Info << "CTableDataVcf::GetStringValue() " << NStr::NumericToString(row) << ":" << NStr::NumericToString(col) << " = " << value);
}


string CTableDataVcf::GetHtmlValue(size_t row, size_t col) const
{
//	LOG_POST(Info << "CTableDataVcf::GetHtmlValue() " << NStr::NumericToString(row) + ":" + NStr::NumericToString(col));
	return "";
}




long CTableDataVcf::GetIntValue(size_t row, size_t col) const
{
	long value(0);
	if(row < GetRowsCount()) {
		CConstRef<CVcfVariant> vcf_variant(x_GetRowData(row));
		if(col == ePosition) {
			value = vcf_variant->GetPos()+1;
		}
	}
//	LOG_POST(Info << "CTableDataVcf::GetIntValue() " << NStr::NumericToString(row) << ":" << NStr::NumericToString(col) << " = " << value);
	return value;
}


double CTableDataVcf::GetRealValue(size_t row, size_t col) const
{
    double value = 0.0;
//	LOG_POST(Info << "CTableDataVcf::GetRealValue() " << NStr::NumericToString(row) + ":" + NStr::NumericToString(col));
	return value;
}


SConstScopedObject CTableDataVcf::GetObjectValue(size_t row, size_t col) const
{
//	LOG_POST(Info << "CTableDataVcf::GetObjectValue() " << NStr::NumericToString(row) + ":" + NStr::NumericToString(col));
	return GetRowObject(row);
}


SConstScopedObject CTableDataVcf::GetRowObject(size_t row) const
{
    SConstScopedObject value;
    if(m_DataLoadStatus == eReady && row < GetRowsCount()) {
        value.scope = m_Scope;
        value.object = x_GetRowData(row);
    }
    return value;
}


CConstRef<CVcfVariant> CTableDataVcf::x_GetRowData(size_t row) const
{
    return row < m_listVcfs->GetData().m_VcfVariants.size() ? m_listVcfs->GetData().m_VcfVariants[row] : CConstRef<CVcfVariant>(NULL);
}


END_NCBI_SCOPE

