/*  $Id: vcf_search_tool.cpp 45888 2021-01-05 23:24:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/data_mining_service.hpp>
#include "vcf_search_tool.hpp"

#include <corelib/ncbitime.hpp>

#include <gui/widgets/wx/message_box.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/seq_graphic/vcf_ds.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/widgets/wx/ui_tool_registry.hpp>

#include <gui/utils/app_job_dispatcher.hpp>

#include <objmgr/util/sequence.hpp>
#include <objmgr/seq_map_ci.hpp>

#include <gui/utils/extension_impl.hpp>

#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/choice.h>
#include <wx/srchctrl.h>


#define ID_COMBOBOX 10003
#define ID_TEXT     11414


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// IDMSearchTool
CVcfSearchTool::CVcfSearchTool()
{
}


string CVcfSearchTool::GetName() const
{
    static string name("VCF Search");
    return name;
}


IUITool* CVcfSearchTool::Clone() const
{
    return new CVcfSearchTool();
}


string CVcfSearchTool::GetDescription() const
{
    return "";
}


CIRef<IDMSearchForm> CVcfSearchTool::CreateSearchForm()
{
    CIRef<IDMSearchForm> form(new CVcfSearchForm(*this));
    return form;
}


bool CVcfSearchTool::IsCompatible(IDataMiningContext* context)
{
	bool isCompatible(false);
    ISeqLocSearchContext* sl_ctx = dynamic_cast<ISeqLocSearchContext*>(context);
    if(sl_ctx) {
		CVcfTrackData* vcf_data(dynamic_cast<CVcfTrackData*>(sl_ctx->GetNonAsnData()));
		isCompatible = vcf_data != NULL;
	}
//	LOG_POST(Info << "CVcfSearchTool::IsCompatible(): " << isCompatible << ":" << " for " << context << " " << context->GetDMContextName());	
	
	return isCompatible;
}


CRef<CSearchJobBase> CVcfSearchTool::x_CreateJob(IDMSearchQuery& query)
{
    CRef<CSearchJobBase> job;
    CVcfSearchQuery* f_query = dynamic_cast<CVcfSearchQuery*>(&query);
    if(f_query) {
        job.Reset(new CVcfSearchJob(*f_query));
    }
    return job;
}



string  CVcfSearchTool::GetExtensionIdentifier() const
{
    return "search_tool::vcf_search_tool";
}

string  CVcfSearchTool::GetExtensionLabel() const
{
    return "Datamining Tool - VCF search";
}


///////////////////////////////////////////////////////////////////////////////
/// CVcfSearchQuery
CVcfSearchQuery::CVcfSearchQuery(vector<ISeqLocSearchContext*> VcfContexts, const string& ids)
	: m_VcfContexts(VcfContexts),
	  m_Ids(ids)
{
}


///////////////////////////////////////////////////////////////////////////////
/// CVcfSearchForm

CVcfSearchForm::CVcfSearchForm(CVcfSearchTool& tool)
:   m_Tool(&tool)
{
}


void CVcfSearchForm::Create()
{
    x_CreateWidgets();
}


void CVcfSearchForm::Init()
{
    CSearchFormBase::Init();
    


  //  x_UpdateData(eDDX_Set);
}


void CVcfSearchForm::UpdateContexts()
{
    CSearchFormBase::UpdateContexts();
}


void CVcfSearchForm::x_CreateWidgets()
{
//    CSearchFormBase::x_CreateWidgets(2, 3, true);
}


IDMSearchTool* CVcfSearchForm::x_GetTool()
{
    return m_Tool.GetPointer();
    
}


wxSizer *  CVcfSearchForm::GetWidget(wxWindow * parent)
{
    if (!m_Sizer) {     
        wxFlexGridSizer * sz = new wxFlexGridSizer(1, 4, 0, 0);     
        sz->AddGrowableCol(3);
        m_Sizer = sz;
        
        m_Sizer->Add(new wxStaticText( parent, wxID_STATIC, wxT("Search Context:"), wxDefaultPosition, wxDefaultSize, 0 ),
            0, wxALIGN_CENTER_VERTICAL|wxALL, 5);   

        m_DbCombo = new wxChoice(parent, ID_COMBOBOX,
                                   wxDefaultPosition, wxDefaultSize,
                                   0, (const wxString*)NULL);
        
        m_Sizer->Add(m_DbCombo,1, wxGROW|wxALIGN_CENTER_VERTICAL|wxALL, 5);     
        
        m_Sizer->Add(new wxStaticText( parent, wxID_STATIC, wxT("Search Expression:"), wxDefaultPosition, wxDefaultSize, 0 ),
            0, wxALIGN_CENTER_VERTICAL|wxALL, 5);   
    
        m_Text = new CSearchControl(parent, ID_TEXT, wxT(""), wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER );
        m_Sizer->Add(m_Text,1, wxGROW|wxALIGN_CENTER_VERTICAL|wxALL, 5);
        
    }
    return m_Sizer;
}

void CVcfSearchForm::Update()
{
    UpdateContextCombo(m_DbCombo);
}

CIRef<IDMSearchQuery> CVcfSearchForm::ConstructQuery()
{
    CIRef<IDMSearchQuery> ref;

	vector<ISeqLocSearchContext*>  VcfContexts;
    for( size_t i = 0;  i < m_Contexts.size();  i++ )   {
        if (m_DbCombo->GetSelection() != 0  &&
            m_DbCombo->GetSelection() != int(i + 1)) {
             continue;
        }
        ISeqLocSearchContext* seq_ctx =
            dynamic_cast<ISeqLocSearchContext*>(m_Contexts[i]);
		VcfContexts.push_back(seq_ctx);
      }

    ref = new CVcfSearchQuery(VcfContexts, ToStdString(m_Text->GetValue()));
    return ref;
}


///////////////////////////////////////////////////////////////////////////////
/// CVcfSearchJob

CVcfSearchJob::CVcfSearchJob(CVcfSearchQuery& query)
:   m_Query(&query)
{
    m_Descr = "Search for VCF variant with ID \"" + m_Query->GetIds() + "\"";
}


bool CVcfSearchJob::x_ValidateParams()
{
    m_Error.Reset();

	if (m_Query->GetIds().empty()) {
        m_Error = new CAppJobError("No sequence ids are provided.");
    }
    return m_Error ? false : true;
}


void CVcfSearchJob::x_SetupColumns(CObjectList& obj_list)
{
	CSearchJobBase::x_SetupColumns(obj_list);
}


IAppJob::EJobState CVcfSearchJob::x_DoSearch()
{
    CStopWatch sw;
    sw.Start();
    {
        CMutexGuard Guard(m_Mutex);
        m_ProgressStr = "Loading data." ;
    }
	const vector<ISeqLocSearchContext*>& vcf_contexts(m_Query->GetVcfContexts());
    string ids = m_Query->GetIds();
	CObjectList * obj_list = m_TempResult->GetObjectList();

    for(auto vcf_context: vcf_contexts)    {
        if(IsCanceled()) {
            break;
        }
		INonAsnTrackData* NonAsnData(vcf_context->GetNonAsnData());
		if(NonAsnData) {
			list<CRef<CObject>> results;
			CRef<CScope> scope(vcf_context->GetSearchScope());
			NonAsnData->DoSearch(vcf_context->GetSearchLoc()->GetId(), ids, results);
			for(auto result: results) {
				obj_list->AddRow(result.GetPointer(), scope.GetPointer());
			}
        }
    }
    return IsCanceled() ? eCanceled : eCompleted;
}


END_NCBI_SCOPE
