/*  $Id: vcf_table_view.cpp 46219 2021-02-05 23:12:28Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Dmitry Rudnev
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_sequence/vcf_table_view.hpp>

#include <gui/core/project_view_impl.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/document.hpp>

#include <gui/framework/workbench.hpp>
#include <gui/framework/status_bar_service.hpp>

#include <gui/utils/extension_impl.hpp>
#include <gui/packages/pkg_sequence/table_data_vcf.hpp>
#include <gui/objutils/obj_event.hpp>
#include <gui/objutils/nonasn_track_data.hpp>
#include <gui/objutils/table_selection.hpp>
#include <gui/widgets/data/query_panel_event.hpp>
#include <gui/widgets/grid_widget/grid_event.hpp>
#include <gui/widgets/grid_widget/grid.hpp>
#include <gui/widgets/grid_widget/grid_widget.hpp>

#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/message_box.hpp>

#include <gui/utils/app_popup.hpp>

#include <objects/general/Object_id.hpp>
#include <objmgr/util/seq_loc_util.hpp>
#include <objmgr/util/sequence.hpp>
#include <wx/msgdlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


///////////////////////////////////////////////////////////////////////////////
/// CVcfTableView

CProjectViewTypeDescriptor s_VcfViewTypeDescr(
    "VCF Table View",       // type name
    "vcf_table_view",       // icon alias TODO
    "Show tabular information about VCFs",
    "The VCF Table View provides a tabular break-down of a loaded VCF file",
    "VCF_TABLE_VIEW",       // help ID
    "Sequence",             // category
    false,                   // not a singleton
    "Seq-loc",
    eOneObjectAccepted
);

// handler for custom Vcf commands injected into the grid menu
class CVcfGridEvtExt : public CObject,
                       public IGridEvtExt
{
public:
    enum EEventType {
//        eCmdLinkRefVcf = 8800,
    };
protected:
    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(CVcfGridEvtExt, wxEvtHandler)
//    EVT_MENU(CVcfGridEvtExt::eCmdLinkRefVcf, CVcfGridEvtExt::OnEventLinkout)
END_EVENT_TABLE()


class CVcfTableSelection : public CObject, public ITableSelection
{
public:
	virtual void GetRows(const ITableData& table, const CSelectionEvent& evt, vector<size_t>& rows) const;
	virtual void GetSelection(const ITableData& table, const vector<size_t>& rows, CSelectionEvent& evt) const;
};


void CVcfTableSelection::GetRows(const ITableData& table, const CSelectionEvent& evt, vector<size_t>& rows) const
{
	size_t nRows(table.GetRowsCount());
	const CSelectionEvent::TVcfVariants& sel_variants(evt.GetVcfVariants());
	size_t n_sel_variants(sel_variants.size());
	for(size_t i=0; i<nRows; i++) {
		SConstScopedObject row_obj(table.GetRowObject(i));
		const CVcfVariant* vcf_variant(dynamic_cast<const CVcfVariant*>(row_obj.object.GetPointer()));
		if(vcf_variant) {
			for(size_t j=0; j<n_sel_variants; j++) {
				if(vcf_variant->WeakMatch(*sel_variants[j])) {
					rows.push_back(i);
					break;
				}
			}
		}
	}

}

void CVcfTableSelection::GetSelection(const ITableData& table, const vector<size_t>& rows, CSelectionEvent& evt) const
{
	size_t n_sel_rows(rows.size());
	for(size_t i=0; i<n_sel_rows; i++) {
		SConstScopedObject row_obj(table.GetRowObject(i));
		const CVcfVariant* vcf_variant(dynamic_cast<const CVcfVariant*>(row_obj.object.GetPointer()));
		if(vcf_variant) {
			evt.AddObjectSelection(*vcf_variant);
		}
	}
}


class CVcfTableView : public CProjectView, public IRegSettings
{
public:
    CVcfTableView() : m_Window() {}

    virtual wxWindow* GetWindow() { return m_Window; }
    virtual wxEvtHandler* GetCommandHandler() { return m_Window->GetEventHandler(); }
    /// @}

    /// @name IView implementation
    /// @(
    virtual const CViewTypeDescriptor&  GetTypeDescriptor() const
    {
        return s_VcfViewTypeDescr;
    }
    virtual void    CreateViewWindow(wxWindow* parent);
    virtual void    DestroyViewWindow();
    /// @}

    /// @name IProjectView implementation
    /// @{
    virtual bool InitView(TConstScopedObjects& objects, const objects::CUser_object* params);
    virtual void OnProjectChanged(CProjectViewEvent& evt);
    /// @}

    /// @name ISelectionClient Implementation
    /// @{
    virtual void GetSelection(CSelectionEvent& evt) const;
    virtual void GetSelection(TConstScopedObjects& objs) const;
    virtual void GetMainObject(TConstScopedObjects& objs) const;
    /// @}

    /// @addtogroup CProjectViewBase overridables
    /// @{
    virtual void GetCompatibleToolBars(vector<string>& names);
    virtual const CObject* x_GetOrigObject() const;
    /// @}

	virtual void    OnWidgetSelChanged();
	/// populate a visible range event for broadcasting
	virtual void GetVisibleRanges(CVisibleRange& vrange) const;

    /// @name IRegSettings interface implementation
    /// @{
    virtual void SetRegistryPath(const string& reg_path);
    virtual void LoadSettings();
    virtual void SaveSettings() const;
    /// @}

protected:
    void x_OnDataChanged(CEvent* evt);
    void x_OnUrlHover(CEvent* evt);
    void x_QueryStatusChange(CEvent* evt);
	virtual void    x_OnSetSelection(CSelectionEvent& evt);

    DECLARE_EVENT_MAP();

    TConstScopedObjects m_OrigObjects;
    wxWindow* m_Window;
	CIRef<INonAsnTrackData>         m_NonAsnTrackData;
};

BEGIN_EVENT_MAP(CVcfTableView, CProjectView)
  ON_EVENT(CVcfTableEvent, CVcfTableEvent::eDataChanged,
           &CVcfTableView::x_OnDataChanged)
  ON_EVENT(CGridWidgetEvent, CGridWidgetEvent::eUrlHover,
           &CVcfTableView::x_OnUrlHover)
  ON_EVENT(CQueryPanelEvent, CQueryPanelEvent::eStatusChange,
           &CVcfTableView::x_QueryStatusChange)
END_EVENT_MAP()


void CVcfTableView::CreateViewWindow(wxWindow* parent)
{
    _ASSERT(!m_Window);

    CGridWidget* widget = new CGridWidget(parent, wxID_ANY, wxDefaultPosition, wxSize(0, 0), wxBORDER_NONE);
	widget->SetTableSelection(new CVcfTableSelection());
    m_Window = widget;

    widget->SetEventExt(new CVcfGridEvtExt());

    widget->AddListener(this, ePool_Parent);
}

void CVcfTableView::DestroyViewWindow()
{
    if( m_Window ){
        m_Window->Destroy();
        m_Window = 0;
    }
}

bool CVcfTableView::InitView(TConstScopedObjects& objects, const objects::CUser_object*)
{
    m_OrigObjects = objects;
	CScope* scope = NULL;
	const CObject* object = NULL;

	if(objects.size() == 1) {
		object = objects[0].object.GetPointer();
		scope = objects[0].scope.GetPointer();

		const CSeq_loc*  loc = dynamic_cast<const CSeq_loc*> (object);
		const CSeq_id*  sid = dynamic_cast<const CSeq_id*> (object);
		const CBioseq*  seq = dynamic_cast<const CBioseq*> (object);
		const CSeq_entry* seq_entry = dynamic_cast<const CSeq_entry*> (object);
		if (seq_entry && seq_entry->IsSeq())
			seq = &seq_entry->GetSeq();

		if (loc) {
			sid = &sequence::GetId(*loc, scope);
		} else if (seq) {
			sid  = seq->GetFirstId();
		}
		if (sid) {
			static const wxString strInitializing = wxT("Initializing...");
			SConstScopedObject s_obj(sid, scope);
			CBioseq_Handle handle;
			try {
				handle = GUI_AsyncExec(
					[&s_obj](ICanceled&)
					{
						const CSeq_id* id = dynamic_cast<const CSeq_id*>(s_obj.object.GetPointer());
						CScope* scope = s_obj.scope.GetPointer();

						CBioseq_Handle handle;
						if (id  &&  scope) {
							handle = scope->GetBioseqHandle(*id);
							// resolving the seq-id in other form to avoid
							// potential future network connection
							CSeq_id_Handle idh = handle.GetSeq_id_Handle();
							idh = sequence::GetId(*idh.GetSeqId(), *scope, sequence::eGetId_Best);
						}
						return handle;
					},
					strInitializing);
			} catch (CException& e) {
				string str("Failed to retrieve sequence: ");
				str += e.GetMsg();
				wxMessageBox(ToWxString(str));
				return false;
			}
			if ( !handle ) {
				string str;
				sid->GetLabel(&str);
				ReportIDError(str, sid->IsLocal(), "VCF Table View");
				return false;
			}
			CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();
			_ASSERT(srv);

			CRef<CGBWorkspace> ws = srv->GetGBWorkspace();
			if (!ws) return false;

			CGBDocument* doc = dynamic_cast<CGBDocument*>(ws->GetProjectFromScope(*objects[0].scope));
			_ASSERT(doc);

			const objects::CProjectFolder& prj_folder(doc->GetData());
		
			if(!prj_folder.IsSetItems())
				return false;

			vector< CIRef<INonAsnTrackDataFactory> >   data_factories;
			GetExtensionAsInterface(EXT_POINT__NONASN_TRACK_DATA_FACTORY, data_factories);

			if (data_factories.empty()) 
				return false;

//			LOG_POST(Info << "CVcfTableView::InitView() " << " for: " << sid->AsFastaString() << ", m_NonAsnTrackData: " <<
//					m_NonAsnTrackData.GetPointer());	
			for (const auto& prj_item : prj_folder.GetItems()) {
//				LOG_POST(Info << "\tChecking item: " << prj_item->GetLabel());	
				if (!prj_item->GetItem().IsOther())
					continue;

				for (auto &factory : data_factories) {
					CIRef<INonAsnTrackData> track_data(factory->CreateTrackData(*sid, *prj_item));
					if (track_data.IsNull()) {
						continue;
					}
					m_NonAsnTrackData.Reset(track_data);
				}
			}
//			LOG_POST(Info << "m_NonAsnTrackData reset to: " << m_NonAsnTrackData.GetPointer());	

			if(m_NonAsnTrackData.IsNull()) {
				return false;
			}
			TSeqRange range;
			if(loc) {
				range = loc->GetTotalRange();
			} else {
			}
			CRef<CTableDataVcf> table_data(new CTableDataVcf(*scope, handle, range, m_NonAsnTrackData->GetBlob()));

			if (table_data) {
				table_data->LoadData();
				table_data->AddListener(this, ePool_Parent);

				CGridWidget* widget = (CGridWidget*)m_Window;

				if (doc) {
					widget->Init(*table_data, &doc->GetUndoManager());
					//!! avoid an auto-resize to check whether all data is requested at once
					//!! widget->GetGrid()->AutoSizeRows();
					x_AttachToProject(*doc);
					x_UpdateContentLabel();
					return true;
				}
				return false;
			}
		}
	}
    // cannot represent the data
    x_ReportInvalidInputData(objects);
    return false;
}

void CVcfTableView::OnProjectChanged(CProjectViewEvent& evt)
{
    if (evt.GetSubtype() == CProjectViewEvent::eData ||
        evt.GetSubtype() == CProjectViewEvent::eBothPropsData) {
        m_Window->Refresh();
    }
}

void CVcfTableView::OnWidgetSelChanged()
{
	if(m_SelectionService) {
		m_SelectionService->OnSelectionChanged(this);
		BroadcastVisibleRange(CVisibleRange::eBasic_Slave);
	}
}

void CVcfTableView::x_OnSetSelection(CSelectionEvent& evt)
{
	if (m_Window  &&  x_HasProject()) {
		CGridWidget* grid = (CGridWidget*)m_Window;
		size_t nFirstRow(grid->SetSelection(evt));
		grid->ScrollToRow(nFirstRow);
	}
}

void CVcfTableView::GetSelection(CSelectionEvent& evt) const
{
    if (m_Window  &&  x_HasProject()) {
        CGridWidget* grid = (CGridWidget*)m_Window;

        TConstScopedObjects sobjs;
        grid->GetSelectedObjects (sobjs);

        TConstObjects objs;
        ITERATE(TConstScopedObjects, it, sobjs) {
            objs.push_back(it->object);
        }

        evt.AddObjectSelection(objs);
    }
}

/// Populate a visible range event prior to broadcasting
void CVcfTableView::GetVisibleRanges(CVisibleRange& vrange) const
{
	TConstScopedObjects sel_objs;
	GetSelection(sel_objs);

	ITERATE(TConstScopedObjects, it, sel_objs)   {
		const CSeq_loc   * loc = dynamic_cast<const CSeq_loc*>(it->object.GetPointer());
		const CSeq_feat  * sft = dynamic_cast<const CSeq_feat*>(it->object.GetPointer());
		const CSeq_annot * ant = dynamic_cast<const CSeq_annot*>(it->object.GetPointer());
		const CVcfVariant *vant = dynamic_cast<const CVcfVariant*>(it->object.GetPointer());
		if(loc) {
			vrange.AddLocation(*loc);
		}
		else if (sft) {
			vrange.AddLocation(sft->GetLocation());
		}
		else if (vant) {
			vrange.AddLocation(vant->GetLocation());
		}
		else if (ant) {
			const CSeq_annot::C_Data::TFtable &feats =ant->GetData().GetFtable();
			ITERATE (CSeq_annot::C_Data::TFtable, feat, feats) {
				vrange.AddLocation((*feat)->GetLocation());
			}
		}
	}
}


void CVcfTableView::GetSelection(TConstScopedObjects& objs) const
{
    if (m_Window  &&  x_HasProject()) {
        CGridWidget* grid = (CGridWidget*)m_Window;
        grid->GetSelectedObjects (objs);
    }
}


void CVcfTableView::GetMainObject(TConstScopedObjects& objs) const
{
    if (m_Window  &&  x_HasProject()) {
        /*CGridWidget* grid = (CGridWidget*)m_Window;*/

        if(m_OrigObjects.size() > 0) {
            const SConstScopedObject& ScopedObject(m_OrigObjects[0]);
            const CObject* object = ScopedObject.object.GetPointer();
            if (object) {
                CScope* scope(const_cast<CScope*>(ScopedObject.scope.GetPointer()));
                objs.push_back(SConstScopedObject(object, scope));
            }
        }
    }
}

// override function to suppress Navigation toolbar
void CVcfTableView::GetCompatibleToolBars(vector<string>& names)
{
    names.push_back(CProjectViewToolBatFactory::sm_SelTBName);
}

const CObject* CVcfTableView::x_GetOrigObject() const
{
    if(m_OrigObjects.size() > 0) {
        return m_OrigObjects[0].object.GetPointer();
    }

    return NULL;
}

void CVcfTableView::x_OnUrlHover(CEvent* evt)
{
    CGridWidgetEvent* gridEvt = dynamic_cast<CGridWidgetEvent*>(evt);
    if (!gridEvt) return;

    CIRef<IStatusBarService> sb_srv = m_Workbench->GetServiceByType<IStatusBarService>();
    sb_srv->SetStatusMessage(ToStdString(gridEvt->GetURL()));
}

void CVcfTableView::x_QueryStatusChange(CEvent* evt)
{
    CQueryPanelEvent* queryEvt = dynamic_cast<CQueryPanelEvent*>(evt);
    if (!queryEvt) return;

    CIRef<IStatusBarService> sb_srv = m_Workbench->GetServiceByType<IStatusBarService>();
    sb_srv->SetStatusMessage(queryEvt->GetStatus());
}

void CVcfTableView::x_OnDataChanged(CEvent* evt)
{
    if (!m_Workbench)
        return;

    CVcfTableEvent* tableEvt(dynamic_cast<CVcfTableEvent*>(evt));
    if (!tableEvt) return;

    CIRef<ITableData> table_data = tableEvt->GetTableData();

    CGridWidget* widget = (CGridWidget*)GetWindow();
    TConstScopedObjects objects;
    GetMainObject(objects);

    widget->Init(*table_data, x_GetUndoManager());
	//!! this probably affects asking for all data at once!
	//!! maybe auto sizing can be done based on the first 100 rows instead of trying to grab all of them?
    if(table_data->GetRowsCount() < 100) {
		widget->GetGrid()->AutoSizeRows();
    }
	widget->Refresh();
}

void CVcfTableView::SetRegistryPath(const string& reg_path)
{
    IRegSettings* reg_set = dynamic_cast<IRegSettings*>(m_Window);
    if (reg_set)
        reg_set->SetRegistryPath(reg_path);
}

void CVcfTableView::LoadSettings()
{
    IRegSettings* reg_set = dynamic_cast<IRegSettings*>(m_Window);
    if (reg_set)
        reg_set->LoadSettings();
}


void CVcfTableView::SaveSettings() const
{
    IRegSettings* reg_set = dynamic_cast<IRegSettings*>(m_Window);
    if (reg_set)
        reg_set->SaveSettings();
}

///////////////////////////////////////////////////////////////////////////////
/// CVcfTableViewFactory
string  CVcfTableViewFactory::GetExtensionIdentifier() const
{
    static string sid("vcf_table_view_factory");
    return sid;
}

string CVcfTableViewFactory::GetExtensionLabel() const
{
    static string slabel("VCF Table View Factory");
    return slabel;
}

void CVcfTableViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
	provider.RegisterFileAlias(ToWxString(alias), wxT("feature_table_view.png"));
}


const CProjectViewTypeDescriptor& CVcfTableViewFactory::GetProjectViewTypeDescriptor() const
{
    return s_VcfViewTypeDescr;
}

vector<CUICommand*> CVcfTableViewFactory::GetCommands()
{
    vector<CUICommand*> Commands;
/*    CUICommand* cmd(new CUICommand(CVcfGridEvtExt::eCmdLinkRefVcf,
                                   "&RefVcf Link",
                                   "RefVcf Link",
                                   ""));
    Commands.push_back(cmd);
*/
    return Commands;
}

IView* CVcfTableViewFactory::CreateInstance() const
{
    return new CVcfTableView();
}


IView* CVcfTableViewFactory::CreateInstanceByFingerprint(const TFingerprint& /*fingerprint*/) const
{
    return NULL;
}


int CVcfTableViewFactory::TestInputObjects(TConstScopedObjects& objects)
{
    bool found_good = false;
    bool found_bad = false;
    for( size_t i = 0; i < objects.size(); i++ ){
        const CObject* obj = objects[i].object;
        const type_info& type = typeid(*obj);
		bool isThisGood(typeid(CSeq_loc) == type);

		if(isThisGood && !found_good) {
			found_good = true;
		}
		if(!isThisGood && !found_bad) {
			found_bad = true;
		}
    }
    if( found_good ){
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);
    }
    return fCanShowNone;
}

END_NCBI_SCOPE
