/*  
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Author: Andrei Shkeda
 *
 * File Description:
 *
 * ===========================================================================
 */

#include <ncbi_pch.hpp>
#include <corelib/test_boost.hpp>
#include <corelib/ncbifile.hpp>
#include <gui/utils/seq_dict.hpp>
#include <random>

static const string ALPHABET_DNA = "ACGTN";
static const string ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXZY";

string gen_random(std::mt19937& gen, std::uniform_int_distribution<>& alphabet_distrib, const int len) 
{
    string tmp_s;
    tmp_s.reserve(len);
    for (int i = 0; i < len; ++i) {
        tmp_s += ALPHABET[alphabet_distrib(gen)];
    }
    return tmp_s;
}


BOOST_AUTO_TEST_SUITE(TestSuiteSeqDictinary)


BOOST_AUTO_TEST_CASE(Test0)
{
    CSeqDictionaryBuilder<> builder;
    
    auto acgt_idx = builder.AddSequence("ACGT");
    auto aaaa_idx = builder.AddSequence("AAAA");
    auto cccc_idx = builder.AddSequence("CCCC");

    BOOST_CHECK_EQUAL(builder.AddSequence("AAAA"), aaaa_idx);
    BOOST_CHECK_EQUAL(builder.AddSequence("CCCC"), cccc_idx);
    BOOST_CHECK_EQUAL(builder.AddSequence("ACGT"), acgt_idx);

    BOOST_CHECK_EQUAL(builder.AddSequence("A"), 0);
    BOOST_CHECK_EQUAL(builder.AddSequence("C"), CSeqDictionaryBuilder<>::TIndex(1) << 32);
    BOOST_CHECK_EQUAL(builder.AddSequence("G"), CSeqDictionaryBuilder<>::TIndex(2) << 32);
    BOOST_CHECK_EQUAL(builder.AddSequence("T"), CSeqDictionaryBuilder<>::TIndex(3) << 32);
    
    auto r_index = builder.AddRegion(55);

    auto sd = builder.Build();
    string s;

    BOOST_CHECK_EQUAL(sd->GetSequence(acgt_idx, s), "ACGT");
    BOOST_CHECK_EQUAL(sd->GetSequence(aaaa_idx, s), "AAAA");
    BOOST_CHECK_EQUAL(sd->GetSequence(cccc_idx, s), "CCCC");

    BOOST_CHECK(sd->IsRegion(acgt_idx) == false);
    BOOST_CHECK_EQUAL(sd->GetSeqLength(acgt_idx), 4);

    BOOST_CHECK(sd->IsRegion(r_index) == true);
    BOOST_CHECK_EQUAL(sd->GetSeqLength(r_index), 55);

}

BOOST_AUTO_TEST_CASE(StressTest)
{
    vector<string> test_set;
    std::random_device rd;  //Will be used to obtain a seed for the random number engine
    std::mt19937 gen(rd()); //Standard mersenne_twister_engine seeded with rd()
    std::uniform_int_distribution<> distrib(1, 5000);
    std::uniform_int_distribution<> alphabet_distrib(0, ALPHABET.size() - 1);
    for (size_t i = 0; i < 1000; ++i) {
        test_set.push_back(gen_random(gen, alphabet_distrib, distrib(gen)));
    }

    CSeqDictionaryBuilder<> builder;
    vector<CSeqDictionaryBuilder<>::TIndex> idx_set;
    {
        for_each(test_set.begin(), test_set.end(), [&](const string& s) { 
            auto index = builder.AddSequence(s);
            idx_set.push_back(index);
        });
    }
/*    
    set<CSeqDictionaryBuilder<>::TIndex> idx_set2;
    copy(idx_set.begin(), idx_set.end(), inserter(idx_set2, idx_set2.begin()));
    bool dump = false;
    for_each(test_set.begin(), test_set.end(), [&](const string& s) { 
        auto i = builder.AddSequence(s);
        BOOST_CHECK_MESSAGE(idx_set2.count(i) != 0, "Offset: " << (i >> 32) << ", len:" << (i & 0x00000000FFFFFFFF) << ", string: '" << s << "'");
    });
*/    
    auto sd = builder.Build();
    for (size_t i = 0; i < idx_set.size(); ++i) {
        string s;
        auto index = idx_set[i];
        sd->GetSequence(index, s);
        BOOST_CHECK(s.compare(test_set[i]) == 0);
    }
    ncbi::CTmpFile tmpfile;
    string tmp_file_name = tmpfile.GetFileName();
    {
        ofstream os(tmp_file_name);
        BOOST_CHECK(sd->Serialize(os));
    }

    ifstream t(tmp_file_name.c_str());
    string data((istreambuf_iterator<char>(t)), istreambuf_iterator<char>());
    {
        BOOST_TEST_CHECKPOINT("Deserialize");
        BOOST_CHECK(sd->Deserialize(data.c_str()) > 0);
    }

}

BOOST_AUTO_TEST_CASE(TestEmpty)
{

    CSeqDictionary<> sd;
    string s;
    CSeqDictionary<>::TIndex idx = 555;
    idx <<= 32;
    idx |= 10;
    sd.GetSequence(idx, s);
    BOOST_CHECK(s.empty());

    idx = 1;
    idx <<= 32;
    sd.GetSequence(idx, s);
    BOOST_CHECK_EQUAL(s, "C");
}


BOOST_AUTO_TEST_SUITE_END();

