/*  $Id: columnar_vcf_iterator.cpp 45881 2021-01-05 03:10:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrea Asztalos
 *
 * File Description:
 *   
 *
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/loaders/columnar_vcf_iterator.hpp>

BEGIN_NCBI_SCOPE


CVCFVariant_CI::CVCFVariant_CI(const CVCFVariantsBase& var_list, const TSeqRange& range)
    : m_Variants(&var_list)
{
    if (range.NotEmpty()) {
        m_Start = m_Variants->GetIndexForPosition(range.GetFrom(), CPosToIndex::eMin);
        m_Stop = m_Variants->GetIndexForPosition(range.GetTo(), CPosToIndex::eMax);

        const auto& descr = m_Variants->GetDescriptors();
        const auto& cols_name = CVCFVariantsBase::s_GetColNames();
        for (const auto& it : cols_name) {
            if (!descr[it].empty()) {
                unique_ptr<TSparseCI> new_iter = make_unique<TSparseCI>(&descr[it], m_Start);
                m_Iterators.emplace(it, move(new_iter));
            }
        }

        auto info_fields = descr.GetInfoFieldNames();
        for (const auto& it : info_fields) {
            if (!descr.GetInfoField(it).empty()) {
                unique_ptr<TInfoCI> new_iter = make_unique<TInfoCI>(&descr.GetInfoField(it), m_Start);
                m_InfoIterators.emplace(it, move(new_iter));
            }
        }

        auto samples = descr.GetSampleNames();
        for (const auto& it : samples) {
            if (!descr.GetSample(it).empty()) {
                unique_ptr<TInfoCI> new_iter = make_unique<TInfoCI>(&descr.GetSample(it), m_Start);
                m_SampleIterators.emplace(it, move(new_iter));
            }
        }
        m_Current = m_Start;
    }
}

CVCFVariant_CI& CVCFVariant_CI::begin()
{
    _ASSERT(!m_Iterators.empty());
    for (auto& it : m_Iterators) {
        it.second->go_to(m_Start);
    }
    for (auto& it : m_InfoIterators) {
        it.second->go_to(m_Start);
    }
    for (auto& it : m_SampleIterators) {
        it.second->go_to(m_Start);
    }
    m_Current = m_Start;
    return *this;
}

CVCFVariant_CI& CVCFVariant_CI::operator++()
{
    _ASSERT(!m_Iterators.empty());
    for (auto& it : m_Iterators) {
        it.second->advance();
    }
    for (auto& it : m_InfoIterators) {
        it.second->advance();
    }
    for (auto& it : m_SampleIterators) {
        it.second->advance();
    }
    ++m_Current;
    return *this;
}

bool CVCFVariant_CI::IsValid() const
{
    if (m_Iterators.empty())
        return false;
    bool valid = true;
    for (auto it = m_Iterators.begin(); valid && it != m_Iterators.end(); ++it) {
        valid = m_Current < m_Stop && (*it).second->valid();
    }
    for (auto it = m_InfoIterators.begin(); valid && it != m_InfoIterators.end(); ++it) {
        valid = m_Current < m_Stop && (*it).second->valid();
    }
    for (auto it = m_SampleIterators.begin(); valid && it != m_SampleIterators.end(); ++it) {
        valid = m_Current < m_Stop && (*it).second->valid();
    }
    return valid;
}

size_t CVCFVariant_CI::GetSize() const
{
    return m_Stop - m_Start;
}

unsigned CVCFVariant_CI::GetPosition() const
{
    return m_Variants->GetPositionForIndex(m_Current);
}

bool CVCFVariant_CI::x_IsSetData(const string& col_name) const
{
    auto it = m_Iterators.find(col_name);
    if (it != m_Iterators.end()) {
        return true;
    }
    if (col_name == CVCFVariantsBase::sm_INFO) {
        return IsSetInfo();
    }
    return false;
}

string CVCFVariant_CI::x_GetData(const string& col_name) const
{
    auto it = m_Iterators.find(col_name);
    if (it != m_Iterators.end()) {
        if (it->second->is_null())
            return CVCFVariantsBase::sm_MissingValue;
        else
            return it->second->value();
    }
    if (col_name == CVCFVariantsBase::sm_INFO) {
        return GetInfo();
    }
    NCBI_THROW(CException, eUnknown, "CVCFVariant_CI::x_GetData() cannot find column: " + col_name);
}

bool CVCFVariant_CI::IsSetInfo() const
{
    return !m_Variants->GetDescriptors().GetInfoFieldNames().empty();
}

string CVCFVariant_CI::GetInfo() const
{
    string info_value;
    for (const auto& it : m_InfoIterators) {
        if (it.second->is_null()) {
            // don't list that key
        }
        else {
            info_value += it.first;
            string key_val = it.second->value();
            if (!key_val.empty()) {
                info_value += "=" + key_val;
            }
            info_value += ";";
        }
    }
    if (info_value.back() == ';') 
        info_value.pop_back();

    return info_value;
}

bool CVCFVariant_CI::IsSetInfoField(const string& field_name) const
{
    auto info_iter = m_InfoIterators.find(field_name);
    return (info_iter != m_InfoIterators.end() && !info_iter->second->is_null());
}

string CVCFVariant_CI::GetInfoField(const string& field_name) const
{
    string info_value;
    auto info_iter = m_InfoIterators.find(field_name);
    if (info_iter == m_InfoIterators.end()) {
        NCBI_THROW(CException, eUnknown, "CVCFVariant_CI::GetInfoField() cannot find INFO field " + field_name);
    }

    if (info_iter->second->is_null()) {
        info_value = "null";
    } else {
        info_value = info_iter->second->value();
    }
    return info_value;
}

bool CVCFVariant_CI::IsSetSample(const string& name) const
{
    auto sample_iter = m_SampleIterators.find(name);
    return (sample_iter != m_SampleIterators.end() && !sample_iter->second->is_null());
}

string CVCFVariant_CI::GetSample(const string& name) const
{
    string sample;
    auto sample_it = m_SampleIterators.find(name);
    if (sample_it == m_SampleIterators.end()) {
        NCBI_THROW(CException, eUnknown, "CVCFVariant_CI::GetSample() cannot find Sample: " + name);
    }

    if (sample_it->second->is_null()) {
        sample = "null";
    }
    else {
        sample = sample_it->second->value();
    }
    return sample;
}

bool CVCFVariant_CI::operator==(const CVCFVariant_CI& other) const
{
    return (*m_Variants == *other.m_Variants &&
        m_Start == other.m_Start &&
        m_Stop == other.m_Stop &&
        m_Current == other.m_Current &&
        m_Iterators == other.m_Iterators &&
        m_InfoIterators == other.m_InfoIterators &&
        m_SampleIterators == other.m_SampleIterators);
}

bool CVCFVariant_CI::operator!=(const CVCFVariant_CI& other) const
{
    return !operator==(other);
}


END_NCBI_SCOPE

