/*  $Id: columnar_vcf_variants_base.cpp 46179 2021-01-29 13:38:50Z asztalos $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrea Asztalos, Anatoliy Kuznetsov
 *
 * File Description:
 *   
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistl.hpp>
#include <corelib/version.hpp>
#include <util/line_reader.hpp>
#include <util/bitset/ncbi_bitset.hpp>
#include <util/bitset/bmdbg.h>
#include <util/bitset/bmtask.h>
#include <util/bitset/bmsparsevec_parallel.h>
#include <util/bitset/bmthreadpool.h>
#include <gui/widgets/loaders/columnar_vcf_ex.hpp>
#include <gui/widgets/loaders/columnar_vcf_iterator.hpp>
#include <gui/widgets/loaders/columnar_vcf_variants_base.hpp>
#include <chrono>
#include <numeric>
#include <thread>

BEGIN_NCBI_SCOPE

namespace {
    template<typename SV>
    void PrintSparseVector(const SV& vec, CNcbiOstream& out)
    {
        typename SV::const_iterator it = vec.begin();
        typename SV::const_iterator it_end = vec.end();

        for (size_t index = 0; it != it_end; ++it, ++index) {
            if (!it.is_null()) {
                out << index << ": " << *it << endl;
            }
            else {
                out << index << ": null" << endl;
            }
        }
    }

    string GenerateColFileName(const string& prefix, const string& col_name)
    {
        return prefix + "_" + col_name + ".bin";
    }

    void PrintToFile(const char* buff, size_t size, const string& fname)
    {
        if (size == 0)
            return;

        CFileIO fio;
        fio.Open(fname, CFileIO::eCreate, CFileIO::eReadWrite);
        fio.Write(buff, size);
        fio.Close();
    }

    void PrintStats(const bm::bv_statistics& sum, CNcbiOstream& out)
    {
        out << "Bit blocks:       " << sum.bit_blocks << endl;
        out << "Gap blocks:       " << sum.gap_blocks << endl;
        out << "Max serialize mem:" << sum.max_serialize_mem << " "
            << (sum.max_serialize_mem / (1024 * 1024)) << "MB" << endl;
        out << "Memory used:      " << sum.memory_used << " "
            << (sum.memory_used / (1024 * 1024)) << "MB" << endl;
    }

    template<class SV>
    void AddStats(bm::bv_statistics& sum, SV& vec, CNcbiOstream& out)
    {
        typename SV::statistics stat;
        vec.calc_stat(&stat);
        sum.add(stat);
        out << "\nCumulative statistics:\n";
        PrintStats(sum, out);
    }

    template<class SV>
    void SerializeColumn(SV& vec, const string& prefix, const string& col_name, CNcbiOstream* out, unsigned& cum_memory_used, unsigned& cum_layout_size)
    {
        auto start = chrono::steady_clock::now();

        typename SV::statistics stat;
        vec.calc_stat(&stat);
        
        if (out) {
            *out << "Used memory for " << col_name << " vector: " << stat.memory_used << endl;
            *out << "Max serialize memory (" << col_name << "): " << stat.max_serialize_mem << endl;
        }

        bm::sparse_vector_serializer<SV> serializer;
        bm::sparse_vector_serial_layout<SV> layout;
        serializer.enable_xor_compression();
        serializer.serialize(vec, layout);
        PrintToFile((char*)layout.buf(), layout.size(), GenerateColFileName(prefix, col_name));

        auto diff = chrono::steady_clock::now() - start;
        if (out) {
            *out << "Serialization time for " << col_name << ": " << chrono::duration_cast<chrono::milliseconds>(diff).count() << " ms " << endl;
            *out << "Serialized layout size " << col_name << ": " << layout.size() << endl << endl;
        }

        cum_memory_used += stat.memory_used;
        cum_layout_size += layout.size();
    }

    template<class SV>
    void DeserializeColumn(SV& vec, const string& prefix, const string& col_name, CNcbiOstream* out)
    {
        auto start = chrono::steady_clock::now();

        const string filename = GenerateColFileName(prefix, col_name);
        CFileIO fio;
        fio.Open(filename, CFileIO::eOpen, CFileIO::eRead);
        char* buffer = new char[fio.GetFileSize()];
        fio.Read(buffer, fio.GetFileSize());

        SV tmp(bm::use_null); // it's ok for both type of vectors
        bm::sparse_vector_deserializer<SV> deserializer;
        deserializer.deserialize(tmp, (unsigned char*)buffer);

        auto diff = chrono::steady_clock::now() - start;
        if (out) {
            *out << "Deserialization time of " << filename << ": "
                << chrono::duration_cast<chrono::milliseconds>(diff).count() << " ms" << endl;
        }

        typename SV::size_type pos;
        bool found = bm::sparse_vector_find_first_mismatch(tmp, vec, pos, bm::use_null);
        if (found) {
            stringstream err_msg;
            err_msg << "Mismatch found at position: " << pos
                << "\nValues are: " + vec[pos] << " and " << tmp[pos] << endl;
            NCBI_THROW(CException, eUnknown,
                "Deserialized and stored " + filename + " columns are not identical: " + err_msg.str());
        }

        delete[] buffer;
        fio.Close();
    }

    bool RemoveFile(const string& fname)
    {
        bool status = false;
        CFile fh(fname);
        if (fh.Exists() && fh.IsFile()) {
            status = fh.Remove(CDirEntry::eEntryOnly);
        }
        return status;
    }
}


/////// CvariantDescriptors

CVariantDescriptors::CVariantDescriptors(bool load_info, const set<string>& info_fields, 
    const map<unsigned, string>& sample_cols)
    : m_LoadAllInfo(load_info), m_LoadInfoFields(info_fields), m_LoadSamples(sample_cols)
{
    auto& fields = CVCFVariantsBase::s_GetColNames();
    auto nr_cols = fields.size();

    m_DescrMatrix.reserve(nr_cols);
    size_t index = 0;
    // Many columns support the MISSING value '.', which will be stored as NULL
    for (const auto& it : fields) {
        m_DescrMatrix.emplace_back(make_unique<TSparseStrVector>(bm::use_null));
        m_DescrLabels.emplace(it, index++);
    }

    index = 0;
    m_SampleFields.reserve(m_LoadSamples.size());
    m_SampleIndices.reserve(m_LoadSamples.size());
    for (const auto& it : m_LoadSamples) {
        m_SampleFields.emplace_back(it.second, make_unique<TSparseOptVector>(bm::use_null));
        m_SampleIndices.emplace(it.second, index++);
    }
}

void CVariantDescriptors::InitInserters()
{
    _ASSERT(m_BackInserters.empty());
    auto nr_cols = CVCFVariantsBase::s_GetColNames().size();
    m_BackInserters.reserve(nr_cols);
    for (const auto& it : m_DescrMatrix) {
        m_BackInserters.emplace_back(make_unique<TInserter>(it.get()));
    }

    _ASSERT(m_SampleInserters.empty());
    m_SampleInserters.reserve(m_SampleFields.size());
    for (const auto& it : m_SampleFields) {
        m_SampleInserters.emplace_back(make_unique<TOptInserter>(it.second.get()));
    }
}

void CVariantDescriptors::PushBack(const string& label, const string& value)
{
    auto it = m_DescrLabels.find(label);
    if (it != m_DescrLabels.end()) {
        size_t index = it->second;
        if (value == CVCFVariantsBase::sm_MissingValue || value.empty()) {
            m_BackInserters[index]->add_null();
        }
        else {
            if (label == CVCFVariantsBase::sm_FORMAT) {
                string decoded = NStr::URLDecode(value);
                *m_BackInserters[index].get() = decoded;
            }
            else {
                *m_BackInserters[index].get() = value;
            }
        }
    } else if (NStr::EqualNocase(label, CVCFVariantsBase::sm_INFO)) {
        if (m_LoadAllInfo || !m_LoadInfoFields.empty()) {
            vector<string> info_vals;
            NStr::Split(value, ";", info_vals);

            unordered_map<string, bool> field_found_map;
            if (!m_InfoFields.empty()) {
                for (const auto& it : m_InfoFields) {
                    field_found_map.emplace(it.first, false);
                }
            }

            for (const auto& it : info_vals) {
                string key = it, key_val;

                auto eq_pos = it.find("=");
                if (eq_pos != NPOS) {
                    key = it.substr(0, eq_pos);
                    key_val = it.substr(eq_pos + 1);
                    NStr::URLDecodeInPlace(key_val);
                }
                
                if (m_LoadAllInfo || (m_LoadInfoFields.find(key) != m_LoadInfoFields.end())) {
                    auto key_it = m_InfoFields.find(key);
                    if (key_it == m_InfoFields.end()) {
                        auto inserted = m_InfoFields.emplace(key, make_unique<TSparseOptVector>(bm::use_null));
                        auto info_insert = make_unique<TOptInserter>(inserted.first->second.get());
                        if (m_Index != 0) {
                            info_insert->add_null(m_Index);
                        }
                        *info_insert = key_val;
                        m_InfoInserters.insert(make_pair(key, move(info_insert)));
                    }
                    else {
                        auto insert_it = m_InfoInserters.find(key);
                        *insert_it->second.get() = key_val;
                        field_found_map[key] = true;
                    }
                }
            }

            for (const auto& it : field_found_map) {
                if (it.second == false) {
                    auto insert_it = m_InfoInserters.find(it.first);
                    _ASSERT(insert_it != m_InfoInserters.end());
                    insert_it->second->add_null();
                }
            }
        }
    }
    else if (NStr::EqualNocase(label, CVCFVariantsBase::sm_SAMPLES)) { 
        if (!m_SampleFields.empty()) {
            if (value.empty()) {
                for (auto& it : m_SampleInserters) {
                    it->add_null();
                }
            }
            else {
                vector<string> vals;
                NStr::Split(value, "\t", vals);

                if (value.find(" ") != NPOS) {
                    NCBI_THROW(CColumnarVCFException, eParseTabExpected, "Sample values are expected to be tab delimited");
                }

                if (vals.size() == m_SampleFields.size()) {
                    for (size_t index = 0; index < vals.size(); ++index) {
                        *m_SampleInserters[index].get() = vals[index];
                    }
                }
                else {
                    auto insert_it = m_SampleInserters.begin();
                    for (unsigned i = 0; i < vals.size() && insert_it != m_SampleInserters.end(); ++i) {
                        auto found_index = m_LoadSamples.find(i);
                        if (found_index != m_LoadSamples.end()) {
                            *insert_it->get() = vals[i];
                            ++insert_it;
                        }
                    }
                }
            }
        }
    } else {
        NCBI_THROW(CException, eUnknown, "CVariantDescriptors::PushBack(): Unknown label - " + label);
    }
    // NOTE: the values are not actually added to the vectors until the insert iterators are flushed
}

void CVariantDescriptors::FinalizeReading()
{
    // Assuming that reading has finished, the insert iterators are flushed and deleted
    for (auto& it : m_BackInserters) {
        it->flush();
    }
    m_BackInserters.clear();

    for (auto& it : m_InfoInserters) {
        it.second->flush();
    }
    m_InfoInserters.clear();
    
    for (auto& it : m_SampleInserters) {
        it->flush();
    }
    m_SampleInserters.clear();

    m_LoadAllInfo = false;
    m_LoadInfoFields.clear();
    
    auto vec_size = m_DescrMatrix[0]->size();
    for (const auto& it : m_InfoFields) {
        if (vec_size != it.second->size()) {
            LOG_POST(Error << "Info field " << it.first << " has different size (" <<
                it.second->size() << ") from the expected one (" << vec_size << ")");
        }
    }

    for (const auto& it : m_SampleFields) {
        if (vec_size != it.second->size()) {
            LOG_POST(Error << "Sample " << it.first << " has different size (" <<
                it.second->size() << ") from the expected one (" << vec_size << ")");
        }
    }
    // Optimize the vectors
    BM_DECLARE_TEMP_BLOCK(tb);
    for (auto&& it : m_DescrMatrix) {
        it->remap();
        it->optimize(tb);
    }

    for (auto&& it : m_InfoFields) {
        it.second->remap();
        it.second->optimize(tb);
    }

    for (auto&& it : m_SampleFields) {
        it.second->remap();
        it.second->optimize(tb);
    }
}

CVariantDescriptors::TSparseStrVector& CVariantDescriptors::operator[] (const string& label)
{
    auto it = m_DescrLabels.find(label);
    if (it != m_DescrLabels.end()) {
        return *m_DescrMatrix[it->second];
    }

    NCBI_THROW(CException, eUnknown, "CVariantDescriptors::operator[]: " + label + " was not recognized\n");
}

const CVariantDescriptors::TSparseStrVector& CVariantDescriptors::operator[] (const string& label) const
{
    auto it = m_DescrLabels.find(label);
    if (it != m_DescrLabels.end()) {
        return *m_DescrMatrix[it->second];
    }

    NCBI_THROW(CException, eUnknown, "CVariantDescriptors::operator[]: " + label + " was not recognized\n");
}

vector<string> CVariantDescriptors::GetInfoFieldNames() const
{
    vector<string> field_names;
    for (const auto& it : m_InfoFields) {
        field_names.push_back(it.first);
    }
    return field_names;
}
const CVariantDescriptors::TSparseOptVector& CVariantDescriptors::GetInfoField(const string& field_name) const
{
    auto info_field = m_InfoFields.find(field_name);
    if (info_field != m_InfoFields.end()) {
        return *info_field->second;
    }
    NCBI_THROW(CException, eUnknown, "CVariantDescriptors::GetInfoField(): " + field_name + " was not recognized\n");
}

void CVariantDescriptors::SetNumberofInfoFields(size_t nr_info)
{
    m_InfoFields.reserve(nr_info);
}

CVariantDescriptors::TSparseOptVector& CVariantDescriptors::SetInfoField(const string& field_name)
{
    auto info_field = m_InfoFields.find(field_name);
    if (info_field != m_InfoFields.end()) {
        return *info_field->second;
    }

    auto inserted = m_InfoFields.emplace(field_name, make_unique<TSparseOptVector>(bm::use_null));
    return *inserted.first->second;
}

vector<string> CVariantDescriptors::GetSampleNames() const
{
    vector<string> names;
    for (const auto& it : m_SampleFields) {
        names.push_back(it.first);
    }
    return names;
}

const CVariantDescriptors::TSparseOptVector& CVariantDescriptors::GetSample(const string& name) const
{
    auto sample_it = m_SampleIndices.find(name);
    if (sample_it != m_SampleIndices.end()) {
        return *m_SampleFields[sample_it->second].second;
    }
    NCBI_THROW(CException, eUnknown, "CVariantDescriptors::GetSample(): " + name + " was not recognized\n");
}

void CVariantDescriptors::SetNumberofSampleCols(size_t nr_samples)
{
    m_SampleFields.reserve(nr_samples);
    m_SampleIndices.reserve(nr_samples);
}

CVariantDescriptors::TSparseOptVector& CVariantDescriptors::SetSample(const string& name)
{
    auto sample_it = m_SampleIndices.find(name);
    if (sample_it != m_SampleIndices.end()) {
        return *m_SampleFields[sample_it->second].second;
    }

    m_SampleFields.emplace_back(name, make_unique<TSparseOptVector>(bm::use_null));
    m_SampleIndices.emplace(name, m_SampleFields.size() - 1);
    return *m_SampleFields.back().second;
}

void CVariantDescriptors::GetIndicesForMissingVarID(TSparseStrVector::bvector_type& indices) const
{
    auto& id_vector = *m_DescrMatrix[0].get();
    const auto& bv_non_null = *id_vector.get_null_bvector();

    indices = bv_non_null;
    indices.resize(bv_non_null.count());
    indices.invert();

    /*
    cout << "Found hits (GetIndicesForMissingVarID): " << indices.count() << endl;
    auto en = indices.first();
    auto en_end = indices.end();
    for (; en < en_end; ++en) {
        cout << *en << endl;
    }
    */
}

bool CVariantDescriptors::GetIndicesForVariant(const string& variant_id, TSparseStrVector::bvector_type& indices) const
{
    auto& id_vector = *m_DescrMatrix[0].get();
    return m_Scanner.find_eq_str(id_vector, variant_id.data(), indices);
}

bool CVariantDescriptors::operator==(const CVariantDescriptors& other) const
{
    if ((m_DescrMatrix.size() != other.m_DescrMatrix.size()) ||
        (m_DescrLabels != other.m_DescrLabels) ||
        (m_InfoFields.size() != other.m_InfoFields.size()) ||
        (m_SampleFields.size() != other.m_SampleFields.size())||
        (m_SampleIndices.size() != other.m_SampleIndices.size()))
        return false;
    {
        auto it_this = m_DescrMatrix.begin();
        auto it_that = other.m_DescrMatrix.begin();
        for (; it_this != m_DescrMatrix.end(); ++it_this, ++it_that) {
            if (!(*it_this)->equal(*it_that->get(), bm::use_null))
                return false;
        }
    }

    for (auto it_this = m_InfoFields.begin(); it_this != m_InfoFields.end(); ++it_this) {
        auto it_that = other.m_InfoFields.find(it_this->first);
        if (it_that == other.m_InfoFields.end()) {
            return false;
        }
        if (!it_this->second->equal(*it_that->second.get(), bm::use_null))
            return false;
    }
    {
        auto it_this = m_SampleFields.begin();
        auto it_that = other.m_SampleFields.begin();
        for (; it_this != m_SampleFields.end(); ++it_this, ++it_that) {
            if (!(*it_this).second->equal(*it_that->second.get(), bm::use_null))
                return false;
        }
    }
    return true;
}

void CVariantDescriptors::Clear()
{
    for (auto& it : m_DescrMatrix) {
        it->clear();
    }
    // TODO - reevaluate this function
    for (auto& it : m_InfoFields) {
        it.second->clear();
    }
    m_InfoFields.clear();

    for (auto& it : m_SampleFields) {
        it.second->clear();
    }
    m_SampleFields.clear();
    m_SampleIndices.clear();
}


/////// CPosToIndex
CPosToIndex::CPosToIndex()
{
    m_Indices.emplace_back(make_unique<TIndexVector>());
}

bool CPosToIndex::Add(const unsigned& index, const unsigned& value)
{
    // initialize back_insert_iterators if necessary
    if (m_BackInserters.empty()) {
        for (const auto& it : m_Indices) {
            m_BackInserters.emplace_back(make_unique<TInserter>(it.get()));
        }
    }

    bool added = false;
    auto insert_it = m_BackInserters.begin();
    if (m_LastIndices.empty()) {
        (*insert_it)->add_null(index);
        (*insert_it)->add(value);
        m_LastIndices.push_back(index);
        added = true;
    }
    else {
        for (size_t i = 0; !added && i < m_LastIndices.size(); ++i, ++insert_it) {
            int diff = index - m_LastIndices[i];
            if (diff > 0) {
                if (diff == 1) {
                    (*insert_it)->add(value);
                }
                else {
                    (*insert_it)->add_null(diff - 1);
                    (*insert_it)->add(value);
                }
                m_LastIndices[i] = index;
                added = true;
            }
            else if (diff == 0) {
                continue;
            }
            else if (diff < 0) {
                NCBI_THROW(CException, eUnknown, "Positions are not ordered: " + NStr::UIntToString(index));
            }
        }

        if (!added) {
            m_Indices.emplace_back(make_unique<TIndexVector>());
            unique_ptr<TInserter> new_inserter = make_unique<TInserter>(m_Indices.back().get());
            new_inserter->add_null(index);
            new_inserter->add(value);
            m_BackInserters.push_back(move(new_inserter));
            m_LastIndices.push_back(index);
            added = true;
        }
    }
    return added;
}

void CPosToIndex::FinalizeReading()
{
    // Assuming that reading has finished, the insert iterators are flushed and deleted
    for (auto& it : m_BackInserters) {
        it->flush();
    }
    m_BackInserters.clear();
    m_LastIndices.clear();
    x_SaveMaximumIndex();

    BM_DECLARE_TEMP_BLOCK(tb);
    for (auto& it : m_Indices) {
        it->optimize(tb);
        it->sync();  // build rank-select index (yields faster read access)
    }
}

void CPosToIndex::x_SaveMaximumIndex()
{
    const auto& first_idx = m_Indices.front();
    unsigned last_pos = first_idx->size() - 1;
    m_MaxStoredIndex = first_idx->at(last_pos);
    for (size_t index = 1; index < m_Indices.size(); ++index) {
        if (last_pos < m_Indices[index]->size() && !m_Indices[index]->is_null(last_pos)) {
            m_MaxStoredIndex = max(m_MaxStoredIndex, (size_t)m_Indices[index]->at(last_pos));
        }
    }
}

void CPosToIndex::Lookup(const TSparseStrVector::bvector_type& values, vector<unsigned>& indices) const
{
    auto en = values.first();
    for (; en.valid(); ++en) {
        for (auto& it : m_Indices) {
            auto& idx_vector = *it.get();
            TIndexVector::bvector_type found;
            m_Scanner.find_eq(idx_vector, *en, found);
            if (found.count() > 0) {
                auto found_en = found.first();
                for (; found_en.valid(); ++found_en) {
                    indices.push_back(*found_en);
                }
            }
        }
    }
}

size_t CPosToIndex::GetIndexForPosition(const unsigned& pos, CPosToIndex::EType type) const
{
    const auto& first_idx = m_Indices.front();

    const auto& non_null_pos = first_idx->get_null_bvector();
    auto pos_iter = non_null_pos->first();

    if (pos < *pos_iter) {
        return 0;
        // corner case as all index vectors have higher starting points
    }

    size_t first_vectsize = first_idx->size();
    size_t value;
    if (pos < first_vectsize) {
        auto first_iter = first_idx->get_const_iterator(pos);
        if (!first_iter.is_null()) {
            value = *first_iter;
            for (size_t index = 1; index < m_Indices.size(); ++index) {
                if (pos < m_Indices[index]->size() && !m_Indices[index]->is_null(pos)) {
                    if (type == eMin) {
                        value = min(value, (size_t)m_Indices[index]->at(pos));
                    }
                    else if (type == eMax) {
                        value = max(value, (size_t)m_Indices[index]->at(pos));
                    }
                }
            }
            if (type == eMax) {
                value++;
            }
        }
        else {
            auto pos_end = non_null_pos->end();
            auto pos_prev = non_null_pos->first();
            pos_iter = pos_prev;
            for (++pos_iter; pos_iter < pos_end && *pos_iter < pos; ++pos_iter, ++pos_prev)
            {
            }

            value = first_idx->at(*pos_iter);
        }
    }
    else {
        // both min and max should be to the maximum stored index + 1
        value = m_MaxStoredIndex + 1;
    }
    return value;
}

unsigned CPosToIndex::GetPositionForIndex(const size_t& index) const
{
    bool found = false;
    unsigned pos = 0;
    for (size_t i = 0; !found && i < m_Indices.size(); ++i) {
        auto& idx_vector = *m_Indices[i].get();
        found = m_Scanner.find_eq(idx_vector, index, pos);
    }
    return pos;
}

bool CPosToIndex::operator==(const CPosToIndex& other) const
{
    if ((m_Indices.size() != other.m_Indices.size()) ||
        (m_MaxStoredIndex != other.m_MaxStoredIndex))
        return false;

    auto it_this = m_Indices.begin();
    auto it_that = other.m_Indices.begin();
    for (; it_this != m_Indices.end(); ++it_this, ++it_that) {
        const auto& this_vect = *(it_this->get());
        const auto& that_vect = *(it_that->get());
        TIndexVector::size_type pos;
        bool found = bm::sparse_vector_find_first_mismatch(this_vect, that_vect, pos);
        if (found) {
            return false;
        }
    }
    return true;
}

void CPosToIndex::ListVectors(CNcbiOstream& out) const
{
    out << "There are " << m_Indices.size() << " index vectors " << endl;
    size_t count = 0;
    for (auto& it : m_Indices) {
        ++count;
        out << "Index vector " << count << ":\n";
        PrintSparseVector(*it.get(), out);
    }
}

void CPosToIndex::GetStatistics(bm::bv_statistics& stat_sum, bool Jaccard_index, CNcbiOstream& out)
{
    for (unsigned i = 0; i < m_Indices.size(); ++i) {
        out << "-----------------------Index vector " << NStr::UIntToString(i) << "-------------------------" << endl;
        LOG_POST(Info << "-----------------------Index vector " << NStr::UIntToString(i) << "-------------------------");
        bm::print_svector_stat(*m_Indices[i].get(), Jaccard_index);
        AddStats(stat_sum, *m_Indices[i].get(), out);
    }
}

void CPosToIndex::SerializeVectors(const string& prefix, CNcbiOstream* out, unsigned& cum_memory_used, unsigned& cum_layout_size)
{
    string label = x_GetFilePrefix(prefix);
    for (unsigned count = 0; count < m_Indices.size(); ++count) {
        auto& index_vec = *m_Indices[count].get();
        SerializeColumn(index_vec, label, NStr::UIntToString(count), out, cum_memory_used, cum_layout_size);
    }
}

void CPosToIndex::DeserializeVectors(const string& prefix, CNcbiOstream* out)
{
    string label = x_GetFilePrefix(prefix);
    for (unsigned count = 0; count < m_Indices.size(); ++count) {
        auto& index_vec = *m_Indices[count].get();
        DeserializeColumn(index_vec, label, NStr::UIntToString(count), out);
    }
}

bool CPosToIndex::RemoveSerializedOutput(const string& prefix)
{
    bool status = true;
    string label = x_GetFilePrefix(prefix);
    for (unsigned count = 0; count < m_Indices.size(); ++count) {
        status = status && RemoveFile(GenerateColFileName(label, NStr::UIntToString(count)));
    }
    return status;
}

string CPosToIndex::x_GetFilePrefix(const string& prefix) const
{
    return prefix + "_index";
}


// CVCFVariantsBase

const string CVCFVariantsBase::sm_MissingValue = ".";
const string CVCFVariantsBase::sm_ID = "ID";
const string CVCFVariantsBase::sm_REF = "REF";
const string CVCFVariantsBase::sm_ALT = "ALT";
const string CVCFVariantsBase::sm_QUAL = "QUAL";
const string CVCFVariantsBase::sm_FILTER = "FILTER";
const string CVCFVariantsBase::sm_FORMAT = "FORMAT";

const vector<string>& CVCFVariantsBase::s_GetColNames()
{
    static const vector<string> s_column_names{
        sm_ID, sm_REF, sm_ALT, sm_QUAL, sm_FILTER, sm_FORMAT
    };
    return s_column_names;
}

const string CVCFVariantsBase::sm_INFO = "INFO";
const string CVCFVariantsBase::sm_SAMPLES = "SAMPLES";

const vector<string>& CVCFVariantsBase::s_GetAllColNames()
{
    static const vector<string> s_column_names{
        sm_ID, sm_REF, sm_ALT, sm_QUAL, sm_FILTER, sm_INFO, sm_FORMAT, sm_SAMPLES
    };
    return s_column_names;
}

CVCFVariantsBase::CVCFVariantsBase(bool load_all_info, 
    const set<string>& info_fields, const map<unsigned, string>& sample_cols)
    : m_Descriptors(load_all_info, info_fields, sample_cols)
{
}

CVCFVariantsBase::~CVCFVariantsBase()
{
    // do not delete the data pointed by m_Buffer
    if (m_Buffer) {
        m_Buffer = nullptr;
    }
}

size_t CVCFVariantsBase::GetIndexForPosition(const unsigned& pos, CPosToIndex::EType type) const
{
    return m_Posindexmap.GetIndexForPosition(pos, type);
}

unsigned CVCFVariantsBase::GetPositionForIndex(const size_t& index) const
{
    return m_Posindexmap.GetPositionForIndex(index);
}

unsigned CVCFVariantsBase::GetNumberOfVariants() const
{
    size_t from = m_Posindexmap.GetIndexForPosition(m_StartPos, CPosToIndex::eMin);
    size_t to = m_Posindexmap.GetIndexForPosition(m_StopPos, CPosToIndex::eMax);
    return to - from;
}

unsigned CVCFVariantsBase::Count(const TSeqRange& range)
{
    unsigned count = 0;
    if (range.NotEmpty()) {
        const auto& indices = m_Posindexmap.GetIndexVectors();
        for (const auto& it : indices) {
            const auto non_null = it->get_null_bvector();
            if (non_null) {
                // TODO: replace it with count_range_notnull(left, right) GB-9046
                count += non_null->count_range(range.GetFrom(), range.GetTo());
            }
        }
    }
    return count;
}

bool CVCFVariantsBase::operator==(const CVCFVariantsBase& other) const
{
    return (m_StartPos == other.m_StartPos) &&
        (m_StopPos == other.m_StopPos) &&
        (m_Descriptors == other.m_Descriptors) &&
        (m_Posindexmap == other.m_Posindexmap);
}

void CVCFVariantsBase::ListPositionVectors(CNcbiOstream& out) const
{
    m_Posindexmap.ListVectors(out);
}

void CVCFVariantsBase::GetHistogram(svector& hist, const svector::size_type& bin_width) const
{
    hist.clear();
    if (bin_width == 0)
        return;

    TIndexVector::size_type from = 0;
    TIndexVector::size_type to = bin_width - 1;

    const auto& indices = m_Posindexmap.GetIndexVectors();
    vector<TIndexVector::bvector_type_const_ptr> non_null_vectors;
    for (const auto& it : indices) {
        non_null_vectors.emplace_back(it->get_null_bvector());
    }

    // the first index vector is always the longest 
    size_t max_size = indices.front()->size();
    auto bit = hist.get_back_inserter();

    do {
        unsigned count = 0;
        for (const auto& it : non_null_vectors) {
            if (from < it->size()) {
                count += it->count_range(from, to); // closed interval [from, to]
            }
        }
        from += bin_width;
        to = from + bin_width - 1;
        bit = count;

    } while (from < max_size);
    bit.flush();

    BM_DECLARE_TEMP_BLOCK(tb);
    hist.optimize(tb);
}

void CVCFVariantsBase::GetHistogram(TIndexVector& hist, const svector::size_type& bin_width)
{
    hist.clear();
    if (bin_width == 0)
        return;

    svector hist_vect(bm::use_null);

    TIndexVector::size_type from = 0;
    TIndexVector::size_type to = bin_width - 1;

    const auto& indices = m_Posindexmap.GetIndexVectors();
    vector<TIndexVector::bvector_type_const_ptr> non_null_vectors;
    for (const auto& it : indices) {
        non_null_vectors.emplace_back(it->get_null_bvector());
    }

    // the first index vector is always the longest 
    size_t max_size = indices.front()->size();
    do {
        unsigned count = 0;
        for (const auto& it : non_null_vectors) {
            if (from < it->size()) {
                count += it->count_range(from, to); // closed interval [from, to]
            }
        }

        hist_vect.set(from, count);
        from += bin_width;
        to = from + bin_width - 1;

    } while (from < max_size);

    hist.load_from(hist_vect);

    BM_DECLARE_TEMP_BLOCK(tb);
    hist.optimize(tb);
    hist.sync();
}

string CVCFVariantsBase::s_GetPreviousVersion()
{
    return kEmptyStr;
}

string CVCFVariantsBase::s_GetCurrentVersion(bool format_only)
{
    // update other code that relies on columnarVCF!!!
    string version_nr = "1.0";
    string model_name = "Index-IRAQFF-SplitInfo-SeparateSamples";
    string sc = "SC" NCBI_AS_STRING(NCBI_SC_VERSION);
    string rev = "R" NCBI_AS_STRING(NCBI_SUBVERSION_REVISION);
    vector<string> parts{ version_nr, model_name };
    if (!format_only) {
        parts.push_back(sc);
        parts.push_back(rev);
    }
    return NStr::Join(parts, "-");
}

bool CVCFVariantsBase::s_ConvertFromPreviousVersion(objects::CAbstractProjectItem& prj_item)
{
    // currently the two versions are the same
    prj_item.SetType("ColumnarVCF:" + s_GetCurrentVersion());
    return true;
}

void CVCFVariantsBase::x_SerializeData() const
{
    auto start = chrono::steady_clock::now();

    if (m_Buffer) {
        m_Buffer->clear();
    }

    vector<pair<string, chrono::duration<double>>> time_to_serialize;

    // serialize the index vectors
    using TIdxLayout = bm::sparse_vector_serial_layout<TIndexVector>;
    vector<unique_ptr<TIdxLayout>> idx_layout_vector;

    bm::sparse_vector_serializer<TIndexVector> idx_serializer;

    for (size_t index = 0; index < m_Posindexmap.m_Indices.size(); ++index) {
        auto& idx_vector = *m_Posindexmap.m_Indices[index].get();
        auto layout = make_unique<TIdxLayout>();

        auto start_serial = chrono::steady_clock::now();
        idx_serializer.serialize(idx_vector, *layout.get());
        auto diff_serial = chrono::steady_clock::now() - start_serial;
        time_to_serialize.emplace_back("Index_" + NStr::UIntToString((unsigned)index), diff_serial);

        idx_layout_vector.push_back(move(layout));
    }
    /*
    LOG_POST(Info << "Index vector layout sizes:");
    for (const auto& it : idx_layout_vector) {
        LOG_POST(Info << it->size());
    }
    */
    

    // serialize the data columns (except the INFO and SAMPLE columns)
    using TLayout = bm::sparse_vector_serial_layout<TSparseStrVector>;
    vector<unique_ptr<TLayout>> layout_vector;

    bm::sparse_vector_serializer<TSparseStrVector> str_serializer;

    str_serializer.set_bookmarks(true, 16);
    str_serializer.enable_xor_compression();
    _ASSERT(str_serializer.is_xor_ref());

    const auto& col_names = CVCFVariantsBase::s_GetColNames();
    for (size_t index = 0; index < col_names.size(); ++index) {
        auto& str_vector = m_Descriptors[col_names[index]];
        auto layout = make_unique<TLayout>();

        auto start_serial = chrono::steady_clock::now();
        str_serializer.serialize(str_vector, *layout.get());
        auto diff_serial = chrono::steady_clock::now() - start_serial;
        time_to_serialize.emplace_back(col_names[index], diff_serial);

        layout_vector.push_back(move(layout));
    }

    /*LOG_POST(Info << "Column layout sizes:");
    for (const auto& it : layout_vector) {
        LOG_POST(Info << it->size());
    }
    */

    // serialize the info fields
    using TOptLayout = bm::sparse_vector_serial_layout<TSparseOptVector>;
    unordered_map<string, unique_ptr<TOptLayout>> info_layout_vector;

    bm::sparse_vector_serializer<TSparseOptVector> opt_serializer;

    opt_serializer.set_bookmarks(true, 16);
    opt_serializer.enable_xor_compression();
    _ASSERT(opt_serializer.is_xor_ref());

    const auto info_fields = m_Descriptors.GetInfoFieldNames();
    for (const auto& it : info_fields) {
        auto& sub_col = m_Descriptors.GetInfoField(it);
        auto layout = make_unique<TOptLayout>();

        auto start_serial = chrono::steady_clock::now();
        opt_serializer.serialize(sub_col, *layout.get());
        auto diff_serial = chrono::steady_clock::now() - start_serial;
        time_to_serialize.emplace_back(it, diff_serial);

        info_layout_vector.insert(make_pair(it, move(layout)));
    }

    /*LOG_POST(Info << "Info field layout sizes:");
    for (auto& it : info_layout_vector) {
        LOG_POST(Info << it.first << ": "<< it.second->size());
    }

    {
        size_t total_info = 0;
        LOG_POST(Info << "Info field layout sizes:");
        for (auto& it : info_layout_vector) {
            LOG_POST(Info << it.first << ": " << it.second->size());
            total_info += it.second->size();
        }
        LOG_POST(Info << "Total INFO column: " << total_info);
    }
    */
    
    // serialize SAMPLES columns
    vector<pair<string, unique_ptr<TOptLayout>>> sample_layout_vector;

    using samples_serializer_type = bm::sparse_vector_serializer<TSparseOptVector>;
    samples_serializer_type sample_serializer;

    sample_serializer.set_bookmarks(true, 16);

    // create a reference vector and attach it to all serializers
    samples_serializer_type::bv_ref_vector_type bitvector_ref;

    // add references in reverse order
    const auto samples = m_Descriptors.GetSampleNames();
    for (auto rit = samples.rbegin(); rit != samples.rend(); ++rit) {
        auto& sample_col = m_Descriptors.GetSample(*rit);
        bitvector_ref.add_sparse_vector(sample_col);
    }

    // connect the reference vector to the serializer
    sample_serializer.set_xor_ref(&bitvector_ref);


    // compute XOR similarity matrix in parallel
    bm::compute_sim_matrix_plan_builder<bm::bvector<>> pbuilder;
    bm::compute_sim_matrix_plan_builder<bm::bvector<>>::task_batch tbatch;
    bm::xor_sim_model<bm::bvector<>> sim_model;
    bm::xor_sim_params  xor_search_params;
    pbuilder.build_plan(tbatch, sim_model, bitvector_ref, xor_search_params);

    typedef bm::thread_pool<bm::task_description*, bm::spin_lock<bm::pad0_struct> > pool_type;
    pool_type tpool; // our thread pool here (no threads created yet)
    
    unsigned thread_count = thread::hardware_concurrency();
    if (thread_count == 0)
        thread_count = 1;
    
    tpool.start(thread_count); // start the threads
    {
        bm::thread_pool_executor<pool_type> exec;
        exec.run(tpool, tbatch, true);
    }
    tpool.set_stop_mode(pool_type::stop_when_done);
    tpool.join();


    // add similarity model to each serializer
    sample_serializer.set_sim_model(&sim_model);

    // then: serialize
    for (const auto& it : samples) {
        auto& sample_col = m_Descriptors.GetSample(it);
        auto layout = make_unique<TOptLayout>();

        auto start_serial = chrono::steady_clock::now();
        sample_serializer.serialize(sample_col, *layout.get());
        auto diff_serial = chrono::steady_clock::now() - start_serial;
        time_to_serialize.emplace_back(it, diff_serial);

        sample_layout_vector.emplace_back(make_pair(it, move(layout)));
    }

    sample_serializer.set_xor_ref(nullptr);

    /*
    {
        size_t total_samples = 0;
        for (auto& it : sample_layout_vector) {
            total_samples += it.second->size();
        }
        LOG_POST(Info << "Total SAMPLES columns: " << total_samples);
    }
    */

    size_t total_size = 0;
    total_size += (sizeof(size_t) * idx_layout_vector.size());
    for (const auto& it : idx_layout_vector) {
        total_size += it->size();
    }
    
    total_size += (sizeof(size_t) * layout_vector.size());
    for (const auto& it : layout_vector) {
        total_size += it->size();
    }

    total_size += (sizeof(size_t) * 2); // for saving two more numbers

    total_size += sizeof(size_t); // number of info columns
    total_size += (sizeof(size_t) * info_layout_vector.size());
    for (const auto& it : info_layout_vector) {
        total_size += it.second->size();
        total_size += it.first.size(); // the string
        total_size += sizeof(size_t);  // the size of the string
    }

    total_size += sizeof(size_t); // number of sample columns
    total_size += (sizeof(size_t) * sample_layout_vector.size());
    for (const auto& it : sample_layout_vector) {
        total_size += it.second->size();
        total_size += it.first.size(); // the name
        total_size += sizeof(size_t); // the size of the string
    }
    LOG_POST(Info << "Total serialized data size for " << GetNumberOfVariants() << " variants: " << total_size);

    m_Buffer = new vector<char>(total_size);
    char* buf_ptr = m_Buffer->data();

    // save the number of index vectors
    auto idx_size = idx_layout_vector.size();
    memcpy(buf_ptr, &idx_size, sizeof(idx_size));
    buf_ptr += sizeof(idx_size);

    // save the number of string columns
    auto col_size = layout_vector.size();
    memcpy(buf_ptr, &col_size, sizeof(col_size));
    buf_ptr += sizeof(col_size);

    for (const auto& it : idx_layout_vector) {
        // save the size of the column
        auto layout_size = it->size();
        memcpy(buf_ptr, &layout_size, sizeof(layout_size));
        buf_ptr += sizeof(layout_size);

        // save the column
        memcpy(buf_ptr, it->data(), layout_size);
        buf_ptr += layout_size;
    }

    for (const auto& it : layout_vector) {
        // save the size of the column
        auto layout_size = it->size();
        memcpy(buf_ptr, &layout_size, sizeof(layout_size));
        buf_ptr += sizeof(layout_size);

        // save the column
        memcpy(buf_ptr, it->data(), layout_size);
        buf_ptr += layout_size;
    }

    // save the number of info fields
    auto info_col_size = info_layout_vector.size();
    memcpy(buf_ptr, &info_col_size, sizeof(info_col_size));
    buf_ptr += sizeof(info_col_size);

    for (const auto& it : info_layout_vector) {
        // save the size of the field name
        size_t name_size = it.first.size();
        memcpy(buf_ptr, &name_size, sizeof(name_size));
        buf_ptr += sizeof(name_size);

        // save the name of the field
        memcpy(buf_ptr, it.first.data(), name_size);
        buf_ptr += name_size;

        // save the size of the field
        auto layout_size = it.second->size();
        memcpy(buf_ptr, &layout_size, sizeof(layout_size));
        buf_ptr += sizeof(layout_size);

        // save the data
        memcpy(buf_ptr, it.second->data(), layout_size);
        buf_ptr += layout_size;
    }

    // save the samples
    auto samples_size = sample_layout_vector.size();
    memcpy(buf_ptr, &samples_size, sizeof(samples_size));
    buf_ptr += sizeof(samples_size);

    for (const auto& it : sample_layout_vector) {
        // save the size of the sample column's name
        size_t name_size = it.first.size();
        memcpy(buf_ptr, &name_size, sizeof(name_size));
        buf_ptr += sizeof(name_size);

        // save the sample column's name
        memcpy(buf_ptr, it.first.data(), name_size);
        buf_ptr += name_size;

        // save the size of the sample data column
        auto layout_size = it.second->size();
        memcpy(buf_ptr, &layout_size, sizeof(layout_size));
        buf_ptr += sizeof(layout_size);

        // save the data
        memcpy(buf_ptr, it.second->data(), layout_size);
        buf_ptr += layout_size;
    }

    auto diff = chrono::steady_clock::now() - start;
    LOG_POST(Info << "Serialized data into a single blob in "
        << chrono::duration_cast<chrono::milliseconds>(diff).count() << " ms ");

    /*
    LOG_POST(Info << "Individual serialization times:");
    for (const auto& it : time_to_serialize) {
        LOG_POST(Info << it.first << ": "
            << chrono::duration_cast<chrono::milliseconds>(it.second).count() << "ms");
    }
    */
}

void CVCFVariantsBase::x_DeserializeAllData()
{
    auto start = chrono::steady_clock::now();

    if (m_Buffer) {
        const unsigned char* buf_ptr = (unsigned char*)m_Buffer->data();
        size_t nr_cols = 0;
        x_DeserializeIndexVectors(buf_ptr, nr_cols);
        x_DeserializeDescr_All(buf_ptr, nr_cols);
    }

    auto diff = chrono::steady_clock::now() - start;
    LOG_POST(Info << "Time to deserialize all data: "
        << chrono::duration_cast<chrono::milliseconds>(diff).count() << " ms ");
}

void CVCFVariantsBase::x_DeserializeIndexVectors(const unsigned char* &buf_ptr, size_t& nr_cols)
{
//    auto start = chrono::steady_clock::now();

    if (buf_ptr == nullptr) {
        NCBI_THROW(CVCFDataException, eDeserializationFailed, 
            "Deserialization of index vectors has failed as data buffer is empty");
    }

    size_t nr_idx = 0;
    memcpy(&nr_idx, buf_ptr, sizeof(size_t));
    buf_ptr += sizeof(size_t);

    nr_cols = 0;
    memcpy(&nr_cols, buf_ptr, sizeof(size_t));
    buf_ptr += sizeof(size_t);

    /*
    LOG_POST(Info << "Number of index vectors to deserialize: " << nr_idx);
    LOG_POST(Info << "Number of columns to deserialize: " << nr_cols);
    */
    vector<pair<string, chrono::duration<double>>> time_to_deserialize;

    // deserialize index vectors
    vector<size_t> idx_layout_sizes;
    bm::sparse_vector_deserializer<TIndexVector> idx_deserializer;

    if (nr_idx > 1) {
        for (size_t i = 1; i < nr_idx; ++i) {
            unique_ptr<TIndexVector> new_vect = make_unique<TIndexVector>(bm::use_null);
            m_Posindexmap.m_Indices.push_back(move(new_vect));
        }
    }

    auto idx_vect_iter = m_Posindexmap.m_Indices.begin();
    auto idx_vect_end = m_Posindexmap.m_Indices.end();
    for (size_t i = 0; i < nr_idx && idx_vect_iter != idx_vect_end; ++i) {
        size_t index_vec_size = 0;
        memcpy(&index_vec_size, buf_ptr, sizeof(size_t));
        buf_ptr += sizeof(size_t);

        auto start_serial = chrono::steady_clock::now();
        idx_deserializer.deserialize(*(*idx_vect_iter).get(), buf_ptr);
        (*idx_vect_iter)->sync(true);
        auto diff_serial = chrono::steady_clock::now() - start_serial;
        time_to_deserialize.emplace_back("Index_" + NStr::UIntToString((unsigned)i), diff_serial);

        buf_ptr += index_vec_size;
        idx_layout_sizes.push_back(index_vec_size);
        ++idx_vect_iter;
    }
    /*
    LOG_POST(Info << "Index vector sizes to deserialize:");
    for (const auto& it : idx_layout_sizes) {
        LOG_POST(Info << it);
    }
    */

    // set internal member
    m_Posindexmap.x_SaveMaximumIndex();
    /*
    auto diff = chrono::steady_clock::now() - start;
    LOG_POST(Info << "Time to deserialize index vectors: "
        << chrono::duration_cast<chrono::milliseconds>(diff).count() << " ms ");

    LOG_POST(Info << "Individual deserialization times:");
    for (const auto& it : time_to_deserialize) {
        LOG_POST(Info << it.first << ": "
            << chrono::duration_cast<chrono::milliseconds>(it.second).count() << "ms");
    }
    */
}

void CVCFVariantsBase::x_DeserializeDescr_All(const unsigned char* buf_ptr, const size_t& nr_cols)
{
    if (buf_ptr == nullptr) {
        NCBI_THROW(CVCFDataException, eDeserializationFailed, 
            "Deserialization of data has failed as data buffer is empty");
    }

    vector<pair<string, chrono::duration<double>>> time_to_deserialize;

    // deserialize data columns (except the INFO and SAMPLE columns)
    vector<size_t> layout_sizes;
    
    const auto& col_names = CVCFVariantsBase::s_GetColNames();
    auto col_it = col_names.begin();
    _ASSERT(col_names.size() == nr_cols);

    for (size_t i = 0; i < nr_cols && col_it != col_names.end(); ++i) {
        size_t col_size = 0;
        memcpy(&col_size, buf_ptr, sizeof(size_t));
        buf_ptr += sizeof(size_t);

        auto& col_data = m_Descriptors[*col_it];
        auto start_serial = chrono::steady_clock::now();
        m_Deserializer.deserialize(col_data, buf_ptr);
        
        auto diff_serial = chrono::steady_clock::now() - start_serial;
        time_to_deserialize.emplace_back(*col_it, diff_serial);

        buf_ptr += col_size;
        layout_sizes.push_back(col_size);
        ++col_it;
    }

    /*
    LOG_POST(Info << "Layout sizes to deserialize:");
    for (auto& it : layout_sizes) {
        LOG_POST(Info << it);
    }
    */

    auto var_it = m_Descriptors[CVCFVariantsBase::sm_ID].begin();
    m_StartPos = m_Posindexmap.GetPositionForIndex(var_it.pos());
    m_StopPos = m_Posindexmap.GetPositionForIndex(m_Posindexmap.GetMaxIndex());
    _ASSERT(m_StartPos <= m_StopPos);

    // deserialize info fields
    size_t info_cols = 0;
    memcpy(&info_cols, buf_ptr, sizeof(size_t));
    buf_ptr += sizeof(size_t);

    //LOG_POST(Info << "Number of info fields to deserialize: " << info_cols);

    unordered_map<string, size_t> info_layout_sizes;
    m_Descriptors.SetNumberofInfoFields(info_cols);
        
    for (size_t i = 0; i < info_cols; ++i) {
        size_t name_size = 0;
        memcpy(&name_size, buf_ptr, sizeof(size_t));
        buf_ptr += sizeof(size_t);

        char* str_ptr = (char*)buf_ptr;
        string info_name(str_ptr, name_size);
        buf_ptr += name_size;

        size_t info_size = 0;
        memcpy(&info_size, buf_ptr, sizeof(size_t));
        buf_ptr += sizeof(size_t);

        auto& info_data = m_Descriptors.SetInfoField(info_name);
        auto start_serial = chrono::steady_clock::now();
        m_OptDeserializer.deserialize(info_data, buf_ptr);

        auto diff_serial = chrono::steady_clock::now() - start_serial;
        time_to_deserialize.emplace_back(info_name, diff_serial);

        buf_ptr += info_size;
        info_layout_sizes.emplace(info_name, info_size);
    }
    /*
    LOG_POST(Info << "Layout sizes of info fields to deserialize:");
    for (const auto& it : info_layout_sizes) {
        LOG_POST(Info << it.first << "  " << it.second);
    }
    */

    // deserialize SAMPLE columns

    // reference vector for XOR deserialization
    bm::sparse_vector_deserializer<TSparseOptVector> sample_deserializer;
    bm::sparse_vector_deserializer<TSparseOptVector>::bv_ref_vector_type  bitvector_ref;

    // first pass: build reference vectors

    size_t sample_cols = 0;
    memcpy(&sample_cols, buf_ptr, sizeof(size_t));
    buf_ptr += sizeof(size_t);

    //LOG_POST(Info << "Number of SAMPLE columns to deserialize: " << sample_cols);

    m_Descriptors.SetNumberofSampleCols(sample_cols);

    const unsigned char* tmp_ptr = buf_ptr;
    for (size_t i = 0; i < sample_cols; ++i) {
        size_t name_size = 0;
        memcpy(&name_size, tmp_ptr, sizeof(size_t));
        tmp_ptr += sizeof(size_t);

        char* str_ptr = (char*)tmp_ptr;
        string sample_name(str_ptr, name_size);
        tmp_ptr += name_size;

        size_t sample_size = 0;
        memcpy(&sample_size, tmp_ptr, sizeof(size_t));
        tmp_ptr += sizeof(size_t);

        auto& sample_data = m_Descriptors.SetSample(sample_name);
        sample_deserializer.deserialize_structure(sample_data, tmp_ptr);

        tmp_ptr += sample_size;
        _ASSERT(sample_data.empty());
    }

    // construct the reference vector
    // Add references in reverse order
    const auto samples = m_Descriptors.GetSampleNames();
    for (auto rit = samples.rbegin(); rit != samples.rend(); ++rit) {
        auto& sample_col = m_Descriptors.GetSample(*rit);
        bitvector_ref.add_vectors(sample_col.get_bmatrix());
    }

    sample_deserializer.set_xor_ref(&bitvector_ref);

    // second pass: data deserialization
    unordered_map<string, size_t> sample_layout_sizes;

    for (size_t i = 0; i < sample_cols; ++i) {
        size_t name_size = 0;
        memcpy(&name_size, buf_ptr, sizeof(size_t));
        buf_ptr += sizeof(size_t);

        char* str_ptr = (char*)buf_ptr;
        string sample_name(str_ptr, name_size);
        buf_ptr += name_size;

        size_t sample_size = 0;
        memcpy(&sample_size, buf_ptr, sizeof(size_t));
        buf_ptr += sizeof(size_t);

        auto& sample_data = m_Descriptors.SetSample(sample_name);
        auto start_serial = chrono::steady_clock::now();
        sample_deserializer.deserialize(sample_data, buf_ptr, false);

        auto diff_serial = chrono::steady_clock::now() - start_serial;
        time_to_deserialize.emplace_back(sample_name, diff_serial);

        buf_ptr += sample_size;
        sample_layout_sizes.emplace(sample_name, sample_size);
    }

    sample_deserializer.set_xor_ref(nullptr);
}

void CVCFVariantsBase::x_DeserializeColumn(const string& col_name, const unsigned char* buf_ptr, const size_t& nr_cols)
{
    if (buf_ptr == nullptr) {
        NCBI_THROW(CVCFDataException, eDeserializationFailed,
            "Deserialization of " + col_name + " vector has failed as data buffer is empty");
    }

    if (col_name.empty()) {
        LOG_POST(Error << "x_DeserializeColumn() failed: empty column name");
        return;
    }

    const auto& col_names = CVCFVariantsBase::s_GetColNames();
    auto col_it = col_names.begin();
    _ASSERT(col_names.size() == nr_cols); //is that always true?

    for (size_t i = 0; i < nr_cols && col_it != col_names.end(); ++i) {
        size_t col_size = 0;
        memcpy(&col_size, buf_ptr, sizeof(size_t));
        buf_ptr += sizeof(size_t);

        if (NStr::EqualNocase(col_name, *col_it)) {
            auto& col_data = m_Descriptors[*col_it];
            m_Deserializer.deserialize(col_data, buf_ptr);
            break;
        }

        buf_ptr += col_size;
        ++col_it;
    }
}

void CVCFVariantsBase::x_DeserializeDescr_Range(const unsigned char* buf_ptr, const size_t& nr_cols, 
    const TSeqRange* range, const set<string>& cols_to_decode)
{
    if (buf_ptr == nullptr) {
        NCBI_THROW(CVCFDataException, eDeserializationFailed,
            "Deserialization of a data range has failed as data buffer is empty");
    }

    auto start = chrono::steady_clock::now();
    if (!range || range->Empty() || cols_to_decode.empty()) {
        m_StartPos = 0;
        m_StopPos = 0;
        return;
    }

    size_t from_index = 0;
    size_t to_index = 0;
    bool deserialize_all = false;

    //map the range onto index coordinates [from_index, to_index)
    from_index = m_Posindexmap.GetIndexForPosition(range->GetFrom(), CPosToIndex::eMin);
    to_index = m_Posindexmap.GetIndexForPosition(range->GetTo(), CPosToIndex::eMax);
    if (from_index == 0 && to_index == m_Posindexmap.m_MaxStoredIndex + 1) {
        deserialize_all = true;
    }
    // from_index == to_index -> effectively an empty range

    vector<pair<string, chrono::duration<double>>> time_to_deserialize;
    if (from_index < to_index) {
        // deserialize data columns (except INFO and SAMPLES columns)
        vector<size_t> layout_sizes;

        const auto& col_names = CVCFVariantsBase::s_GetColNames();
        auto col_it = col_names.begin();
        _ASSERT(col_names.size() == nr_cols); //is that always true?

        for (size_t i = 0; i < nr_cols && col_it != col_names.end(); ++i, ++col_it) {
            size_t col_size = 0;
            memcpy(&col_size, buf_ptr, sizeof(size_t));
            buf_ptr += sizeof(size_t);

            if (cols_to_decode.find(*col_it) != cols_to_decode.end()) {
                auto& col_data = m_Descriptors[*col_it];
                auto start_serial = chrono::steady_clock::now();
                if (deserialize_all) {
                    m_Deserializer.deserialize(col_data, buf_ptr);
                }
                else {
                    m_Deserializer.deserialize(col_data, buf_ptr, from_index, to_index - 1);
                }
                auto diff_serial = chrono::steady_clock::now() - start_serial;
                time_to_deserialize.emplace_back(*col_it, diff_serial);
                layout_sizes.push_back(col_size);
            }

            buf_ptr += col_size;
        }

        /*
        LOG_POST(Info << "Layout sizes to deserialize:");
        for (auto& it : layout_sizes) {
            LOG_POST(Info << it << ", ");
        }
        */

        // deserialize info fields
        size_t info_cols = 0;
        memcpy(&info_cols, buf_ptr, sizeof(size_t));
        buf_ptr += sizeof(size_t);

        /*
        LOG_POST(Info << "Number of info fields to deserialize: " << info_cols);
        */
        bool need_all_info_fields = (cols_to_decode.find(CVCFVariantsBase::sm_INFO) != cols_to_decode.end());
        
        unordered_map<string, size_t> info_layout_sizes;
        if (need_all_info_fields) {
            m_Descriptors.SetNumberofInfoFields(info_cols);
        }

        for (size_t i = 0; i < info_cols; ++i) {
            size_t name_size = 0;
            memcpy(&name_size, buf_ptr, sizeof(size_t));
            buf_ptr += sizeof(size_t);

            char* str_ptr = (char*)buf_ptr;
            string info_name(str_ptr, name_size);
            buf_ptr += name_size;

            size_t info_size = 0;
            memcpy(&info_size, buf_ptr, sizeof(size_t));
            buf_ptr += sizeof(size_t);

            if (need_all_info_fields || cols_to_decode.find(info_name) != cols_to_decode.end()) {
                auto& info_data = m_Descriptors.SetInfoField(info_name);
                auto start_serial = chrono::steady_clock::now();
                if (deserialize_all) {
                    m_OptDeserializer.deserialize(info_data, buf_ptr);
                }
                else {
                    m_OptDeserializer.deserialize(info_data, buf_ptr, from_index, to_index - 1);
                }

                auto diff_serial = chrono::steady_clock::now() - start_serial;
                time_to_deserialize.emplace_back(info_name, diff_serial);
                info_layout_sizes.emplace(info_name, info_size);
            }
            buf_ptr += info_size;
        }

        /*LOG_POST(Info << "Layout sizes of info fields to deserialize:");
        for (const auto& it : info_layout_sizes) {
            LOG_POST(Info << it.first << "  " << it.second);
        }
        */
        

        if (cols_to_decode.find(CVCFVariantsBase::sm_SAMPLES) != cols_to_decode.end()) {
            // deserialize SAMPLEs columns
            // reference vector for XOR deserialization
            bm::sparse_vector_deserializer<TSparseOptVector> sample_deserializer;
            bm::sparse_vector_deserializer<TSparseOptVector>::bv_ref_vector_type  bitvector_ref;

            // first pass: build reference vectors

            size_t sample_cols = 0;
            memcpy(&sample_cols, buf_ptr, sizeof(size_t));
            buf_ptr += sizeof(size_t);

            m_Descriptors.SetNumberofSampleCols(sample_cols);

            const unsigned char* tmp_ptr = buf_ptr;
            for (size_t i = 0; i < sample_cols; ++i) {
                size_t name_size = 0;
                memcpy(&name_size, tmp_ptr, sizeof(size_t));
                tmp_ptr += sizeof(size_t);

                char* str_ptr = (char*)tmp_ptr;
                string sample_name(str_ptr, name_size);
                tmp_ptr += name_size;

                size_t sample_size = 0;
                memcpy(&sample_size, tmp_ptr, sizeof(size_t));
                tmp_ptr += sizeof(size_t);

                auto& sample_data = m_Descriptors.SetSample(sample_name);
                sample_deserializer.deserialize_structure(sample_data, tmp_ptr);

                tmp_ptr += sample_size;
                _ASSERT(sample_data.empty());
            }

            // construct the reference vector
            // Add references in reverse order
            const auto samples = m_Descriptors.GetSampleNames();
            for (auto rit = samples.rbegin(); rit != samples.rend(); ++rit) {
                auto& sample_col = m_Descriptors.GetSample(*rit);
                bitvector_ref.add_vectors(sample_col.get_bmatrix());
            }

            sample_deserializer.set_xor_ref(&bitvector_ref);

            // second pass: data deserialization
            unordered_map<string, size_t> sample_layout_sizes;

            for (size_t i = 0; i < sample_cols; ++i) {
                size_t name_size = 0;
                memcpy(&name_size, buf_ptr, sizeof(size_t));
                buf_ptr += sizeof(size_t);

                char* str_ptr = (char*)buf_ptr;
                string sample_name(str_ptr, name_size);
                buf_ptr += name_size;

                size_t sample_size = 0;
                memcpy(&sample_size, buf_ptr, sizeof(size_t));
                buf_ptr += sizeof(size_t);

                auto& sample_data = m_Descriptors.SetSample(sample_name);
                auto start_serial = chrono::steady_clock::now();
                if (deserialize_all) {
                    sample_deserializer.deserialize(sample_data, buf_ptr, false);
                }
                else {
                    sample_deserializer.deserialize_range(sample_data, buf_ptr, from_index, to_index - 1, false);
                }

                auto diff_serial = chrono::steady_clock::now() - start_serial;
                time_to_deserialize.emplace_back(sample_name, diff_serial);

                buf_ptr += sample_size;
                sample_layout_sizes.emplace(sample_name, sample_size);
            }

            sample_deserializer.set_xor_ref(nullptr);
        }
    }

    /*
    auto diff = chrono::steady_clock::now() - start;
    LOG_POST(Info << "Deserialization time for a given range took: "
        << chrono::duration_cast<chrono::milliseconds>(diff).count() << " ms ");

    LOG_POST(Info << "Individual deserialization times:");
    for (const auto& it : time_to_deserialize) {
        LOG_POST(Info << it.first << ": "
            << chrono::duration_cast<chrono::milliseconds>(it.second).count() << "ms");
    }
    */

    m_StartPos = range->GetFrom();
    m_StopPos = range->GetTo();

    _ASSERT(m_StartPos <= m_StopPos);
}


END_NCBI_SCOPE

