/*  $Id: demo_read_large_vcf.cpp 46001 2021-01-21 12:55:04Z grichenk $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrea Asztalos
 *
 * File Description:
 *    Demo application for reading large (> 100MB) VCF files
 */


#include <ncbi_pch.hpp>

#include <corelib/ncbiapp.hpp>
#include <corelib/ncbireg.hpp>
#include <corelib/ncbi_limits.hpp>
#include <util/compress/stream_util.hpp>
#include <gui/widgets/loaders/columnar_vcf_variants.hpp>
#include <gui/widgets/loaders/columnar_vcf_reader.hpp>
#include <chrono>

USING_NCBI_SCOPE;

class CColumnarVCFReaderApp : public CNcbiApplication
{
public:
    virtual void Init(void);
    virtual int  Run(void);
    virtual void Exit(void);
};

void CColumnarVCFReaderApp::Init(void)
{
    unique_ptr<CArgDescriptions> arg_desc(new CArgDescriptions);
    arg_desc->SetUsageContext(GetArguments().GetProgramBasename(),
        "Read large VCF files");

    arg_desc->AddOptionalKey("i", "InFile", "Input VCF file", CArgDescriptions::eInputFile);

    arg_desc->AddOptionalKey("iserial", "InFile", "Input File containing the serialized data", CArgDescriptions::eInputFile);

    arg_desc->AddOptionalKey("o", "OutFile", "Single Output File", CArgDescriptions::eOutputFile);

    arg_desc->AddFlag("sv_cols", "Output only the first five columns from the data section of the input file");

    // for serializing the entire list of variants
    arg_desc->AddOptionalKey("serial", "OutputFile", "Output file for serialized data", CArgDescriptions::eOutputFile);
    
    // to deserialize and compare
    arg_desc->AddFlag("deserialize_and_compare", "Deserialize data and compare it with the existing variant data");
    
    // for serializing individual columns
    arg_desc->AddOptionalKey("serialize_cols", "OutputFile", "Output file with logging serialization and deserialization", CArgDescriptions::eOutputFile);

    arg_desc->AddOptionalKey("index", "OutFile", "Output file for listing index vectors", CArgDescriptions::eOutputFile);

    arg_desc->AddOptionalKey("stats", "OutputFile", "Output File for listing statistics", CArgDescriptions::eOutputFile);

    arg_desc->AddFlag("hist", "Compute histogram");

    SetupArgDescriptions(arg_desc.release());
}

// To serialize the data into one file:
// demo_read_large_vcf -i bbb.gz -serial serial_bbb.bin -o output_bbb.txt -sv_cols

// To read the serialized data from a file:
// demo_read_large_vcf -iserial serial_bbb.bin -o output_bbb.txt -sv_cols

// demo_read_large_vcf -i bbb.gz -serial serial_bbb.bin -deserialize_and_compare

int CColumnarVCFReaderApp::Run(void)
{
    const CArgs& args = GetArgs();
    if (args["i"]) {
        const string& fname = args["i"].AsString();

        CNcbiIstream* istr = nullptr;
        if (NStr::EndsWith(fname, ".gz", NStr::eNocase)) {
            CDecompressIStream* decompressIstr =
                new CDecompressIStream(args["i"].AsInputFile(), CDecompressIStream::eGZipFile);
            istr = decompressIstr;
        }
        else {
            istr = &args["i"].AsInputFile();
        }

        if (!istr) {
            cerr << "Error: Could not open file " + fname + "\n";
            return false;
        }

        cout << "Starting to read file: " << fname << endl;
        CColumnarVCFReaderTest reader;
        try {
            reader.ReadData(*istr);
        }
        catch (const CException& e) {
            cout << "Parsing the file has failed: " << e.GetMsg() << endl;
            return 1;
        }

        cout << "VCF file was successfully loaded into memory" << endl;
        
        vector<string> chr_names = reader.GetChromosomeNames();
        for (const auto& it : chr_names) {
            auto var = reader.GetVariantsForChr(it);
            vector<unsigned> positions;
            var->GetPositionsForMissingVarID(positions);
            if (!positions.empty()) {
                cout << "Positions where missing IDs appear:" << endl;
                for (const auto& pos_it : positions) {
                    cout << pos_it << ", ";
                }
                cout << endl;
            }
            cout << it << "\t" << var->GetNumberOfVariants() << " variants\n";
        }

        // FOR NOW: only for files that contain a single chromosome
        if (args["serial"]) {
            const string& serial_fname = args["serial"].AsString();
            vector<string> chr_names = reader.GetChromosomeNames();
            
            if (chr_names.size() == 1) {
                auto first_var = reader.GetVariantsForChr(chr_names.front());
                string chr_name = first_var->GetChrName();
                cout << "Starting to serialize data for chr " << chr_name << endl;

                first_var->WriteSerializedData(serial_fname);

                if (args["deserialize_and_compare"]) {
                    CRef<CVCFVariantList> new_variant(new CVCFVariantList(chr_name, serial_fname));
                    
                    if (*new_variant == *first_var) {
                        cout << "The two are the same" << endl;
                    }
                    else {
                        cout << "ERROR: The two are not the same " << endl;
                        return 5;
                    }
                }
            }
        }
        if (args["serialize_cols"]) {
            reader.SerializeToDisk(fname, &args["serialize_cols"].AsOutputFile());
            reader.Deserialize(fname, &args["serialize_cols"].AsOutputFile());
            cout << "Finished serializing data and deserialization was successful" << endl;
        }

        if (args["o"]) {
            try {
                reader.ListColumns(args["o"].AsOutputFile(), args["sv_cols"]);
                cout << "Finished listing columns" << endl;
            }
            catch (const CException& e) {
                cout << "Listing data columns has failed" << endl;
                return 2;
            }
        }

        if (args["index"]) {
            try {
                reader.ListIndexVectors(args["index"].AsOutputFile());
                cout << "Finished listing index vectors " << endl;
            }
            catch (const CException& e) {
                cout << "Listing index vectors has failed" << endl;
                return 3;
            }
        }

        if (args["stats"]) {
            reader.GetStatistics(args["stats"].AsOutputFile());
            cout << "Finished printing statistics" << endl;
            try {
            }
            catch (const CException& e) {
                cout << "Printing statistics of vectors has failed" << endl;
                return 4;
            }
        }
    }
    else if (args["iserial"]) {
        CRef<CVCFVariantList> new_variant(new CVCFVariantList("chr1", args["iserial"].AsString()));
        cout << "Number of variants:" << new_variant->GetNumberOfVariants() << endl;

        /*
        vector<pair<string, unsigned>> copies;
        bool unique = new_variant->AreVariantIdsUnique(copies);
        cout << "Variant ids are unique:" << NStr::BoolToString(unique) << endl;
        for (const auto& it : copies) {
            cout << it.first << " appears " << it.second << " times " << endl;
        }
        */

        CVCFVariantsBase::TIndexVector hist;
        for (size_t bin = 0; bin < 250000000; bin += 100000) {
            auto start = chrono::steady_clock::now();
            new_variant->GetHistogram(hist, bin);
            auto diff = chrono::steady_clock::now() - start;
            cout << "Hist bin: " << bin << "time to compute : "
                << chrono::duration_cast<chrono::milliseconds>(diff).count() << " ms " << endl;
        }
    }

    return 0;
}

void CColumnarVCFReaderApp::Exit(void)
{
    SetDiagStream(0);
#ifdef NCBI_OS_MSWIN
    cout << "Press any key to end the demo app" << endl;
    getchar();
#endif
}


/////////////////////////////////////////////////////////////////////////////
//  MAIN
int NcbiSys_main(int argc, ncbi::TXChar* argv[])
{
    // Execute main application function
    return CColumnarVCFReaderApp().AppMain(argc, argv);
}
