/*  $Id: test_columnar_vcf.cpp 45881 2021-01-05 03:10:02Z rudnev $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Author:  Andrea Asztalos
*
* File Description:
* 
*/

#include <ncbi_pch.hpp>
#include <corelib/ncbifile.hpp>
#include <objects/seqloc/Seq_id.hpp>
#include <gui/widgets/loaders/columnar_vcf_iterator.hpp>
#include <gui/widgets/loaders/columnar_vcf_variants.hpp>
#include <gui/widgets/loaders/columnar_vcf_reader.hpp>
// This header must be included before all Boost.Test headers if there are any
#include <corelib/test_boost.hpp>
#include <util/bitset/bmsparsevec_algo.h>


USING_NCBI_SCOPE;

NCBITEST_AUTO_INIT()
{
}

NCBITEST_AUTO_FINI()
{
    getchar();
}

void TestNonInfoColumns(CVCFVariantList& var_list)
{
    BOOST_CHECK(var_list.GetStart() == 12877850);
    BOOST_CHECK(var_list.GetStop() == 12877850);

    BOOST_CHECK(var_list.GetNumberOfVariants() == 1);
    BOOST_CHECK(var_list.GetChrName() == "1");
    BOOST_CHECK(var_list.GetMaxIndex() == 0);

    vector<string> values{ "nssv16159936", "C", "<DEL>", "", "","GT:GQ:DP:HQ" };

    const auto& col_names = CVCFVariantsBase::s_GetColNames();
    const auto& descr = var_list.GetDescriptors();
    auto col_it = col_names.begin();
    auto values_it = values.begin();
    BOOST_CHECK(values.size() == col_names.size());

    for (; col_it != col_names.end(); ++col_it, ++values_it) {
        auto& col_vector = descr[*col_it];
        BOOST_CHECK(col_vector.size() == 1);
        auto col_it = col_vector.begin();
        BOOST_CHECK(*col_it == *values_it); // NULL values are effectively empty values
    }

    col_it = col_names.begin();
    auto var_id_iter = descr[*col_it].get_const_iterator(0);
    vector<unsigned> positions;
    BOOST_CHECK(var_list.GetPositionsForVariant(*var_id_iter, positions) == true);
    BOOST_CHECK(positions.size() == 1);
    BOOST_CHECK(positions[0] == 12877850);

    auto samples = descr.GetSampleNames();
    vector<string> test_samples{ "NA1", "NA2", "NA3" };
    BOOST_CHECK(samples == test_samples);
    auto s1_it = descr.GetSample("NA1").begin();
    BOOST_CHECK(string(*s1_it) == "0|0:48:1:51,51");
}

BOOST_AUTO_TEST_CASE(Test_ParseLine)
{
    cout << "\n**************Test_ParseLine************\n" << endl;
    string line = "1	12877850	nssv16159936	C	<DEL>	.	.	DBVARID;SVTYPE=DEL;END=12933342;SVLEN=-55493;OTH_AN=190	GT:GQ:DP:HQ	0|0:48:1:51,51	1|0:48:8:51,51	1/1:43:5:.,.";
    map<unsigned, string> sample_cols{ {0, "NA1"}, {1, "NA2"}, {2, "NA3"} };
    {
        // loading each info field
        bool load_all_info = true;
        CVCFVariantList var_list("1", load_all_info, set<string>(), sample_cols);
        var_list.ParseLine(line);
        var_list.FinalizeReading();

        TestNonInfoColumns(var_list);

        const auto& descr = var_list.GetDescriptors();
        vector<string> info_fields{ "DBVARID", "SVTYPE", "END", "SVLEN", "OTH_AN" };
        BOOST_CHECK(info_fields.size() == descr.GetInfoFieldNames().size());

        vector<string> info_values{ kEmptyStr, "DEL", "12933342", "-55493", "190" };
        size_t index = 0;
        for (const auto& info_it : info_fields) {
            const auto& info_vector = descr.GetInfoField(info_it);
            BOOST_CHECK(info_vector.size() == 1);
            auto col_it = info_vector.begin();
            BOOST_CHECK(*col_it == info_values[index]);
            ++index;
        }
    }

    {
        // load none of the info fields
        bool load_all_info = false;
        CVCFVariantList var_list("1", load_all_info, set<string>(), sample_cols);
        var_list.ParseLine(line);
        var_list.FinalizeReading();

        TestNonInfoColumns(var_list);

        const auto& descr = var_list.GetDescriptors();
        BOOST_CHECK(descr.GetInfoFieldNames().empty());
    }

    {
        // load some of the info fields
        bool load_all_info = false;
        set<string> info_fields{ "SVTYPE", "END", "SVLEN" };
        CVCFVariantList var_list("1", load_all_info, info_fields, sample_cols);
        var_list.ParseLine(line);
        var_list.FinalizeReading();

        TestNonInfoColumns(var_list);

        const auto& descr = var_list.GetDescriptors();
        BOOST_CHECK(descr.GetInfoFieldNames().size() == 3);
        auto svtype_vec = descr.GetInfoField("SVTYPE");
        BOOST_CHECK(NStr::EqualCase(svtype_vec.begin().value(), "DEL"));

        svtype_vec = descr.GetInfoField("END");
        BOOST_CHECK(NStr::EqualCase(svtype_vec.begin().value(), "12933342"));

        svtype_vec = descr.GetInfoField("SVLEN");
        BOOST_CHECK(NStr::EqualCase(svtype_vec.begin().value(), "-55493"));
    }
}

BOOST_AUTO_TEST_CASE(Test_ParseLine_ReadAllSamples)
{
    cout << "\n**************Test_ParseLine_ReadAllSamples************\n" << endl;
    string line = "1	12877850	nssv16159936;.;rs12222	C	<DEL>	.	.	DBVARID;SVTYPE=DEL;END=12933342;SVLEN=-55493;OTH_AN=190	GT:GQ:DP:HQ	0|0:48:1:51,51	1|0:48:8:51,51	1/1:43:5:.,.";
    
    map<unsigned, string> sample_cols{ {0, "NA1"}, {1, "NA2"}, {2, "NA3"} };
    CVCFVariantList var_list("1", true, set<string>(), sample_cols);
    var_list.ParseLine(line);
    var_list.FinalizeReading();

    BOOST_CHECK(var_list.GetStart() == 12877850);
    BOOST_CHECK(var_list.GetStop() == 12877850);

    BOOST_CHECK(var_list.GetNumberOfVariants() == 3);
    BOOST_CHECK(var_list.GetChrName() == "1");
    BOOST_CHECK(var_list.GetMaxIndex() == 2);

    vector<string> values_first { "",             "C", "<DEL>", "", "", "GT:GQ:DP:HQ" };
    vector<string> values_second{ "nssv16159936", "C", "<DEL>", "", "", "GT:GQ:DP:HQ" };
    vector<string> values_third { "rs12222",      "C", "<DEL>", "", "", "GT:GQ:DP:HQ" };
    // the missing variant id is stored first

    const auto& col_names = CVCFVariantsBase::s_GetColNames();
    const auto& descr = var_list.GetDescriptors();

    size_t index = 0;
    for (auto col_it = col_names.begin(); col_it != col_names.end(); ++col_it, ++index) {
        auto& col_vector = descr[*col_it];
        BOOST_CHECK(col_vector.size() == 3);
        BOOST_CHECK(string(col_vector[0]) == values_first[index]);
        BOOST_CHECK(string(col_vector[1]) == values_second[index]);
        BOOST_CHECK(string(col_vector[2]) == values_third[index]);
    }

    //var_list.List(cout);

    vector<unsigned> positions;
    var_list.GetPositionsForMissingVarID(positions);
    BOOST_CHECK(positions.size() == 1);
    BOOST_CHECK(positions[0] == 12877850);

    auto var_id_iter = descr[CVCFVariantsBase::sm_ID].begin();
    auto var_id_end = descr[CVCFVariantsBase::sm_ID].end();
    for (; var_id_iter != var_id_end; ++var_id_iter) {
        if (!var_id_iter.is_null()) {
            vector<unsigned> positions;
            BOOST_CHECK(var_list.GetPositionsForVariant(*var_id_iter, positions) == true);
            BOOST_CHECK(positions.size() == 1);
            BOOST_CHECK(positions[0] == 12877850);
        }
    }
}

BOOST_AUTO_TEST_CASE(Test_ParseLine_ReadNoSamples)
{
    cout << "\n**************Test_ParseLine_ReadNoSamples************\n" << endl;
    string line = "1	12877850	nssv16159936;.;rs12222	C	<DEL>	.	.	DBVARID;SVTYPE=DEL;END=12933342;SVLEN=-55493;OTH_AN=190	GT:GQ:DP:HQ	0|0:48:1:51,51	1|0:48:8:51,51	1/1:43:5:.,.";

    map<unsigned, string> sample_cols;
    CVCFVariantList var_list("1", true, set<string>(), sample_cols);
    var_list.ParseLine(line);
    var_list.FinalizeReading();

    BOOST_CHECK(var_list.GetStart() == 12877850);
    BOOST_CHECK(var_list.GetStop() == 12877850);

    BOOST_CHECK(var_list.GetNumberOfVariants() == 3);
    BOOST_CHECK(var_list.GetChrName() == "1");
    BOOST_CHECK(var_list.GetMaxIndex() == 2);

    TSeqRange range(var_list.GetStart(), var_list.GetStop());
    for (CVCFVariant_CI var_iter(var_list, range); var_iter; ++var_iter) {
        BOOST_CHECK(var_iter.IsSetFormat());
        BOOST_CHECK(var_iter.GetFormat() == "GT:GQ:DP:HQ");

    }
    auto samples = var_list.GetDescriptors().GetSampleNames();
    BOOST_CHECK(samples.empty());
}

BOOST_AUTO_TEST_CASE(Test_ParseLine_ReadSomeSamples)
{
    cout << "\n**************Test_ParseLine_ReadSomeSamples************\n" << endl;
    string line = "1	12877850	nssv16159936;.;rs12222	C	<DEL>	.	.	DBVARID;SVTYPE=DEL;END=12933342;SVLEN=-55493;OTH_AN=190	GT:GQ:DP:HQ	0|0:48:1:51,51	1|0:48:8:51,51	1/1:43:5:.,.";

    map<unsigned, string> sample_cols{ {0, "NA1"}, {2, "NA3"} };
    CVCFVariantList var_list("1", true, set<string>(), sample_cols);
    var_list.ParseLine(line);
    var_list.FinalizeReading();

    BOOST_CHECK(var_list.GetStart() == 12877850);
    BOOST_CHECK(var_list.GetStop() == 12877850);

    BOOST_CHECK(var_list.GetNumberOfVariants() == 3);
    BOOST_CHECK(var_list.GetChrName() == "1");
    BOOST_CHECK(var_list.GetMaxIndex() == 2);

    auto samples = var_list.GetDescriptors().GetSampleNames();
    BOOST_CHECK(samples.size() == 2);

    TSeqRange range(var_list.GetStart(), var_list.GetStop());
    for (CVCFVariant_CI var_iter(var_list, range); var_iter; ++var_iter) {
        BOOST_CHECK(var_iter.IsSetFormat());
        BOOST_CHECK(var_iter.GetFormat() == "GT:GQ:DP:HQ");
        BOOST_CHECK(var_iter.IsSetSample(samples[0]));
        BOOST_CHECK(var_iter.GetSample("NA1") == "0|0:48:1:51,51");
        BOOST_CHECK(var_iter.IsSetSample(samples[1]));
        BOOST_CHECK(var_iter.GetSample("NA3") == "1/1:43:5:.,.");
    }
}

BOOST_AUTO_TEST_CASE(Test_AddPosition)
{
    cout << "\n****************Test_AddPosition*************\n" << endl;
    CPosToIndex pos_to_index;
    BOOST_CHECK(pos_to_index.GetNumberOfIndexVecs() == 1);

    vector<unsigned> positions{ 1, 3, 4, 4, 5, 6, 7, 100, 1344, 1344, 1344, 1344, 1344, 1400, 1400, 1401, 1401, 1401 };
    for (unsigned index = 0; index < positions.size(); ++index) {
        pos_to_index.Add(positions[index], index);
    }
    pos_to_index.FinalizeReading();

    BOOST_CHECK(pos_to_index.GetNumberOfIndexVecs() == 5);
    BOOST_CHECK(pos_to_index.GetMaxIndex() == positions.size() - 1);
    //pos_to_index.ListVectors(cout);

    auto& vectors = pos_to_index.GetIndexVectors();
    const auto& first_idx = *vectors[0].get();
    BOOST_CHECK(first_idx[1] == 0);
    BOOST_CHECK(first_idx[3] == 1);
    BOOST_CHECK(first_idx[6] == 5);
    BOOST_CHECK(first_idx[1344] == 8);
    BOOST_CHECK(first_idx[1400] == 13);
    BOOST_CHECK(first_idx[1401] == 15);

    const auto& second_idx = *vectors[1].get();
    BOOST_CHECK(second_idx[4] == 3);
    BOOST_CHECK(second_idx[1344] == 9);
    BOOST_CHECK(second_idx[1400] == 14);

    const auto& third_idx = *vectors[2].get();
    BOOST_CHECK(third_idx[1344] == 10);
    BOOST_CHECK(third_idx[1401] == 17);

    const auto& fourth_idx = *vectors[3].get();
    BOOST_CHECK(fourth_idx[1344] == 11);

    const auto& fifth_idx = *vectors[4].get();
    BOOST_CHECK(fifth_idx[1344] == 12);
}


BOOST_AUTO_TEST_CASE(Test_ReadVCF)
{
    cout << "\n**************Test_ReadVCF************\n" << endl;
    const string file_name = "Test_VCF/short.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadData(instr) == true);

    vector<string> names = reader.GetChromosomeNames();
    BOOST_CHECK(names.size() == 2);
    auto it_first = find(names.begin(), names.end(), "1");
    BOOST_CHECK(it_first != names.end());

    auto it_second = find(names.begin(), names.end(), "chr2");
    BOOST_CHECK(it_second != names.end());


    vector<CRef<CVCFVariantList>> variants{ reader.GetVariantsForChr("1"), reader.GetVariantsForChr("chr2") };
    vector<unsigned> nr_variants {17, 16};
    vector<size_t> nr_indexes{ 5, 3 };
    vector<size_t> max_indices{16, 15};
    vector<string> test_samples{"NA00001", "NA00002", "NA00003"};
    size_t i = 0;
    for (auto& it : variants) {
        auto nr_vars = it->GetNumberOfVariants();
        BOOST_CHECK(nr_vars == nr_variants[i]);

        auto nr_index_vectors = it->GetNumberOfIndexVecs();
        BOOST_CHECK(nr_index_vectors == nr_indexes[i]);
        BOOST_CHECK(it->GetMaxIndex() == max_indices[i]);
        ++i;

        vector<string> var_list = it->GetAllVariantIDS();
        BOOST_CHECK(var_list.size() == nr_vars);

        // check that variant_ids are unique
        set<string> tmp_set;
        for (unsigned i = 0; i < var_list.size(); ++i) {
            tmp_set.insert(var_list[i]);
        }
        BOOST_CHECK(tmp_set.size() == var_list.size());

        auto samples = it->GetDescriptors().GetSampleNames();
        BOOST_CHECK(samples == test_samples);

        BOOST_CHECK(it->SerializeVariantData("short_vcf") == true);
        BOOST_CHECK(it->DeserializeAndCheck("short_vcf") == true);
        BOOST_CHECK(it->RemoveSerializedOutput("short_vcf") == true);
    }


}

BOOST_AUTO_TEST_CASE(Test_MultipleVariantIds)
{
    cout << "\n**************Test_MultipleVariantIds************\n" << endl;
    string file_name = "Test_VCF/short.vcf";
    CNcbiFstream instr_1(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadHeader(instr_1) == true);
    BOOST_CHECK(reader.GetVCFversion() == "VCFv4.1");
    BOOST_CHECK(reader.GetAssembly() == "GCF_000001405.13");

    file_name.assign("Test_VCF/short_multiple.vcf");
    CNcbiFstream instr_2(file_name.c_str(), IOS_BASE::in);
    reader.ReadData(instr_2);

    {
        CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
        BOOST_REQUIRE(!first_var.IsNull());
        vector<pair<string, unsigned>> copies;
        first_var->AreVariantIdsUnique(copies);
        BOOST_REQUIRE(copies.size() == 2);
        BOOST_CHECK(copies[0].first == "nssv16177038" && copies[0].second == 2);
        BOOST_CHECK(copies[1].first == "nssv16111133" && copies[1].second == 3);

        vector<unsigned> positions;
        first_var->GetPositionsForMissingVarID(positions);
        BOOST_REQUIRE(positions.size() == 2);
        vector<unsigned> test_positions{ 12931850, 12938000 };
        BOOST_CHECK(positions == test_positions);

        {
            positions.clear();
            first_var->GetPositionsForVariant("nssv16168081", positions);
            BOOST_CHECK(positions.size() == 1 && positions[0] == 12895000);
        }
        {
            positions.clear();
            first_var->GetPositionsForVariant("rs4567789", positions);
            BOOST_CHECK(positions.size() == 1 && positions[0] == 12896425);
        }
        {
            positions.clear();
            first_var->GetPositionsForVariant("nssv16177038", positions);
            vector<unsigned> test_positions{ 12926800, 12931850 };
            BOOST_CHECK(positions == test_positions);
        }
        {
            positions.clear();
            first_var->GetPositionsForVariant("nssv16111133", positions);
            BOOST_CHECK(positions.size() == 1 && positions[0] == 12938300);
        }
        //first_var->List(cout, true);
    }
    {
        CRef<CVCFVariantList> second_var = reader.GetVariantsForChr("chr2");
        BOOST_REQUIRE(!second_var.IsNull());
        vector<pair<string, unsigned>> copies;
        second_var->AreVariantIdsUnique(copies);
        BOOST_REQUIRE(copies.size() == 2);
        BOOST_CHECK(copies[0].first == "rs161" && copies[0].second == 4);
        BOOST_CHECK(copies[1].first == "rs89999777" && copies[1].second == 3);

        vector<unsigned> positions;
        second_var->GetPositionsForMissingVarID(positions);
        vector<unsigned> test_positions{ 1299, 1360 };
        BOOST_CHECK(positions == test_positions);

        {
            positions.clear();
            second_var->GetPositionsForVariant("rs456", positions);
            BOOST_CHECK(positions.size() == 1 && positions[0] == 1368);
        }
        {
            positions.clear();
            second_var->GetPositionsForVariant("rs16161387", positions);
            BOOST_CHECK(positions.size() == 1 && positions[0] == 1289);
        }
        {
            positions.clear();
            second_var->GetPositionsForVariant("rs161", positions);
            vector<unsigned> test_positions{ 1284, 1293, 1367, 1368 };
            BOOST_CHECK(positions == test_positions);
        }
        {
            positions.clear();
            second_var->GetPositionsForVariant("rs89999777", positions);
            vector<unsigned> test_positions{ 1389, 1390, 1391 };
            BOOST_CHECK(positions == test_positions);
        }
        //second_var->List(cout, true);
    }

}


BOOST_AUTO_TEST_CASE(Test_ReadVariantsByChrName)
{
    cout << "\n**************Test_ReadVariantsByChrName************\n" << endl;
    string file_name = "Test_VCF/short_multiple.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadHeader(instr) == true);
    BOOST_CHECK(reader.GetVCFversion() == "VCFv4.2");
    BOOST_CHECK(reader.GetAssembly() == "GCF_000001405.13");

    vector<pair<CConstRef<objects::CSeq_id>, vector<string>>> syns
    { {CConstRef<objects::CSeq_id>(new objects::CSeq_id("lcl|2")), {"chr2"}},
      {CConstRef<objects::CSeq_id>(new objects::CSeq_id("lcl|Mm")),   {"M", "chromM", "VCF_MM", "chrM"}},
      {CConstRef<objects::CSeq_id>(new objects::CSeq_id("lcl|X")), {"X"}},
      {CConstRef<objects::CSeq_id>(new objects::CSeq_id("lcl|1")), {"1", "chr1"}}
    };

    // variants for "X" are not found in the file
    auto variants = reader.ReadVariantsForChrs(instr, syns);
    BOOST_CHECK(variants.size() == 3);
    for (const auto& var_it : variants) {
        const auto& vcf_vars = *var_it.second;
        if (vcf_vars.GetChrName() == "1") {
            BOOST_CHECK(vcf_vars.GetNumberOfVariants() == 21); // includes duplicates
            BOOST_CHECK(vcf_vars.GetMaxIndex() == 20);
            
        }
        else if (vcf_vars.GetChrName() == "chr2") {
            BOOST_CHECK(vcf_vars.GetNumberOfVariants() == 19);
            BOOST_CHECK(vcf_vars.GetMaxIndex() == 18);
        }
        else if (vcf_vars.GetChrName() == "chrM") {
            BOOST_CHECK(vcf_vars.GetNumberOfVariants() == 9);
            BOOST_CHECK(vcf_vars.GetMaxIndex() == 8);
            BOOST_CHECK(vcf_vars.GetChrName() == "chrM");
            BOOST_CHECK(vcf_vars.GetStart() == 10064);
            BOOST_CHECK(vcf_vars.GetStop() == 13409);
        }
    }
}


BOOST_AUTO_TEST_CASE(Test_ReadVariantsByChrName_Async)
{
    cout << "\n**************Test_ReadVariantsByChrName_Async************\n" << endl;
    string file_name = "Test_VCF/short_multiple.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadHeader(instr) == true);

    vector<pair<CConstRef<objects::CSeq_id>, vector<string>>> syns
    { {CConstRef<objects::CSeq_id>(new objects::CSeq_id("lcl|2")), {"chr2"}},
      {CConstRef<objects::CSeq_id>(new objects::CSeq_id("lcl|Mm")),   {"M", "chromM", "VCF_MM", "chrM"}},
      {CConstRef<objects::CSeq_id>(new objects::CSeq_id("lcl|X")), {"X"}},
      {CConstRef<objects::CSeq_id>(new objects::CSeq_id("lcl|1")), {"1", "chr1"}}
    };

    unsigned vars_count{ 0 };
    auto CallOptimizeVarsList = [&vars_count](CVCFVariantList &vcf_vars) {
        if (vcf_vars.GetChrName() == "1") {
            BOOST_CHECK(vcf_vars.GetNumberOfVariants() == 21); // includes duplicates
            BOOST_CHECK(vcf_vars.GetMaxIndex() == 20);
        }
        else if (vcf_vars.GetChrName() == "chr2") {
            BOOST_CHECK(vcf_vars.GetNumberOfVariants() == 19);
            BOOST_CHECK(vcf_vars.GetMaxIndex() == 18);
        }
        else if (vcf_vars.GetChrName() == "chrM") {
            BOOST_CHECK(vcf_vars.GetNumberOfVariants() == 9);
            BOOST_CHECK(vcf_vars.GetMaxIndex() == 8);
            BOOST_CHECK(vcf_vars.GetChrName() == "chrM");
            BOOST_CHECK(vcf_vars.GetStart() == 10064);
            BOOST_CHECK(vcf_vars.GetStop() == 13409);
        }
        vars_count++;
    };

    // variants for "X" are not found in the file
    CColumnarVCFReader::TReportProgress progress;
    reader.ReadVariantsForChrs(instr, syns, nullptr, nullptr, progress, CallOptimizeVarsList);
    BOOST_CHECK(vars_count == 3);
}


BOOST_AUTO_TEST_CASE(Test_ReadData_Async)
{
    cout << "\n**************Test_ReadData_Async************\n" << endl;
    string file_name = "Test_VCF/short_multiple.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadHeader(instr) == true);

    unsigned vars_count{ 0 };
    auto CallOptimizeVarsList = [&vars_count](CVCFVariantList &vcf_vars) {
        if (vcf_vars.GetChrName() == "1") {
            BOOST_CHECK(vcf_vars.GetNumberOfVariants() == 21); // includes duplicates
            BOOST_CHECK(vcf_vars.GetMaxIndex() == 20);
        }
        else if (vcf_vars.GetChrName() == "chr2") {
            BOOST_CHECK(vcf_vars.GetNumberOfVariants() == 19);
            BOOST_CHECK(vcf_vars.GetMaxIndex() == 18);
        }
        else if (vcf_vars.GetChrName() == "chrM") {
            BOOST_CHECK(vcf_vars.GetNumberOfVariants() == 9);
            BOOST_CHECK(vcf_vars.GetMaxIndex() == 8);
            BOOST_CHECK(vcf_vars.GetChrName() == "chrM");
            BOOST_CHECK(vcf_vars.GetStart() == 10064);
            BOOST_CHECK(vcf_vars.GetStop() == 13409);
        }
        vars_count++;
    };

    CColumnarVCFReader::TReportProgress progress;
    reader.ReadData(instr, nullptr, nullptr, progress, CallOptimizeVarsList);
    BOOST_CHECK(vars_count == 3);
}


BOOST_AUTO_TEST_CASE(Test_SerializeIntoSingleBlob)
{
    cout << "\n**************Test_SerializeIntoSingleBlob************\n" << endl;
    const string file_name = "Test_VCF/short.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    reader.ReadData(instr);

    {
        CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
        BOOST_REQUIRE(!first_var.IsNull());
        BOOST_CHECK(first_var->GetNumberOfVariants() == 17);

        const auto& buffer = first_var->GetSerializedData();
        cout << "\nBuild new variant list from serialized data for chr1\n";
        CRef<CVCFVariantList> new_variant(new CVCFVariantList("1", buffer));
        
        BOOST_CHECK(new_variant->GetNumberOfVariants() == first_var->GetNumberOfVariants());
        BOOST_CHECK(new_variant->GetNumberOfIndexVecs() == first_var->GetNumberOfIndexVecs());
        BOOST_CHECK(new_variant->GetAllVariantIDS() == first_var->GetAllVariantIDS());
        BOOST_CHECK(new_variant->GetStart() == 12877850);
        BOOST_CHECK(new_variant->GetStop() == 12938000);

        BOOST_CHECK(*new_variant == *first_var);
        //new_variant->List(cout);
        //new_variant->ListPositionVectors(cout);
    }

    {
        CRef<CVCFVariantList> second_var = reader.GetVariantsForChr("chr2");
        BOOST_REQUIRE(!second_var.IsNull());
        BOOST_CHECK(second_var->GetNumberOfVariants() == 16);

        cout << "\nBuild new variant list from serialized data for chr2\n";
        const auto& buffer = second_var->GetSerializedData();
        CRef<CVCFVariantList> new_variant(new CVCFVariantList("chr2", buffer));
        BOOST_CHECK(new_variant->GetNumberOfVariants() == second_var->GetNumberOfVariants());
        BOOST_CHECK(new_variant->GetNumberOfIndexVecs() == second_var->GetNumberOfIndexVecs());
        BOOST_CHECK(new_variant->GetAllVariantIDS() == second_var->GetAllVariantIDS());
        BOOST_CHECK(new_variant->GetStart() == 1284);
        BOOST_CHECK(new_variant->GetStop() == 1389);
        BOOST_CHECK(*new_variant == *second_var);
        //new_variant->List(cout);
        //new_variant->ListPositionVectors(cout);
    }
}

BOOST_AUTO_TEST_CASE(Test_SerializeAndSave)
{
    cout << "\n**************Test_SerializeAndSave************\n" << endl;
    const string file_name = "Test_VCF/short.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    reader.ReadData(instr);

    {
        CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
        BOOST_CHECK(!first_var.IsNull());
        BOOST_CHECK(first_var->GetNumberOfVariants() == 17);
       
        const string fname = "saved_data.bin";
        first_var->WriteSerializedData(fname);
        CRef<CVCFVariantList> new_variant(new CVCFVariantList("1", fname));
        BOOST_CHECK(new_variant->GetNumberOfVariants() == first_var->GetNumberOfVariants());
        BOOST_CHECK(new_variant->GetNumberOfIndexVecs() == first_var->GetNumberOfIndexVecs());
        BOOST_CHECK(new_variant->GetAllVariantIDS() == first_var->GetAllVariantIDS());

        BOOST_CHECK(*new_variant == *first_var);
        //new_variant->List(cout);
        //new_variant->ListPositionVectors(cout);

        CFile fh(fname);
        if (fh.Exists() && fh.IsFile()) {
            BOOST_CHECK(fh.Remove(CDirEntry::eEntryOnly) == true);
        }
    }

    {
        CRef<CVCFVariantList> second_var = reader.GetVariantsForChr("chr2");
        BOOST_CHECK(!second_var.IsNull());
        BOOST_CHECK(second_var->GetNumberOfVariants() == 16);

        const string fname = "saved_data.bin";
        second_var->WriteSerializedData(fname);
        CRef<CVCFVariantList> new_variant(new CVCFVariantList("chr2", fname));
        BOOST_CHECK(new_variant->GetNumberOfVariants() == second_var->GetNumberOfVariants());
        BOOST_CHECK(new_variant->GetNumberOfIndexVecs() == second_var->GetNumberOfIndexVecs());
        BOOST_CHECK(new_variant->GetAllVariantIDS() == second_var->GetAllVariantIDS());

        BOOST_CHECK(*new_variant == *second_var);
        //new_variant->List(cout);
        //new_variant->ListPositionVectors(cout);

        CFile fh(fname);
        if (fh.Exists() && fh.IsFile()) {
            BOOST_CHECK(fh.Remove(CDirEntry::eEntryOnly) == true);
        }
    }
}


BOOST_AUTO_TEST_CASE(Test_Serialization_NoINFOColumn)
{
    cout << "\n**************Test_Serialization_NoINFOColumn************\n" << endl;
    const string file_name = "Test_VCF/short.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    reader.ReadHeader(instr);
    auto info_fields = reader.GetInfoFields();
    BOOST_CHECK(info_fields.size() == 44);

    {
        string name = "DBVARID";
        auto found_it = find_if(info_fields.begin(), info_fields.end(),
            [&name](const CConstRef<SVcfFieldData>& data) { return (data->m_Name == name); });
        BOOST_CHECK(found_it != info_fields.end());
        BOOST_CHECK((*found_it)->m_Description == "ID is a dbVar accession");
    }

    // skip the entire INFO column
    CNcbiFstream instr_1(file_name.c_str(), IOS_BASE::in);
    reader.LoadAllInfo(false);
    reader.ReadData(instr_1);

    
    CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
    BOOST_REQUIRE(!first_var.IsNull());
    BOOST_CHECK(first_var->GetNumberOfVariants() == 17);
    BOOST_CHECK(first_var->GetDescriptors().GetInfoFieldNames().empty());

    TSeqRange range(first_var->GetStart(), first_var->GetStop());
    CVCFVariant_CI var_iter(*first_var, range);
    for (; var_iter; ++var_iter) {
        BOOST_CHECK(!var_iter.GetVariantID().empty());
        BOOST_CHECK(!var_iter.IsSetInfo());
        BOOST_CHECK(!var_iter.IsSetInfoField("DBVARID"));
    }

    const auto& buffer = first_var->GetSerializedData();
    cout << "\nBuild new variant list from serialized data for chr1\n";
    CRef<CVCFVariantList> new_variant(new CVCFVariantList("1", buffer));

    BOOST_CHECK(new_variant->GetNumberOfVariants() == first_var->GetNumberOfVariants());
    BOOST_CHECK(new_variant->GetNumberOfIndexVecs() == first_var->GetNumberOfIndexVecs());
    BOOST_CHECK(new_variant->GetAllVariantIDS() == first_var->GetAllVariantIDS());
    BOOST_CHECK(new_variant->GetStart() == 12877850);
    BOOST_CHECK(new_variant->GetStop() == 12938000);

    BOOST_CHECK(*new_variant == *first_var);
    //new_variant->List(cout);
    //new_variant->ListPositionVectors(cout);

    CVCFVariant_CI var_iter2(*new_variant, range);
    for (; var_iter2; ++var_iter2) {
        BOOST_CHECK(!var_iter2.GetVariantID().empty());
        BOOST_CHECK(!var_iter2.IsSetInfo());
        BOOST_CHECK(!var_iter2.IsSetInfoField("DBVARID"));
    }
}

BOOST_AUTO_TEST_CASE(Test_Serialization_SomeINFOColumn)
{
    cout << "\n**************Test_Serialization_SomeINFOColumn************\n" << endl;
    const string file_name = "Test_VCF/front_bbb.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    
    set<string> load_info_fields{"AC", "DP", "NS", "AA", "END"};

    CColumnarVCFReader reader;
    reader.LoadAllInfo(false);
    reader.LoadSelectedInfoFields() = load_info_fields;
    reader.ReadData(instr);

    CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
    BOOST_REQUIRE(!first_var.IsNull());
    BOOST_CHECK(first_var->GetNumberOfVariants() == 215);
    BOOST_CHECK(first_var->GetDescriptors().GetInfoFieldNames().size() == 4);
    BOOST_CHECK(first_var->GetDescriptors().GetSampleNames().size() == 2504);

    TSeqRange range(54669, 55);
    CVCFVariant_CI var_iter(*first_var, range);
    for (; var_iter; ++var_iter) {
        if (var_iter.GetPosition() == 54669) {
            BOOST_CHECK(var_iter.IsSetInfo());
            BOOST_CHECK(!var_iter.IsSetInfoField("AC")); // not present in that row
            BOOST_CHECK(!var_iter.IsSetInfoField("EAS_AF")); // not read
            BOOST_CHECK(!var_iter.IsSetInfoField("END")); // not present in the file, though it's present in the header
            BOOST_CHECK(var_iter.IsSetInfoField("NS"));
            BOOST_CHECK(var_iter.GetInfoField("NS") == "2504");
        }
        else if (var_iter.GetPosition() == 55136) {
            BOOST_CHECK(!var_iter.IsSetInfoField("NS"));
            BOOST_CHECK(var_iter.IsSetInfoField("DP"));
            BOOST_CHECK(var_iter.GetInfoField("DP") == "17042");
            
        }
    }

    const auto& buffer = first_var->GetSerializedData();
    CRef<CVCFVariantList> new_variant(new CVCFVariantList("1", buffer));

    BOOST_CHECK(new_variant->GetNumberOfVariants() == first_var->GetNumberOfVariants());
    BOOST_CHECK(new_variant->GetNumberOfIndexVecs() == first_var->GetNumberOfIndexVecs());
    BOOST_CHECK(new_variant->GetAllVariantIDS() == first_var->GetAllVariantIDS());
    BOOST_CHECK(new_variant->GetStart() == first_var->GetStart());
    BOOST_CHECK(new_variant->GetStop() == first_var->GetStop());

    BOOST_CHECK(*new_variant == *first_var);
    {
        CVCFVariant_CI var_iter(*new_variant, range);
        for (; var_iter; ++var_iter) {
            if (var_iter.GetPosition() == 54669) {
                BOOST_CHECK(var_iter.IsSetInfo());
                BOOST_CHECK(!var_iter.IsSetInfoField("AC")); // not present in that row
                BOOST_CHECK(!var_iter.IsSetInfoField("EAS_AF")); // not read
                BOOST_CHECK(!var_iter.IsSetInfoField("END")); // not present in the file, though it's present in the header
                BOOST_CHECK(var_iter.IsSetInfoField("NS"));
                BOOST_CHECK(var_iter.GetInfoField("NS") == "2504");
            }
            else if (var_iter.GetPosition() == 55136) {
                BOOST_CHECK(!var_iter.IsSetInfoField("NS"));
                BOOST_CHECK(var_iter.IsSetInfoField("DP"));
                BOOST_CHECK(var_iter.GetInfoField("DP") == "17042");

            }
        }
    }
}

BOOST_AUTO_TEST_CASE(Test_MissingVariantID)
{
    cout << "\n**************Test_MissingVariantID************\n" << endl;
    const string file_name = "Test_VCF/short_withdot.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    reader.ReadData(instr);

    {
        CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
        BOOST_REQUIRE(!first_var.IsNull());
        BOOST_CHECK(first_var->GetNumberOfVariants() == 18);
        BOOST_CHECK(first_var->GetAllVariantIDS().size() == 18);
        BOOST_CHECK(first_var->GetNumberOfIndexVecs() == 6);

        vector<pair<string, unsigned>> copies;
        BOOST_CHECK(first_var->AreVariantIdsUnique(copies) == true);

        vector<unsigned> positions;
        first_var->GetPositionsForMissingVarID(positions);
        BOOST_CHECK(positions.size() == 3);
        vector<unsigned> test_positions{ 12926800, 12931850, 12938000 };
        BOOST_CHECK(positions == test_positions);
    }

    {
        CRef<CVCFVariantList> second_var = reader.GetVariantsForChr("chr2");
        BOOST_REQUIRE(!second_var.IsNull());
        vector<pair<string, unsigned>> copies;
        BOOST_CHECK(second_var->AreVariantIdsUnique(copies) == true);

        vector<unsigned> positions;
        second_var->GetPositionsForMissingVarID(positions);
        BOOST_CHECK(positions.size() == 3);
        vector<unsigned> test_positions{ 1289, 1293, 1360 };
        BOOST_CHECK(positions == test_positions);
    }
}

BOOST_AUTO_TEST_CASE(Test_GetIndexForPosition)
{
    cout << "\n**************Test_GetIndexForPosition****************\n";
    CPosToIndex pos_to_index;
    BOOST_CHECK(pos_to_index.GetNumberOfIndexVecs() == 1);

    vector<unsigned> positions{ 1284, 1289, 1289, 1289, 1292, 1293, 1293, 1299, 1340, 1360, 1360, 1367,
        1368, 1368, 1368, 1389, 1389, 1389, 1389, 1389 };

    for (unsigned index = 0; index < positions.size(); ++index) {
        pos_to_index.Add(positions[index], index);
    }
    pos_to_index.FinalizeReading();

    BOOST_CHECK(pos_to_index.GetNumberOfIndexVecs() == 5);

    // [1280, 1400] -> [0, 20)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1280, CPosToIndex::eMin) == 0);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1400, CPosToIndex::eMax) == 20);

    // [1230, 1280] -> [0, 0)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1230, CPosToIndex::eMin) == 0);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1280, CPosToIndex::eMax) == 0);

    // [1230, 1284] -> [0, 1)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1230, CPosToIndex::eMin) == 0);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1284, CPosToIndex::eMax) == 1);

    // [1289, 1360] -> [1, 11)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1289, CPosToIndex::eMin) == 1);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1360, CPosToIndex::eMax) == 11);

    // [1400, 1400] -> [20, 20)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1400, CPosToIndex::eMin) == 20);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1400, CPosToIndex::eMax) == 20);

    // [1400, 1440] -> [20, 20)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1400, CPosToIndex::eMin) == 20);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1440, CPosToIndex::eMax) == 20);

    // [1389, 1440] -> [19, 20)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1389, CPosToIndex::eMin) == 15);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1440, CPosToIndex::eMax) == 20);

    // [1289, 1289] -> [1, 4)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1289, CPosToIndex::eMin) == 1);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1289, CPosToIndex::eMax) == 4);

    // [1367, 1367] -> [11, 12)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1367, CPosToIndex::eMin) == 11);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1367, CPosToIndex::eMax) == 12);

    // [1250, 1291] -> [0, 4)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1250, CPosToIndex::eMin) == 0);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1291, CPosToIndex::eMax) == 4);

    // [1283, 1295] -> [0, 7)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1283, CPosToIndex::eMin) == 0);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1295, CPosToIndex::eMax) == 7);

    // [1350, 1389] -> [9, 20)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1350, CPosToIndex::eMin) == 9);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1389, CPosToIndex::eMax) == 20);

    // [1320, 1292] -> [8, 5)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1320, CPosToIndex::eMin) == 8);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1292, CPosToIndex::eMax) == 5);

    // [1310, 1320] -> [8, 8)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1310, CPosToIndex::eMin) == 8);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1320, CPosToIndex::eMax) == 8);

    // [1310, 1340] -> [8, 9)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1310, CPosToIndex::eMin) == 8);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1340, CPosToIndex::eMax) == 9);

    // [1287, 1380] -> [1, 15)
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1287, CPosToIndex::eMin) == 1);
    BOOST_CHECK(pos_to_index.GetIndexForPosition(1380, CPosToIndex::eMax) == 15);

}

BOOST_AUTO_TEST_CASE(Test_VariantIter1)
{
    cout << "\n****Test_VariantIter\n";
    const string file_name = "Test_VCF/short.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadData(instr));
    {
        CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
        BOOST_REQUIRE(!first_var.IsNull());
        BOOST_REQUIRE(first_var->GetStart() == 12877850);
        BOOST_REQUIRE(first_var->GetStop() == 12938000);

        TSeqRange range(first_var->GetStart(), first_var->GetStop());
        CVCFVariant_CI var_iter(*first_var, range);
        BOOST_CHECK(var_iter.GetSize() == 17);

        BOOST_CHECK(var_iter.GetVariantID() == "nssv16159936");
        BOOST_CHECK(var_iter.GetPosition() == 12877850);

        for (; var_iter; ++var_iter) {}
        BOOST_CHECK(var_iter.IsValid() == false);
        
        var_iter.begin();
        ++var_iter;
        BOOST_CHECK(var_iter.IsValid() == true);
        BOOST_CHECK(var_iter.GetVariantID() == "nssv16168081");
        

        vector<string> alts{ "<DEL>", "<DUP>", "<DEL>", "<DEL>", "<CNV>", "<CNV>", "<CNV>",
        "<CNV>", "<CNV>", "<CNV>", "<CNV>", "<CNV>", "<CNV>", "<CNV>", "<CNV>", "<CNV>", "<CNV>" };

        size_t index = 0;
        for (var_iter.begin(); var_iter && index < alts.size(); ++var_iter, ++index) {
            BOOST_CHECK(var_iter.GetAlt() == alts[index]);
        }
    }

    {
        CRef<CVCFVariantList> second_var = reader.GetVariantsForChr("chr2");
        BOOST_REQUIRE(!second_var.IsNull());

        vector<string> ids{ "rs16168081", "rs16161387", "rs4567789", "nssv16170723", "rs161", "rs1678",
            "rs16177038", "rs16177460", ".", "rs16161309" };

        {
            cout << "\n****Test_VariantIter: RANGE_INT \n";
            TSeqRange range_int(1289, 1360);
            CVCFVariant_CI var_iter(*second_var, range_int);
            BOOST_CHECK(var_iter.GetSize() == 10);
            
            size_t index = 0;
            for (; var_iter && index < ids.size(); ++var_iter, ++index) {
                BOOST_CHECK(var_iter.GetVariantID() == ids[index]);
            }
        }
    
        {
            cout << "\n****Test_VariantIter: RANGE_POS\n";
            TSeqRange range_pos(1289, 1289);
            CVCFVariant_CI var_iter_pos(*second_var, range_pos);
            BOOST_CHECK(var_iter_pos.GetSize() == 3);

            size_t index = 0;
            for (var_iter_pos.begin(); var_iter_pos && index < 3; ++var_iter_pos, ++index) {
                BOOST_CHECK(var_iter_pos.GetPosition() == 1289);
                BOOST_CHECK(var_iter_pos.GetVariantID() == ids[index]);
            }
        }

        {
            cout << "\n****Test_VariantIter: RANGE_EMPTY\n";
            TSeqRange range_empty(1310, 1320);
            CVCFVariant_CI var_iter_empty(*second_var, range_empty);
            BOOST_CHECK(var_iter_empty.GetSize() == 0);
            BOOST_CHECK(var_iter_empty.IsValid() == false);
        }

        {
            cout << "\n****Test_VariantIter: RANGE_EMPTY2\n";
            TSeqRange range_empty;
            CVCFVariant_CI var_iter_empty(*second_var, range_empty);
            BOOST_CHECK(var_iter_empty.GetSize() == 0);
            BOOST_CHECK(var_iter_empty.IsValid() == false);
        }

        {
            cout << "\n****Test_VariantIter: RANGE_ONE\n";
            TSeqRange range_one(1310, 1340);
            CVCFVariant_CI var_iter_one(*second_var, range_one);
            BOOST_CHECK(var_iter_one.GetSize() == 1);
            BOOST_CHECK(var_iter_one.GetPosition() == 1340);
            BOOST_CHECK(var_iter_one.GetAlt() == "<CNV>");
            ++var_iter_one;
            BOOST_CHECK(var_iter_one.IsValid() == false);
        }
    }
}

BOOST_AUTO_TEST_CASE(Test_VariantIter2)
{
    cout << "\n**************Test_VariantIter2************\n" << endl;
    const string file_name = "Test_VCF/short_multiple.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    reader.ReadData(instr);

    { // testing sample names
        auto samples = reader.GetSamples();
        map<unsigned, string> test_samples = { {0, "NA00001"}, {1, "NA00002"}, {2, "NA00003"} };
        BOOST_CHECK(samples == test_samples);
    }

    CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
    BOOST_REQUIRE(!first_var.IsNull());

    TSeqRange range(12926800, 12931855);
    CVCFVariant_CI var_iter(*first_var, range);
    BOOST_CHECK(var_iter.GetSize() == 8);

    size_t index = 0;
    for (; var_iter; ++var_iter, ++index) {
        if (index == 2 || index == 7) {
            BOOST_CHECK(var_iter.GetVariantID() == "nssv16177038");
        }
        if (index == 2) {
            BOOST_CHECK(var_iter.GetPosition() == 12926800);
            BOOST_CHECK(var_iter.GetSample("NA00001") == "1|2:21:6:23,27");
            BOOST_CHECK(var_iter.GetSample("NA00002") == "2|1:2:0:18,4");
            BOOST_CHECK(var_iter.GetSample("NA00003") == "2/2:35:4");
        }
        else if (index == 7) {
            BOOST_CHECK(var_iter.GetPosition() == 12931850);
            BOOST_CHECK(var_iter.GetSample("NA00001") == "0|0:54:7:56,6");
            BOOST_CHECK(var_iter.GetSample("NA00002") == "0|0:48:4:51,5");
            BOOST_CHECK(var_iter.GetSample("NA00003") == "0/0:60:2");
        }
    }

    CRef<CVCFVariantList> sec_var = reader.GetVariantsForChr("chr2");
    {
        TSeqRange range(1360, 1500);
        CVCFVariant_CI var_iter(*sec_var, range);
        for (; var_iter; ++var_iter) {
            if (var_iter.GetPosition() == 1389) {
                BOOST_CHECK(var_iter.GetFormat() == "GT:GQ:DP:HQ");
                BOOST_CHECK(!var_iter.IsSetSample("NA00001"));
                BOOST_CHECK(!var_iter.IsSetSample("NA00002"));
                BOOST_CHECK(!var_iter.IsSetSample("NA00003"));
            }
            else if (var_iter.GetPosition() == 1390) {
                BOOST_CHECK(var_iter.GetFormat() == CVCFVariantsBase::sm_MissingValue);
                BOOST_CHECK(!var_iter.IsSetSample("NA00001"));
                BOOST_CHECK(!var_iter.IsSetSample("NA00002"));
                BOOST_CHECK(!var_iter.IsSetSample("NA00003"));
            }
            else if (var_iter.GetPosition() == 1391) {
                BOOST_CHECK(var_iter.GetFormat() == "GT:GQ:DP:HQ");
                BOOST_CHECK(var_iter.GetSample("NA00001") == "0|0:54:7:57,60");
                BOOST_CHECK(var_iter.GetSample("NA00002") == "0|0:48:4:52,54");
                BOOST_CHECK(var_iter.GetSample("NA00003") == "0/0:61:4");
            }
        }
    }
}


BOOST_AUTO_TEST_CASE(Test_FindEq)
{
    using bvector_type = bm::bvector<>;
    using TSparseStrVector = bm::str_sparse_vector<char, bvector_type, 64>;
    
    TSparseStrVector str_vector;
    auto in_iter = str_vector.get_back_inserter();

    /*
    // Example #1:
    in_iter = "rs123456";
    in_iter = "rs23456";
    in_iter = ".";
    in_iter = "esv4567";
    in_iter = "esv89000";
    in_iter = ".";
    in_iter = "esv4444";
    in_iter = "rs22222";
    in_iter = "rs";
    in_iter = ".";
    in_iter.flush();
    */

    
    // Example #2:
    //This example doesn't work
    in_iter = "nssv16159936";
    in_iter = "nssv16168081";
    in_iter = "nssv16161387";
    in_iter = "rs4567789";
    in_iter = ".";
    in_iter = "nssv16175917";
    in_iter = "nssv16177038";
    in_iter = "nssv16177460";
    in_iter = ".";
    in_iter = "nssv16161309";
    in_iter.flush();
    
   
    str_vector.optimize();

    // print them out:
    /*auto it = str_vector.begin();
    auto it_end = str_vector.end();
    for (; it != it_end; ++it) {
        cout << *it << endl;
    }
    */
    bm::sparse_vector_scanner<TSparseStrVector> scanner;
    TSparseStrVector::bvector_type result;
    scanner.find_eq_str(str_vector, ".", result);

    BOOST_CHECK(result.count() == 2);
    auto en = result.first();
    BOOST_CHECK(*en == 4);
    ++en;
    BOOST_CHECK(*en == 8);
}


BOOST_AUTO_TEST_CASE(Test_DeserializeSingleRange)
{
    cout << "\n**************Test_DeserializeSingleRange************\n" << endl;
    const string file_name = "Test_VCF/short.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadData(instr) == true);

    CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
    BOOST_REQUIRE(!first_var.IsNull());
    BOOST_CHECK(first_var->GetNumberOfVariants() == 17);

    const auto& buffer = first_var->GetSerializedData();
    vector<char> local_cpy(buffer.begin(), buffer.end());
    TSeqRange range;

    {
        // no range, therefore no data will be deserialized, only the index vectors
        CRef<CVCFVariantsBase> sliced_vars(new CVCFSlicedVariants(local_cpy));
        BOOST_CHECK(sliced_vars->GetNumberOfVariants() == 0);
    }
    {
        // set the range but don't specify any data columns to decode,
        // therefore no data will be deserialized, only the index vectors
        range.Set(30, 12938301);
        CRef<CVCFVariantsBase> new_varlist(new CVCFSlicedVariants(local_cpy, &range));
        BOOST_CHECK(new_varlist->GetStart() == 0);
        BOOST_CHECK(new_varlist->GetStop() == 0);
        BOOST_CHECK(new_varlist->GetNumberOfIndexVecs() == 5);
        BOOST_CHECK(new_varlist->GetMaxIndex() == 16);
        BOOST_CHECK(new_varlist->GetNumberOfVariants() == 0);
    }
    {
        range.Set(30, 12938301); // include all variants
        set<string> data_cols{ CVCFVariantsBase::sm_ID, CVCFVariantsBase::sm_SAMPLES };
        CRef<CVCFVariantsBase> new_varlist(new CVCFSlicedVariants(local_cpy, &range, data_cols));
        BOOST_CHECK(new_varlist->GetStart() == 30);
        BOOST_CHECK(new_varlist->GetStop() == 12938301);
        BOOST_CHECK(first_var->GetNumberOfVariants() == new_varlist->GetNumberOfVariants());
        BOOST_CHECK(new_varlist->Count(range) == 17);

        vector<unsigned> positions;
        BOOST_CHECK(new_varlist->GetPositionsForVariant("rs1234567", positions) == true);
        BOOST_CHECK(positions.size() == 1);
        BOOST_CHECK(positions.front() == 12938000);

        auto samples = new_varlist->GetDescriptors().GetSampleNames();
        BOOST_CHECK(samples.size() == 3);
    }
    {
        range.Set(12896460, 12896600); // an empty range, no variants should be included
        set<string> data_cols{ CVCFVariantsBase::sm_ID, CVCFVariantsBase::sm_SAMPLES };
        CRef<CVCFVariantsBase> new_varlist(new CVCFSlicedVariants(local_cpy, &range, data_cols));
        BOOST_CHECK(new_varlist->GetStart() == 12896460);
        BOOST_CHECK(new_varlist->GetStop() == 12896600);
        BOOST_CHECK(new_varlist->GetNumberOfVariants() == 0);
        BOOST_CHECK(new_varlist->Count(range) == 0);
        auto info_fields = new_varlist->GetDescriptors().GetInfoFieldNames();
        BOOST_CHECK(info_fields.empty());
        auto samples = new_varlist->GetDescriptors().GetSampleNames();
        BOOST_CHECK(samples.empty());

        CVCFVariant_CI var_iter(*new_varlist, range);
        BOOST_CHECK(var_iter.IsValid() == false);
        BOOST_CHECK(var_iter.GetSize() == 0);

        // even though the range does not have any variants, it can find the relevant variantID
        vector<unsigned> positions;
        BOOST_CHECK(new_varlist->GetPositionsForVariant("rs1234567", positions) == true);
        BOOST_CHECK(positions.size() == 1);
        BOOST_CHECK(positions.front() == 12938000);
    }
    {
        range.Set(12894460, 12895000);  // contains a single element
        set<string> data_cols{ CVCFVariantsBase::sm_ID, CVCFVariantsBase::sm_INFO, CVCFVariantsBase::sm_FORMAT, CVCFVariantsBase::sm_SAMPLES };
        CRef<CVCFVariantsBase> new_varlist(new CVCFSlicedVariants(local_cpy, &range, data_cols));
        BOOST_CHECK(new_varlist->GetStart() == 12894460);
        BOOST_CHECK(new_varlist->GetStop() == 12895000);
        BOOST_CHECK(new_varlist->GetNumberOfVariants() == 1);
        BOOST_CHECK(new_varlist->Count(range) == 1);

        CVCFVariant_CI var_iter(*new_varlist, range);
        BOOST_CHECK(var_iter.IsValid());
        BOOST_CHECK(var_iter.GetSize() == 1);
        BOOST_CHECK(var_iter.GetPosition() == 12895000);

        // check the info sub-columns
        auto info_fields = new_varlist->GetDescriptors().GetInfoFieldNames();
        BOOST_CHECK(info_fields.size() == 27);
        BOOST_CHECK(var_iter.GetInfoField("AMR_AN") == "1896");
        string value;
        try {
            value = var_iter.GetInfoField("Amr_an");
        }
        catch (const CException&) {}
        BOOST_CHECK(value.empty());
        BOOST_CHECK(var_iter.GetInfoField("DBVARID") == kEmptyStr);
        BOOST_CHECK(var_iter.GetInfoField("SVTYPE") == "DUP");
        BOOST_CHECK(var_iter.GetInfoField("END") == "12940000");

        // check sample values
        BOOST_CHECK(var_iter.IsSetFormat());
        BOOST_CHECK(var_iter.GetFormat() == "GT:GQ:DP:HQ");
        auto samples = new_varlist->GetDescriptors().GetSampleNames();
        BOOST_CHECK(samples.size() == 3);
        BOOST_CHECK(var_iter.IsSetSample(samples[0]));
        BOOST_CHECK(var_iter.GetSample(samples[0]) == "0|0:49:3:58,50");
        BOOST_CHECK(var_iter.IsSetSample(samples[2]));
        BOOST_CHECK(var_iter.GetSample(samples[2]) == "0/0:41:3");

        vector<unsigned> positions;
        BOOST_CHECK(new_varlist->GetPositionsForVariant("nssv16168081", positions) == true);
        BOOST_CHECK(positions.front() == 12895000);

        positions.clear();
        BOOST_CHECK(new_varlist->GetPositionsForVariant("nssv16161309", positions) == true);
        BOOST_CHECK(positions.front() == 12931850);
    }
    {
        range.Set(12894460, 12895001);  // a single element
        set<string> data_cols{ CVCFVariantsBase::sm_ID };
        CRef<CVCFSlicedVariants> new_varlist(new CVCFSlicedVariants(local_cpy, &range, data_cols));
        BOOST_CHECK(new_varlist->GetStart() == 12894460);
        BOOST_CHECK(new_varlist->GetStop() == 12895001);
        BOOST_CHECK(new_varlist->GetNumberOfVariants() == 1);
        BOOST_CHECK(new_varlist->Count(range) == 1);

        CVCFVariant_CI var_iter(*new_varlist, range);
        BOOST_CHECK(var_iter.IsValid());
        BOOST_CHECK(var_iter.GetSize() == 1);
        BOOST_CHECK(var_iter.GetPosition() == 12895000);
    }
    {
        range.Set(12896420, 12931855); // 10 elements
        set<string> data_cols{ CVCFVariantsBase::sm_ID, "SVTYPE", "END", "DESC" };
        CRef<CVCFSlicedVariants> new_varlist(new CVCFSlicedVariants(local_cpy, &range, data_cols));
        BOOST_CHECK(new_varlist->GetStart() == 12896420);
        BOOST_CHECK(new_varlist->GetStop() == 12931855);
        BOOST_CHECK(new_varlist->GetNumberOfVariants() == 10);
        BOOST_CHECK(new_varlist->Count(range) == 10);

        CVCFVariant_CI var_iter(*new_varlist, range);
        BOOST_CHECK(var_iter.IsValid());
        BOOST_CHECK(var_iter.GetSize() == 10);

        for (; var_iter; ++var_iter) {
            if (var_iter.GetVariantID() == CVCFVariantsBase::sm_MissingValue) {
                BOOST_CHECK(var_iter.GetInfoField("SVTYPE") == "CNV");
                BOOST_CHECK(var_iter.GetInfoField("END") == "12938000");
                BOOST_CHECK(var_iter.GetInfoField("DESC") == "gnomAD Structural Variants");
            }
        }

        vector<unsigned> positions;
        BOOST_CHECK(new_varlist->GetPositionsForVariant("16175917", positions) == false);
        
        positions.clear();
        BOOST_CHECK(new_varlist->GetPositionsForVariant(".", positions) == true);
        BOOST_CHECK(positions.front() == 12931850);

        positions.clear();
        BOOST_CHECK(new_varlist->GetPositionsForVariant("rs345678", positions) == true);
        BOOST_CHECK(positions.front() == 12938000);

    }
    {
        range.Set(12938001, 12939000); // a range with no variants
        set<string> data_cols{ CVCFVariantsBase::sm_ID };
        CRef<CVCFSlicedVariants> new_varlist(new CVCFSlicedVariants(local_cpy, &range, data_cols));
        BOOST_CHECK(new_varlist->GetStart() == 12938001);
        BOOST_CHECK(new_varlist->GetStop() == 12939000);
        BOOST_CHECK(new_varlist->GetNumberOfVariants() == 0);
        BOOST_CHECK(new_varlist->Count(range) == 0);
        BOOST_CHECK(first_var->Count(range) == 0);

        CVCFVariant_CI var_iter(*new_varlist, range);
        BOOST_CHECK(var_iter.IsValid() == false);
        BOOST_CHECK(var_iter.GetSize() == 0);
    }
}

BOOST_AUTO_TEST_CASE(Test_FindVariants)
{
    cout << "\n**************Test_FindVariants************\n" << endl;
    const string file_name = "Test_VCF/short_multiple.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;

    vector<pair<CConstRef<objects::CSeq_id>, vector<string>>> syns{ {CConstRef<objects::CSeq_id>(nullptr), {"chr2"}} };

    auto variants = reader.ReadVariantsForChrs(instr, syns);
    BOOST_REQUIRE(!variants.empty());
    auto second_var = variants.front().second;
    BOOST_REQUIRE(!second_var.IsNull());
    BOOST_CHECK(second_var->GetNumberOfVariants() == 19);

    vector<unsigned> positions;
    second_var->GetPositionsForMissingVarID(positions);
    {
        vector<unsigned> test_positions{ 1299, 1360 };
        BOOST_CHECK(positions == test_positions);
    }
   

    positions.clear();
    second_var->GetPositionsForVariant("rs161", positions);
    {
        vector<unsigned> test_positions{ 1284, 1293, 1367, 1368 };
        BOOST_CHECK(positions == test_positions);
    }

    CVCFVariantList& sec_vars = dynamic_cast<CVCFVariantList&>(second_var.GetNCObject());
    const auto& buffer = sec_vars.GetSerializedData();

    vector<char> local_cpy(buffer.begin(), buffer.end());
    TSeqRange range(1340, 1390);
    set<string> data_cols{ CVCFVariantsBase::sm_ID };
    CVCFSlicedVariants sliced(local_cpy, &range, data_cols);
    BOOST_CHECK(sliced.GetNumberOfVariants() == 10);

    positions.clear();
    BOOST_CHECK(sliced.GetPositionsForVariant("rs161", positions));
    {
        vector<unsigned> test_positions{ 1367, 1368 };
        BOOST_CHECK(positions == test_positions);
    }

    sliced.SetRange(1230, 1300);
    BOOST_CHECK(sliced.GetStart() == 1230);
    BOOST_CHECK(sliced.GetStop() == 1300);
    
    positions.clear();
    BOOST_CHECK(sliced.GetPositionsForVariant("rs161", positions));
    {
        vector<unsigned> test_positions{ 1284, 1293};
        BOOST_CHECK(positions == test_positions);
    }

    // when searching for the missing variantID, it will 
    // always deserialize the entire ID vector
    positions.clear();
    sliced.GetPositionsForMissingVarID(positions);
    {
        vector<unsigned> test_positions{ 1299, 1360 };
        BOOST_CHECK(positions == test_positions);
    }

    sliced.SetRange(1390, 1500);
    positions.clear();
    sliced.GetPositionsForMissingVarID(positions);
    {
        vector<unsigned> test_positions{ 1299, 1360 };
        BOOST_CHECK(positions == test_positions);
    }
    // deserializing the ID vector does not change the start and stop positions
    BOOST_CHECK(sliced.GetStart() == 1390);
    BOOST_CHECK(sliced.GetStop() == 1500);
    
}

template<typename SV>
void PrintSparseVector(const SV& vec, CNcbiOstream& out)
{
    typename SV::const_iterator it = vec.begin();
    typename SV::const_iterator it_end = vec.end();

    for (size_t index = 0; it != it_end; ++it, ++index) {
        if (!it.is_null()) {
            out << index << ": " << *it;
        }
        else {
            out << index << ": " << "null";
        }
        out << endl;
    }
    out << endl;
}

BOOST_AUTO_TEST_CASE(Test_ScrollToRange)
{
    cout << "\n**************Test_ScrollToRange************\n" << endl;
    const string file_name = "Test_VCF/short.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadData(instr) == true);

    {
        CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
        BOOST_REQUIRE(!first_var.IsNull());
        BOOST_CHECK(first_var->GetNumberOfVariants() == 17);
        const auto& buffer = first_var->GetSerializedData();

        vector<char> local_cpy(buffer.begin(), buffer.end());
        TSeqRange range;
        set<string> data_cols{CVCFVariantsBase::sm_ID, CVCFVariantsBase::sm_INFO};
        CVCFSlicedVariants sliced(local_cpy, nullptr, data_cols);
        BOOST_CHECK(sliced.GetNumberOfVariants() == 0);

        {
            CVCFVariant_CI var_iter(sliced, range);
            BOOST_CHECK(var_iter.IsValid() == false);
            // when range is empty -> don't deserialize any of the string sparse vectors
        }

        range.Set(12896400, 1292690);  // empty range
        sliced.SetRange(range);
        BOOST_CHECK(sliced.GetNumberOfVariants() == 0);
        {
            CVCFVariant_CI var_iter(sliced, range);
            BOOST_CHECK(var_iter.IsValid() == false);
        }

        sliced.SetRange(12896400, 12926900);
        BOOST_CHECK(sliced.GetNumberOfVariants() == 6);
        const auto& descrs = sliced.GetDescriptors();
        const auto& id_vector = descrs[CVCFVariantsBase::sm_ID];

        auto it = id_vector.begin();
        BOOST_CHECK(it.is_null());
        ++it;  BOOST_CHECK(it.is_null());
        ++it;
        vector<string> ids = { "nssv16161387", "rs4567789", "nssv16170723", "nssv16175917", "nssv16177038", "nssv16177460" };
        for (auto str_it = ids.begin(); str_it != ids.end(); ++str_it, ++it) {
            BOOST_CHECK(NStr::EqualNocase(*str_it, *it));
        }
        BOOST_CHECK(it.is_null());
        {
            auto info_fields = sliced.GetDescriptors().GetInfoFieldNames();
            BOOST_CHECK(info_fields.size() == 27);
            range.Set(12896400, 12926900);
            CVCFVariant_CI var_iter(sliced, range);
            while (var_iter.GetVariantID() != "rs4567789") {
                ++var_iter;
            }
            BOOST_CHECK(var_iter.GetInfoField("END") == "12930275");
            BOOST_CHECK(var_iter.GetInfoField("LINKS") == "dbVar:nsv4517266");
            BOOST_CHECK(var_iter.GetInfoField("OTH_AN") == "190");

            // to check the partial deserialization of the vector
            auto& end_info = sliced.GetDescriptors().GetInfoField("END");
            PrintSparseVector(end_info, cout);
        }

        sliced.SetRange(12877800, 12938300); // all variants
        BOOST_CHECK(sliced.GetNumberOfVariants() == 17);
        BOOST_CHECK(sliced.GetStart() == 12877800);
        BOOST_CHECK(sliced.GetStop() == 12938300);

        sliced.SetRange(12926900, 12931880); // first variant is the missing variant
        BOOST_CHECK(sliced.GetNumberOfVariants() == 4);
        BOOST_CHECK(sliced.GetStart() == 12926900);
        BOOST_CHECK(sliced.GetStop() == 12931880);
    }
}

BOOST_AUTO_TEST_CASE(Test_Histogram1)
{
    cout << "\n**************Test_Histogram1************\n" << endl;
    const string file_name = "Test_VCF/short_hist.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadData(instr) == true);

    CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("chr2");
    BOOST_REQUIRE(!first_var.IsNull());
    BOOST_CHECK(first_var->GetNumberOfVariants() == 21);
    BOOST_CHECK(first_var->GetMaxIndex() == 20);
    BOOST_CHECK(first_var->GetNumberOfIndexVecs() == 4);

    CVCFVariantsBase::TIndexVector hist;
    CVCFVariantsBase::svector hist_sv;
    {
        first_var->GetHistogram(hist, 10);
        //PrintSparseVector(hist, cout);

        CVCFVariantsBase::TIndexVector tmp_hist;
        CVCFVariantsBase::svector check_hist_vect(bm::use_null);
        for (size_t i = 0; i < 1421; i += 10) {
            check_hist_vect[i] = 0;
        }
        check_hist_vect[1280] = 4;
        check_hist_vect[1290] = 5;
        check_hist_vect[1340] = 1;
        check_hist_vect[1360] = 7;
        check_hist_vect[1380] = 1;
        check_hist_vect[1420] = 3;

        tmp_hist.load_from(check_hist_vect);

        BM_DECLARE_TEMP_BLOCK(tb);
        tmp_hist.optimize(tb);
        tmp_hist.sync();
        BOOST_CHECK(hist.equal(tmp_hist));
    }

    {
        first_var->GetHistogram(hist_sv, 10);
        //PrintSparseVector(hist_sv, cout);

        CVCFVariantsBase::svector check_hist_vect;
        for (size_t i = 0; i < 142; ++i) {
            check_hist_vect[i] = 0;
        }
        check_hist_vect[128] = 4;
        check_hist_vect[129] = 5;
        check_hist_vect[134] = 1;
        check_hist_vect[136] = 7;
        check_hist_vect[138] = 1;
        check_hist_vect[142] = 3;

        BM_DECLARE_TEMP_BLOCK(tb);
        check_hist_vect.optimize(tb);
        BOOST_CHECK(hist_sv.equal(check_hist_vect));
    }

    {
        first_var->GetHistogram(hist, 50);
        //PrintSparseVector(hist, cout);

        CVCFVariantsBase::TIndexVector tmp_hist;
        CVCFVariantsBase::svector check_hist_vect(bm::use_null);
        for (size_t i = 0; i < 1401; i += 50) {
            check_hist_vect[i] = 0;
        }
        check_hist_vect[1250] = 9;
        check_hist_vect[1300] = 1;
        check_hist_vect[1350] = 8;
        check_hist_vect[1400] = 3;
        
        tmp_hist.load_from(check_hist_vect);

        BM_DECLARE_TEMP_BLOCK(tb);
        tmp_hist.optimize(tb);
        tmp_hist.sync();
        BOOST_CHECK(hist.equal(tmp_hist));
    }

    {
        first_var->GetHistogram(hist_sv, 50);
        //PrintSparseVector(hist_sv, cout);
        BOOST_CHECK(hist_sv.size() == 1420/50 + 1);

        CVCFVariantsBase::svector check_hist_vect;
        for (size_t i = 0; i < 29; ++i) {
            check_hist_vect[i] = 0;
        }
        check_hist_vect[1250/50] = 9;
        check_hist_vect[1300/50] = 1;
        check_hist_vect[1350/50] = 8;
        check_hist_vect[1400/50] = 3;

        BM_DECLARE_TEMP_BLOCK(tb);
        check_hist_vect.optimize(tb);
        BOOST_CHECK(hist_sv.equal(check_hist_vect));
    }
    
    first_var->GetHistogram(hist, 0);
    BOOST_CHECK(hist.empty());

    first_var->GetHistogram(hist_sv, 0);
    BOOST_CHECK(hist_sv.empty());

    first_var->GetHistogram(hist, 1500);
    {
        BOOST_CHECK(hist.size() == 1);
        auto hist_it = hist.begin();
        BOOST_CHECK(*hist_it == 21);
    }

    first_var->GetHistogram(hist_sv, 1500);
    {
        BOOST_CHECK(hist.size() == 1);
        auto hist_it = hist_sv.begin();
        BOOST_CHECK(*hist_it == 21);
    }
}

BOOST_AUTO_TEST_CASE(Test_Histogram2)
{
    cout << "\n**************Test_Histogram2************\n" << endl;
    const string file_name = "Test_VCF/front_bbb.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadData(instr) == true);

    CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("1");
    BOOST_REQUIRE(!first_var.IsNull());
    BOOST_CHECK(first_var->GetNumberOfVariants() == 215);
    BOOST_CHECK(first_var->GetMaxIndex() == 214);
    BOOST_CHECK(first_var->GetNumberOfIndexVecs() == 2);
    
    TSeqRange range;
    BOOST_CHECK(first_var->Count(range) == 0);
    range.Set(17709, 17748);
    BOOST_CHECK(first_var->Count(range) == 2);
    range.Set(17700, 20000);
    BOOST_CHECK(first_var->Count(range) == 7);
    //first_var->ListPositionVectors(cout);

    CVCFVariantsBase::TIndexVector hist;

    {
        first_var->GetHistogram(hist, 1000);
       // PrintSparseVector(hist, cout);

        CVCFVariantsBase::TIndexVector tmp_hist;
        CVCFVariantsBase::svector check_hist_vect(bm::use_null);
        for (size_t i = 0; i < 56000; i += 1000) {
            check_hist_vect[i] = 0;
        }
        check_hist_vect[10000] = 11;
        check_hist_vect[11000] = 3;
        check_hist_vect[13000] = 21;
        check_hist_vect[14000] = 14;
        check_hist_vect[15000] = 26;
        check_hist_vect[16000] = 17;
        check_hist_vect[17000] = 12;
        check_hist_vect[18000] = 2;
        check_hist_vect[19000] = 1;
        check_hist_vect[20000] = 2;
        check_hist_vect[28000] = 1;
        check_hist_vect[29000] = 1;
        check_hist_vect[30000] = 5;
        check_hist_vect[46000] = 5;
        check_hist_vect[47000] = 5;
        check_hist_vect[48000] = 5;
        check_hist_vect[49000] = 13;
        check_hist_vect[50000] = 2;
        check_hist_vect[51000] = 21;
        check_hist_vect[52000] = 7;
        check_hist_vect[53000] = 3;
        check_hist_vect[54000] = 29;
        check_hist_vect[55000] = 9;

        tmp_hist.load_from(check_hist_vect);

        BM_DECLARE_TEMP_BLOCK(tb);
        tmp_hist.optimize(tb);
        tmp_hist.sync();
        BOOST_CHECK(hist.equal(tmp_hist));
    }

    first_var->GetHistogram(hist, 0);
    BOOST_CHECK(hist.empty());

    first_var->GetHistogram(hist, 55330);
    BOOST_CHECK(hist.size() == 1);
    auto hist_it = hist.begin();
    BOOST_CHECK(*hist_it == 215);
}

BOOST_AUTO_TEST_CASE(Test_SplitInfoColumn)
{
    cout << "\n**************Test_SplitInfoColumn************\n" << endl;
    string file_name = "Test_VCF/rw-893.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    {
        CNcbiFstream instr1(file_name.c_str(), IOS_BASE::in);
        CColumnarVCFReader reader;
        BOOST_CHECK(reader.ReadHeader(instr1) == true);
        BOOST_CHECK(reader.GetVCFversion() == "VCFv4.0");
        BOOST_CHECK(reader.GetAssembly() == "1000GenomesPilot-NCBI36");
        auto info_fields = reader.GetInfoFields();
        for (const auto& it : info_fields) {
            if (it->m_Name == "DB") {
                BOOST_CHECK(it->m_Description == "dbSNP membership, build 129");
            }
            else if (it->m_Name == "H2") {
                BOOST_CHECK(it->m_Description == "HapMap2 membership, >5% minor allele frequency in 1+ populations");
            }
        }
    }

    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadData(instr) == true);

    CRef<CVCFVariantList> first_var = reader.GetVariantsForChr("20");
    BOOST_REQUIRE(!first_var.IsNull());
    auto info_fields =  first_var->GetDescriptors().GetInfoFieldNames();
    BOOST_CHECK(info_fields.size() == 6);
    //first_var->List(cout);

    TSeqRange range(first_var->GetStart(), first_var->GetStop());
    CVCFVariant_CI var_iter(*first_var, range);
    BOOST_CHECK(var_iter.GetSize() == first_var->GetNumberOfVariants());

    size_t index = 0;
    for (; var_iter; ++var_iter) {
        //cout << var_iter.GetInfo() << endl;
        if (index == 0) {
            BOOST_CHECK(var_iter.GetInfoField("NS") == "3");
            BOOST_CHECK(var_iter.GetInfoField("DB").empty());
            BOOST_CHECK(var_iter.GetInfoField("DP") == "14");
            BOOST_CHECK(var_iter.GetInfoField("AF") == "0.5");

            BOOST_CHECK(!var_iter.IsSetInfoField("AA"));
            BOOST_CHECK(var_iter.GetInfoField("AA") == "null");

            BOOST_CHECK(var_iter.GetInfoField("H2").empty());
        }
        else if (index == 1) {
            BOOST_CHECK(var_iter.GetInfoField("NS") == "3");
            BOOST_CHECK(!var_iter.IsSetInfoField("DB"));
            BOOST_CHECK(var_iter.GetInfoField("DB") == "null");
            BOOST_CHECK(var_iter.GetInfoField("DP") == "11");
            BOOST_CHECK(var_iter.GetInfoField("AF") == "0.017");
            BOOST_CHECK(var_iter.GetInfoField("AA") == "null");
            BOOST_CHECK(var_iter.GetInfoField("H2") == "null");
        }
        else if (index == 2) {
            BOOST_CHECK(var_iter.GetInfoField("NS") == "2");
            BOOST_CHECK(var_iter.GetInfoField("DB").empty());
            BOOST_CHECK(var_iter.GetInfoField("DP") == "10");
            BOOST_CHECK(var_iter.GetInfoField("AF") == "0.333,0.667");
            BOOST_CHECK(var_iter.GetInfoField("AA") == "T");

            BOOST_CHECK(!var_iter.IsSetInfoField("H2"));
            BOOST_CHECK(var_iter.GetInfoField("H2") == "null");
        }
        else if (index == 3) {
            BOOST_CHECK(var_iter.GetInfoField("NS") == "3");
            BOOST_CHECK(var_iter.GetInfoField("DB") == "null");
            BOOST_CHECK(var_iter.GetInfoField("DP") == "13");
            BOOST_CHECK(var_iter.GetInfoField("AF") == "null");
            BOOST_CHECK(var_iter.GetInfoField("AA") == "T");
            BOOST_CHECK(var_iter.GetInfoField("H2") == "null");
        }
        else if (index == 4) {
            BOOST_CHECK(var_iter.GetInfoField("NS") == "3");
            BOOST_CHECK(var_iter.GetInfoField("DB") == "null");
            BOOST_CHECK(var_iter.GetInfoField("DP") == "9");
            BOOST_CHECK(var_iter.GetInfoField("AF") == "null");
            BOOST_CHECK(var_iter.GetInfoField("AA") == "G");

            BOOST_CHECK(!var_iter.IsSetInfoField("H2"));
            BOOST_CHECK(var_iter.GetInfoField("H2") == "null");
        }
        ++index;
    }
}


BOOST_AUTO_TEST_CASE(Test_ReadSalmon)
{
    cout << "\n**************Test_ReadSalmon************\n" << endl;
    string file_name = "Test_VCF/salmon.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);
    
    CColumnarVCFReader reader;
    BOOST_CHECK(reader.ReadHeader(instr) == true);
    BOOST_CHECK(reader.GetVCFversion() == "VCFv4.2");
    BOOST_CHECK(reader.GetAssembly() == "GCA_000233375.4");
}

BOOST_AUTO_TEST_CASE(Test_Encoding)
{
    cout << "\n**************Test_Encoding************\n" << endl;
    string file_name = "Test_VCF/grch38.vcf";
    CNcbiFstream instr(file_name.c_str(), IOS_BASE::in);

    CColumnarVCFReader reader;
    // INFO lines may appear also after the #CHROM header line
    // They may also be duplicated
    auto info_fields = reader.GetInfoFields();
    BOOST_CHECK(info_fields.empty());

    reader.ReadHeader(instr);
    info_fields = reader.GetInfoFields();
    BOOST_CHECK(info_fields.size() == 26); // don't count the duplicate lines, 
    
    reader.ReadData(instr);
    auto variants = reader.GetVariantsForChr("1");
    BOOST_CHECK(variants->GetNumberOfVariants() == 14);
    TSeqRange range(variants->GetStart(), variants->GetStop());
    for (CVCFVariant_CI var_iter(*variants, range); var_iter; ++var_iter) {
        string var_id = var_iter.GetVariantID();
        if (var_id == "nssv1749199") {
            BOOST_CHECK(var_iter.GetInfoField("DESC") == "lineage specific duplication -  fixed_Ggod-Ggog");
        }
        else if (var_id == "nssv1749212") {
            BOOST_CHECK(var_iter.GetInfoField("DESC") == "lineage specific,comma_here -  fixed");
        }
        else if (var_id == "nssv1749256") {
            BOOST_CHECK(var_iter.GetInfoField("DESC") == "lineage specific=456 and -  fixed_Ggod");
        }
    }
}
