/*  $Id: vcf_assembly_sel_panel.cpp 45881 2021-01-05 03:10:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladislav Evgeniev
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/loaders/vcf_assembly_sel_panel.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/compressed_file.hpp>
#include <gui/utils/job_future.hpp>

#include <corelib/ncbifile.hpp>
#include <wx/dirdlg.h>
#include <wx/filename.h>
#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/statbox.h>

#include <gui/widgets/loaders/columnar_vcf_reader.hpp>
#include <gui/widgets/loaders/assembly_sel_panel.hpp>
#include <gui/widgets/loaders/map_assembly_params.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/*!
 * CVcfAssemblySelectionPanel type definition
 */

IMPLEMENT_DYNAMIC_CLASS( CVcfAssemblySelectionPanel, wxPanel )


/*!
 * CVcfAssemblySelectionPanel event table definition
 */

BEGIN_EVENT_TABLE( CVcfAssemblySelectionPanel, wxPanel )

////@begin CVcfAssemblySelectionPanel event table entries
    EVT_IDLE( CVcfAssemblySelectionPanel::OnIdle )
////@end CVcfAssemblySelectionPanel event table entries

END_EVENT_TABLE()


/*!
 * CVcfAssemblySelectionPanel constructors
 */

CVcfAssemblySelectionPanel::CVcfAssemblySelectionPanel()
{
    Init();
}

CVcfAssemblySelectionPanel::CVcfAssemblySelectionPanel( wxWindow* parent, CColumnarVCFReader* vcf_reader,
    const wxString &input_vcf, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
    : m_VCFReader(vcf_reader), m_InputVcf(input_vcf)
{
    Init();
    Create(parent, id, pos, size, style);
}


/*!
 * CVcfAssemblySelectionPanel creator
 */

bool CVcfAssemblySelectionPanel::Create( wxWindow* parent,
    wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
////@begin CVcfAssemblySelectionPanel creation
    SetExtraStyle(wxWS_EX_VALIDATE_RECURSIVELY);
    wxPanel::Create( parent, id, pos, size, style );

    CreateControls();
    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }
    Centre();
////@end CVcfAssemblySelectionPanel creation
    return true;
}


/*!
 * CVcfAssemblySelectionPanel destructor
 */

CVcfAssemblySelectionPanel::~CVcfAssemblySelectionPanel()
{
////@begin CVcfAssemblySelectionPanel destruction
////@end CVcfAssemblySelectionPanel destruction
}


/*!
 * Member initialisation
 */

void CVcfAssemblySelectionPanel::Init()
{
////@begin CVcfAssemblySelectionPanel member initialisation
    m_VcfAssembly = NULL;
    m_AssemblyPanel = NULL;
////@end CVcfAssemblySelectionPanel member initialisation
}


/*!
 * Control creation for CVcfAssemblySelectionPanel
 */

void CVcfAssemblySelectionPanel::CreateControls()
{    
////@begin CVcfAssemblySelectionPanel content construction
    // Generated by DialogBlocks, 25/09/2020 13:45:56 (unregistered)

    CVcfAssemblySelectionPanel* itemPanel1 = this;

    wxStaticBox* itemStaticBoxSizer2Static = new wxStaticBox(itemPanel1, wxID_ANY, _("Assembly Options"));
    wxStaticBoxSizer* itemStaticBoxSizer2 = new wxStaticBoxSizer(itemStaticBoxSizer2Static, wxVERTICAL);
    itemPanel1->SetSizer(itemStaticBoxSizer2);

    wxStaticBox* itemStaticBoxSizer3Static = new wxStaticBox(itemStaticBoxSizer2->GetStaticBox(), wxID_ANY, _("VCF File Assembly Identifier"));
    wxStaticBoxSizer* itemStaticBoxSizer3 = new wxStaticBoxSizer(itemStaticBoxSizer3Static, wxHORIZONTAL);
    itemStaticBoxSizer2->Add(itemStaticBoxSizer3, 0, wxGROW|wxALL, 5);

    m_VcfAssembly = new wxStaticText( itemStaticBoxSizer3->GetStaticBox(), wxID_STATIC, _("N/A"), wxDefaultPosition, wxSize(213, -1), wxALIGN_LEFT|wxSTATIC_BORDER );
    itemStaticBoxSizer3->Add(m_VcfAssembly, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);
    x_ReadAssembly();

    wxBoxSizer* itemBoxSizer5 = new wxBoxSizer(wxHORIZONTAL);
    itemStaticBoxSizer2->Add(itemBoxSizer5, 0, wxGROW|wxLEFT|wxRIGHT, 5);

    m_AssemblyPanel = new CAssemblySelPanel( itemStaticBoxSizer2->GetStaticBox(), ID_PANEL9, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL );
    itemBoxSizer5->Add(m_AssemblyPanel, 1, wxALIGN_CENTER_VERTICAL|wxTOP|wxBOTTOM, 5);

////@end CVcfAssemblySelectionPanel content construction
    m_AssemblyPanel->SetUseMapping();
    m_AssemblyPanel->EnableUseMapping(false);
}

static string s_ReadVCFAssembly(CColumnarVCFReader* reader, const wxString& input_vcf, ICanceled& canceled)
{
    string vcf_assm;
    if (!reader || canceled.IsCanceled())
        return vcf_assm;

    try {
        CCompressedFile file(input_vcf);
        reader->ReadHeader(file.GetIstream(false), &canceled);
        vcf_assm = reader->GetAssembly();
    }
    catch (const CException& e) {
        LOG_POST(Error << "s_ReadVCFAssembly failed: " << e.GetMsg());
    }
    catch (const exception& e) {
        LOG_POST(Error << "s_ReadVCFAssembly failed: " << e.what());
    }
    return vcf_assm;
}

void CVcfAssemblySelectionPanel::x_ReadAssembly()
{
    if (!m_VCFReader)
        return;
    m_VcfAssemblyFuture = job_async(bind(s_ReadVCFAssembly, m_VCFReader, m_InputVcf, placeholders::_1), 
        "Retrieve Assembly name/accession present in VCF file");

    m_VcfAssembly->SetLabel(wxT("Pending..."));
    m_VcfAssembly->SetToolTip(wxT("VCF Assembly tag is being retrieved..."));
}

bool CVcfAssemblySelectionPanel::IsInputValid()
{
    return m_AssemblyPanel->TransferDataFromWindow();
}

void CVcfAssemblySelectionPanel::SetNewFile(const wxString &input_vcf)
{
    m_InputVcf = input_vcf;
    x_ReadAssembly();
}
/*!
 * Should we show tooltips?
 */

bool CVcfAssemblySelectionPanel::ShowToolTips()
{
    return true;
}

/*!
 * Get bitmap resources
 */

wxBitmap CVcfAssemblySelectionPanel::GetBitmapResource(const wxString& name)
{
    // Bitmap retrieval
////@begin CVcfAssemblySelectionPanel bitmap retrieval
    wxUnusedVar(name);
    return wxNullBitmap;
    ////@end CVcfAssemblySelectionPanel bitmap retrieval
}

/*!
 * Get icon resources
 */

wxIcon CVcfAssemblySelectionPanel::GetIconResource(const wxString& name)
{
    // Icon retrieval
////@begin CVcfAssemblySelectionPanel icon retrieval
    wxUnusedVar(name);
    return wxNullIcon;
    ////@end CVcfAssemblySelectionPanel icon retrieval
}


/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_BROWSEBUTTON
 */


static const char* kMappingAssembly = ".MapAssembly";

void CVcfAssemblySelectionPanel::LoadSettings()
{
    if (!m_RegPath.empty()) {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CRegistryReadView view = gui_reg.GetReadView(m_RegPath + kMappingAssembly);

        m_AssemblyPanel->GetData().LoadSettings(view);
    }
}

void CVcfAssemblySelectionPanel::SaveSettings() const
{
    if (!m_RegPath.empty()) {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CRegistryWriteView view = gui_reg.GetWriteView(m_RegPath + kMappingAssembly);

        m_AssemblyPanel->GetData().SaveSettings(view);
    }
}


/*!
 * wxEVT_IDLE event handler for ID_CVcfAssemblySelectionPanel
 */

void CVcfAssemblySelectionPanel::OnIdle(wxIdleEvent& event)
{
    if (m_VcfAssemblyFuture.IsComplete()) {
        try {
            CMapAssemblyParams params;
            params.SetUseMapping(true);

            string vcf_assembly = m_VcfAssemblyFuture();
            if (!vcf_assembly.empty()) {
                wxString str = ToWxString(vcf_assembly);
                m_VcfAssembly->SetLabel(str);
                m_VcfAssembly->SetToolTip(str);
                
                params.SetSearchTerm(vcf_assembly);
            }
            else {
                m_VcfAssembly->SetLabel(wxT("N/A"));
                m_VcfAssembly->SetToolTip(wxEmptyString);
            }
            m_AssemblyPanel->SetData(params);
            m_AssemblyPanel->TransferDataToWindow();
        }
        catch (const exception&) {
        }
        m_VcfAssemblyFuture.reset(); // not neccessary but recommended to release resources
    }
    else {
        event.Skip();
    }
}


END_NCBI_SCOPE




