/*  $Id: vcf_column_sel_panel.cpp 45881 2021-01-05 03:10:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Andrea Asztalos
 */


#include <ncbi_pch.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/checkedlistctrl.hpp>
#include <gui/widgets/loaders/columnar_vcf_reader.hpp>
#include <gui/widgets/loaders/vcf_column_sel_panel.hpp>
#include <wx/statbox.h>
#include <wx/sizer.h>
#include <wx/button.h>
#include <wx/wupdlock.h>
#include <wx/stattext.h>


BEGIN_NCBI_SCOPE

/*!
 * CVcfColumnSelectionPanel type definition
 */

IMPLEMENT_DYNAMIC_CLASS( CVcfColumnSelectionPanel, wxPanel )


/*!
 * CVcfColumnSelectionPanel event table definition
 */

BEGIN_EVENT_TABLE( CVcfColumnSelectionPanel, wxPanel )
    EVT_BUTTON(ID_SelectAllINFO, CVcfColumnSelectionPanel::OnSelectAllClick)
    EVT_BUTTON(ID_DeselectAllINFO, CVcfColumnSelectionPanel::OnDeselectAllClick)
END_EVENT_TABLE()


/*!
 * CVcfColumnSelectionPanel constructors
 */

CVcfColumnSelectionPanel::CVcfColumnSelectionPanel()
{
    Init();
}

CVcfColumnSelectionPanel::CVcfColumnSelectionPanel( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
    Init();
    Create(parent, id, pos, size, style);
}


/*!
 * CVcfColumnSelectionPanel creator
 */

bool CVcfColumnSelectionPanel::Create( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
////@begin CVcfColumnSelectionPanel creation
    SetExtraStyle(wxWS_EX_VALIDATE_RECURSIVELY);
    wxPanel::Create( parent, id, pos, size, style );

    CreateControls();
    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }
    Centre();
////@end CVcfColumnSelectionPanel creation
    return true;
}


/*!
 * CVcfColumnSelectionPanel destructor
 */

CVcfColumnSelectionPanel::~CVcfColumnSelectionPanel()
{
////@begin CVcfColumnSelectionPanel destruction
////@end CVcfColumnSelectionPanel destruction
}


/*!
 * Member initialisation
 */

void CVcfColumnSelectionPanel::Init()
{
////@begin CVcfColumnSelectionPanel member initialisation
    m_InfoSubCols = NULL;
////@end CVcfColumnSelectionPanel member initialisation
}


/*!
 * Control creation for CVcfColumnSelectionPanel
 */

void CVcfColumnSelectionPanel::CreateControls()
{    
////@begin CVcfColumnSelectionPanel content construction
    CVcfColumnSelectionPanel* itemPanel1 = this;

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxVERTICAL);
    itemPanel1->SetSizer(itemBoxSizer2);

    wxStaticText* itemStaticText1 = new wxStaticText( itemPanel1, wxID_STATIC, _("INFO Fields"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer2->Add(itemStaticText1, 0, wxALIGN_LEFT|wxALL, 5);

    m_InfoSubCols = new wxCheckedListCtrl(itemPanel1, ID_INFOFIELDSLIST, wxDefaultPosition, wxDefaultSize, wxLC_REPORT);
    itemBoxSizer2->Add(m_InfoSubCols, 1, wxGROW | wxALL, 5);
    m_InfoSubCols->InsertColumn(0, wxT("Name"));
    m_InfoSubCols->InsertColumn(1, wxT("Description"));

    wxBoxSizer* itemBoxSizer3 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer2->Add(itemBoxSizer3, 0, wxALIGN_LEFT|wxTOP, 5);

    wxButton* itemButton4 = new wxButton( itemPanel1, ID_SelectAllINFO, _("Select All"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemButton4, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxButton* itemButton5 = new wxButton( itemPanel1, ID_DeselectAllINFO, _("Deselect All"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemButton5, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

////@end CVcfColumnSelectionPanel content construction
}

void CVcfColumnSelectionPanel::SetInfoFields(const set<CConstRef<SVcfFieldData>>& fields)
{
    wxWindowUpdateLocker upd(m_InfoSubCols);
    m_InfoSubCols->DeleteAllItems();

    m_FieldData = fields;
    int row = 0;
    for (const auto& it : m_FieldData) {

        // Insert a row
        wxListItem item;
        item.SetId(row);
        m_InfoSubCols->InsertItem(item);

        // Column to indicate selected or not
        m_InfoSubCols->SetChecked(row, it->m_Number != ".");
        m_InfoSubCols->SetItem(row, 0, it->m_Name);

        m_InfoSubCols->SetItem(row, 1, it->m_Description);

        ++row;
    }
    m_InfoSubCols->SetColumnWidth(0, wxLIST_AUTOSIZE_USEHEADER);
    m_InfoSubCols->SetColumnWidth(1, wxLIST_AUTOSIZE);
}

void CVcfColumnSelectionPanel::GetSelectedInfoFields(set<string>& selected)
{
    selected.clear();
    int row = 0;
    for (const auto& it : m_FieldData) {
        if (m_InfoSubCols->IsChecked(row)) {
            selected.insert(it->m_Name);
        }
        ++row;
    }
}

bool CVcfColumnSelectionPanel::AllAreChecked() const
{
    bool checked = true;
    int row = 0;
    for (auto it = m_FieldData.begin(); checked && it != m_FieldData.end(); ++it) {
        checked = m_InfoSubCols->IsChecked(row);
        ++row;
    }
    return checked;
}

/*!
 * Should we show tooltips?
 */

bool CVcfColumnSelectionPanel::ShowToolTips()
{
    return true;
}

/*!
 * Get bitmap resources
 */

wxBitmap CVcfColumnSelectionPanel::GetBitmapResource( const wxString& name )
{
    // Bitmap retrieval
////@begin CVcfColumnSelectionPanel bitmap retrieval
    wxUnusedVar(name);
    return wxNullBitmap;
////@end CVcfColumnSelectionPanel bitmap retrieval
}

/*!
 * Get icon resources
 */

wxIcon CVcfColumnSelectionPanel::GetIconResource( const wxString& name )
{
    // Icon retrieval
////@begin CVcfColumnSelectionPanel icon retrieval
    wxUnusedVar(name);
    return wxNullIcon;
////@end CVcfColumnSelectionPanel icon retrieval
}

/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_SelectAllINFO
 */

void CVcfColumnSelectionPanel::OnSelectAllClick(wxCommandEvent& event)
{
    int row = 0;
    for (const auto& it : m_FieldData) {
        m_InfoSubCols->SetChecked(row, true);
        ++row;
    }
}


/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_DeselectAllINFO
 */

void CVcfColumnSelectionPanel::OnDeselectAllClick(wxCommandEvent& event)
{
    int row = 0;
    for (const auto& it : m_FieldData) {
        m_InfoSubCols->SetChecked(row, false);
        ++row;
    }
}

END_NCBI_SCOPE

