/*  $Id: vcf_ref_sequences_panel.cpp 46239 2021-02-16 23:00:17Z evgeniev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Vladislav Evgeniev 
 */


#include <ncbi_pch.hpp>
#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/icon.h>
#include <wx/utils.h>
#include <wx/button.h>
#include <wx/wupdlock.h>
#include <wx/radiobox.h>

#include <gui/objutils/label.hpp>
#include <gui/objutils/utils.hpp>
#include <objmgr/scope.hpp>
#include <objmgr/object_manager.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/async_call.hpp>
#include <gui/widgets/loaders/vcf_ref_sequences_panel.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/*!
 * CVcfRefSequencesPanel type definition
 */

IMPLEMENT_DYNAMIC_CLASS( CVcfRefSequencesPanel, wxPanel )


/*!
 * CVcfRefSequencesPanel event table definition
 */

BEGIN_EVENT_TABLE( CVcfRefSequencesPanel, wxPanel )

////@begin CVcfRefSequencesPanel event table entries
    EVT_IDLE( CVcfRefSequencesPanel::OnIdle )
    EVT_RADIOBOX( ID_RADIOBOX, CVcfRefSequencesPanel::OnRadioboxSelected )
    EVT_CHECKLISTBOX( ID_CHECKLISTBOX, CVcfRefSequencesPanel::OnChecklistboxToggled )
    EVT_BUTTON( ID_SELECT_ALL_BTN, CVcfRefSequencesPanel::OnSelectAllBtnClick )
    EVT_BUTTON( ID_DESELECT_ALL_BTN, CVcfRefSequencesPanel::OnDeselectAllBtnClick )
    EVT_SEARCHCTRL_SEARCH_BTN( ID_SEARCHCTRL, CVcfRefSequencesPanel::OnSearchButtonClick )
    EVT_SEARCHCTRL_CANCEL_BTN( ID_SEARCHCTRL, CVcfRefSequencesPanel::OnSearchCancelClick )
    EVT_TEXT_ENTER( ID_SEARCHCTRL, CVcfRefSequencesPanel::OnSearchEnter )
    EVT_UPDATE_UI( ID_SEARCHCTRL, CVcfRefSequencesPanel::OnSearchCtrlUpdate )
////@end CVcfRefSequencesPanel event table entries

END_EVENT_TABLE()


/*!
 * CVcfRefSequencesPanel constructors
 */

CVcfRefSequencesPanel::CVcfRefSequencesPanel()
{
    Init();
}

CVcfRefSequencesPanel::CVcfRefSequencesPanel( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
    Init();
    Create(parent, id, pos, size, style);
}


/*!
 * CVcfRefSequencesPanel creator
 */

bool CVcfRefSequencesPanel::Create( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
////@begin CVcfRefSequencesPanel creation
    SetExtraStyle(wxWS_EX_VALIDATE_RECURSIVELY);
    wxPanel::Create( parent, id, pos, size, style );

    CreateControls();
    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }
    Centre();
////@end CVcfRefSequencesPanel creation
    return true;
}


/*!
 * CVcfRefSequencesPanel destructor
 */

CVcfRefSequencesPanel::~CVcfRefSequencesPanel()
{
////@begin CVcfRefSequencesPanel destruction
////@end CVcfRefSequencesPanel destruction
}


/*!
 * Member initialisation
 */

void CVcfRefSequencesPanel::Init()
{
////@begin CVcfRefSequencesPanel member initialisation
    m_LoadSequencesRadio = NULL;
    m_RefSeqList = NULL;
    m_SelectAllBtn = NULL;
    m_DeselctAllBtn = NULL;
    m_SearchCtrl = NULL;
    m_SeqSelectText = NULL;
////@end CVcfRefSequencesPanel member initialisation
}


/*!
 * Control creation for CVcfRefSequencesPanel
 */

void CVcfRefSequencesPanel::CreateControls()
{    
////@begin CVcfRefSequencesPanel content construction
    // Generated by DialogBlocks, 14/12/2020 16:33:26 (unregistered)

    CVcfRefSequencesPanel* itemPanel1 = this;

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxVERTICAL);
    itemPanel1->SetSizer(itemBoxSizer2);

    wxArrayString m_LoadSequencesRadioStrings;
    m_LoadSequencesRadioStrings.Add(_("&Load all sequences referenced in the file (may take a while if many sequences are referenced)"));
    m_LoadSequencesRadioStrings.Add(_("L&oad selected sequences only"));
    m_LoadSequencesRadio = new wxRadioBox( itemPanel1, ID_RADIOBOX, _("Select sequences for which to load variation data"), wxDefaultPosition, wxDefaultSize, m_LoadSequencesRadioStrings, 1, wxRA_SPECIFY_COLS );
    m_LoadSequencesRadio->SetSelection(0);
    itemBoxSizer2->Add(m_LoadSequencesRadio, 0, wxGROW|wxALL, 5);

    m_SeqSelectText = new wxStaticText( itemPanel1, ID_SEQ_SELECT_TEXT, _("The following chromosomes are part of the assembly. Please confirm the sequences to add to the project, or search for scaffolds not on the list. "), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer2->Add(m_SeqSelectText, 0, wxALIGN_LEFT|wxALL, 5);

    wxArrayString m_RefSeqListStrings;
    m_RefSeqList = new wxCheckListBox( itemPanel1, ID_CHECKLISTBOX, wxDefaultPosition, wxDefaultSize, m_RefSeqListStrings, wxLB_SINGLE );
    m_RefSeqList->Enable(false);
    itemBoxSizer2->Add(m_RefSeqList, 1, wxGROW|wxALL, 5);

    wxBoxSizer* itemBoxSizer4 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer2->Add(itemBoxSizer4, 0, wxGROW|wxTOP, 5);

    m_SelectAllBtn = new wxButton( itemPanel1, ID_SELECT_ALL_BTN, _("Select All"), wxDefaultPosition, wxDefaultSize, 0 );
    m_SelectAllBtn->Enable(false);
    itemBoxSizer4->Add(m_SelectAllBtn, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_DeselctAllBtn = new wxButton( itemPanel1, ID_DESELECT_ALL_BTN, _("Deselect All"), wxDefaultPosition, wxDefaultSize, 0 );
    m_DeselctAllBtn->Enable(false);
    itemBoxSizer4->Add(m_DeselctAllBtn, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    itemBoxSizer4->Add(5, 5, 1, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText8 = new wxStaticText( itemPanel1, wxID_STATIC, _("Search/Add:"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer4->Add(itemStaticText8, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_SearchCtrl = new wxSearchCtrl( itemPanel1, ID_SEARCHCTRL, wxEmptyString, wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER );
    m_SearchCtrl->Enable(false);
    itemBoxSizer4->Add(m_SearchCtrl, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

////@end CVcfRefSequencesPanel content construction
}


/*!
 * Should we show tooltips?
 */

bool CVcfRefSequencesPanel::ShowToolTips()
{
    return true;
}

/*!
 * Get bitmap resources
 */

wxBitmap CVcfRefSequencesPanel::GetBitmapResource( const wxString& name )
{
    // Bitmap retrieval
////@begin CVcfRefSequencesPanel bitmap retrieval
    wxUnusedVar(name);
    return wxNullBitmap;
////@end CVcfRefSequencesPanel bitmap retrieval
}

/*!
 * Get icon resources
 */

wxIcon CVcfRefSequencesPanel::GetIconResource( const wxString& name )
{
    // Icon retrieval
////@begin CVcfRefSequencesPanel icon retrieval
    wxUnusedVar(name);
    return wxNullIcon;
////@end CVcfRefSequencesPanel icon retrieval
}

void CVcfRefSequencesPanel::SetReferenceSeqIds(const vector<CConstRef<CSeq_id> >& refSeqIds)
{
    m_SeqSelectText->SetLabel(wxString::Format("The following %zu chromosomes are part of the assembly. Please confirm the sequences to add to the project, or search for scaffolds not on the list.", refSeqIds.size()));
    m_SeqSelectText->GetContainingSizer()->Fit(m_SeqSelectText);
    Layout();
    m_RefSeqIds = refSeqIds;
    for (auto& seq_id : refSeqIds) {
        m_SelectedSeqIds.insert(seq_id.GetPointerOrNull());
    }

    if (!m_Scope) {
        CRef<CObjectManager> obj_mgr = CObjectManager::GetInstance();
        m_Scope.Reset(new CScope(*obj_mgr));
        m_Scope->AddDefaults();
    }

    x_FillList(NcbiEmptyString);
}

void CVcfRefSequencesPanel::SetTopLevelSeqCount(size_t count)
{
    _ASSERT(m_LoadSequencesRadio);
    wxString label = wxString::Format("&Load all sequences referenced in the file (may take a while if many sequences are referenced; up to %zu known for this assembly)", count);
    m_LoadSequencesRadio->SetString(0, label);
}

void CVcfRefSequencesPanel::GetSelectedReferenceSeqIds(vector<CConstRef<CSeq_id> >& selSeqIds) const
{
    selSeqIds.clear();
    for (auto seq_id : m_SelectedSeqIds) {
        CConstRef<CSeq_id> seqId(seq_id);
        selSeqIds.push_back(seqId);
    }
}

bool CVcfRefSequencesPanel::ImportAll() const
{
    return (m_LoadSequencesRadio->GetSelection() == 0);
}

void CVcfRefSequencesPanel::SetGenCollAssembly(const CRef<objects::CGC_Assembly>& gc_assm)
{
    m_GenCollAssm = gc_assm;
}

void CVcfRefSequencesPanel::OnSelectAllBtnClick(wxCommandEvent& event)
{
    for (unsigned i = 0; i < m_RefSeqList->GetCount(); ++i) {
        m_RefSeqList->Check(i);
        m_SelectedSeqIds.insert((const CSeq_id*)m_RefSeqList->GetClientData(i));
    }
    x_FillList(NcbiEmptyString);
}

void CVcfRefSequencesPanel::OnDeselectAllBtnClick(wxCommandEvent& event)
{
    for (unsigned i = 0; i < m_RefSeqList->GetCount(); ++i)
        m_RefSeqList->Check(i, false);
    m_SelectedSeqIds.clear();
    x_FillList(NcbiEmptyString);
}

void CVcfRefSequencesPanel::OnSearchButtonClick(wxCommandEvent& event)
{
    x_FillList(string(event.GetString().ToUTF8()));
}

void CVcfRefSequencesPanel::OnSearchCancelClick(wxCommandEvent& event)
{
    x_FillList(NcbiEmptyString);
}

void CVcfRefSequencesPanel::OnSearchEnter(wxCommandEvent& event)
{
    x_FillList(string(m_SearchCtrl->GetValue().ToUTF8()));
}

void CVcfRefSequencesPanel::x_FillList(const string& filter)
{
    m_IdsWithNoDescriptions.clear();

    m_CurrentFilter = filter;
    NStr::TruncateSpacesInPlace(m_CurrentFilter);

    sort(m_RefSeqIds.begin(), m_RefSeqIds.end(), [](const CConstRef<CSeq_id>& a, const CConstRef<CSeq_id>& b) {
        int a_local = a->IsLocal() ? 1 : 0;
        int b_local = b->IsLocal() ? 1 : 0;
        if (a_local != b_local)
            return a_local < b_local;
        return a->GetSeqIdString(true) < b->GetSeqIdString(true);
    });

    // TODO: Resolve flickering of the list
    m_RefSeqList->Hide();
    wxWindowUpdateLocker upd(m_RefSeqList);
    m_RefSeqList->Clear();

    for (const auto& id : m_RefSeqIds) {
        string seqIdLabel(id->GetSeqIdString(true));
        if (!filter.empty() && NStr::FindNoCase(seqIdLabel, m_CurrentFilter) == NPOS)
            continue;

        int newItem = m_RefSeqList->Append(seqIdLabel, (void*)id.GetPointerOrNull());
        if (m_SelectedSeqIds.count(id.GetPointerOrNull()) > 0) {
            m_RefSeqList->Check(newItem);
        }
        if (!id->IsLocal()) {
            m_IdsWithNoDescriptions.push_back(newItem);
        }
    }
    if (m_RefSeqList->GetCount() == 0) {
        if (m_GenCollAssm && !m_IdMapper) {
            GUI_AsyncExec([this](ICanceled& canceled)
            {
                m_IdMapper.reset(CGencollIdMapperAdapter::GetIdMapper(m_GenCollAssm));
            }, wxT("Querying GenColl..."));
        }
        auto seq_id = CSeqUtils::MapStringId(filter, m_IdMapper.get());
        if (m_IdMapper && seq_id->IsLocal()) {
            m_RefSeqList->Show();
            return;
        }
        string seqIdLabel(filter);
        seqIdLabel += " (";
        seqIdLabel += seq_id->GetSeqIdString(true);
        seqIdLabel += ')';
        int newItem = m_RefSeqList->Append(seqIdLabel, (void*)seq_id.GetPointerOrNull());
        m_RefSeqList->Check(newItem);
        m_RefSeqIds.push_back(seq_id);
        m_SelectedSeqIds.insert(seq_id.GetPointerOrNull());
    }
    m_RefSeqList->Show();
}

void CVcfRefSequencesPanel::OnSearchCtrlUpdate(wxUpdateUIEvent& event)
{
    m_SearchCtrl->ShowCancelButton(!m_CurrentFilter.empty());
}


/*!
 * wxEVT_IDLE event handler for ID_CVCFRefSequencesPanel
 */

void CVcfRefSequencesPanel::OnIdle(wxIdleEvent& event)
{
    if (!m_IdsWithNoDescriptions.empty()) {
        auto item_id = m_IdsWithNoDescriptions.front();
        try {
            CConstRef<CSeq_id> id((CSeq_id*)m_RefSeqList->GetClientData(item_id));
            string seqIdDesc;
            CLabel::GetLabel(*id, &seqIdDesc, CLabel::eDescription, m_Scope);
            string seqIdLabel(id->GetSeqIdString(true));
            seqIdLabel += " (";
            seqIdLabel += seqIdDesc;
            seqIdLabel += ")";
            m_RefSeqList->SetString(item_id, ToWxString(seqIdLabel));
        }
        catch (exception& e) {
            ERR_POST(Error << e.what());
        }
        m_IdsWithNoDescriptions.pop_front();
    }
    event.Skip();
}

/*!
 * wxEVT_COMMAND_RADIOBOX_SELECTED event handler for ID_RADIOBOX
 */

void CVcfRefSequencesPanel::OnRadioboxSelected(wxCommandEvent& event)
{
    if (1 == m_LoadSequencesRadio->GetSelection()) {
        m_RefSeqList->Enable(true);
        m_SelectAllBtn->Enable(true);
        m_DeselctAllBtn->Enable(true);
        m_SearchCtrl->Enable(true);
    }
    else {
        m_RefSeqList->Enable(false);
        m_SelectAllBtn->Enable(false);
        m_DeselctAllBtn->Enable(false);
        m_SearchCtrl->Enable(false);
    }   
}

/*!
 * wxEVT_COMMAND_CHECKLISTBOX_TOGGLED event handler for ID_CHECKLISTBOX
 */

void CVcfRefSequencesPanel::OnChecklistboxToggled( wxCommandEvent& event )
{
    auto item = event.GetInt();
    if (m_RefSeqList->IsChecked(item)) {
        m_SelectedSeqIds.insert((const CSeq_id*)m_RefSeqList->GetClientData(item));
    }
    else {
        auto it = m_SelectedSeqIds.find((const CSeq_id*)m_RefSeqList->GetClientData(item));
        m_SelectedSeqIds.erase(it);
    }
}

END_NCBI_SCOPE
