/* 
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrei Shkeda
 *
 * File Description:
 *    CIntronGlyph -- utility class representing intron feature
 *                
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/intron_glyph.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/general/User_object.hpp>
#include <algo/sequence/consensus_splice.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


// CIntronGlyph::CIntronGlyph()
//
CIntronGlyph::CIntronGlyph(const CMappedFeat& f, ELinkedFeatDisplay LinkedFeatDisplay)
    : CFeatGlyph(f, LinkedFeatDisplay)
{
    x_Init();
}

CIntronGlyph::CIntronGlyph(const CMappedFeat& f, const CSeq_loc& loc, ELinkedFeatDisplay LinkedFeatDisplay)
    : CFeatGlyph(f, loc,LinkedFeatDisplay)
{
    x_Init();
}

void CIntronGlyph::x_Init()
{
    if (GetFeature().IsSetExts()) {
        const CSeq_feat::TExts& exts = GetFeature().GetExts();
        ITERATE (CSeq_feat::TExts, iter, exts) {
            if ((*iter)->GetType().IsStr()  &&
                (*iter)->GetType().GetStr() == "Support"  &&
                (*iter)->GetFieldRef("Total") &&
                !(*iter)->GetFieldRef("Total").IsNull()) {
                const auto& d = (*iter)->GetFieldRef("Total")->GetData();
                if (d.IsInt())
                    m_ReadCount = d.GetInt();
                else if (d.IsReal()) 
                    m_ReadCount = float(d.GetReal() + 0.5);
            }
        }
    }
}

bool CIntronGlyph::IsConsensus() const
{
    // return true if intron is consensus or if data to check consensus (splice sequence)
    // is unavailable.
    if (GetFeature().IsSetExts()) {
        const CSeq_feat::TExts& exts = GetFeature().GetExts();
        ITERATE (CSeq_feat::TExts, iter, exts) {
            if ((*iter)->GetType().IsStr() && NStr::EqualNocase((*iter)->GetType().GetStr(), "Splice Sequence")) {
                    CConstRef<CUser_field> donor = (*iter)->GetFieldRef("Donor");
                    CConstRef<CUser_field> acceptor = (*iter)->GetFieldRef("Acceptor");
                    if (!donor.IsNull() && donor->GetData().IsStr() && !acceptor.IsNull() && acceptor->GetData().IsStr()) {
                        const string& donor_str =  donor->GetData().GetStr();
                        const string& acceptor_str =  acceptor->GetData().GetStr();
                        return IsConsensusSplice(donor_str, acceptor_str);
                    }
                    else return true;
            }
        }
    }
    return true;
}

void CIntronGlyph::GetAccessoryLabel(string& accessory_label) const
{
    accessory_label = NStr::IntToString(m_ReadCount);
}



END_NCBI_SCOPE
