/*  $Id: vcf_glyph.cpp 45881 2021-01-05 03:10:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio, Liangshou Wu
 *
 * File Description:
 *    CVcfGlyph 
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/vcf_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/seq_graphic/layout_track_impl.hpp>
#include <gui/widgets/seq_graphic/layout_group.hpp>
#include <gui/widgets/seq_graphic/config_utils.hpp>

#include <gui/widgets/gl/ruler.hpp>
#include <gui/objutils/obj_fingerprint.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/gui_object_info_seq_feat.hpp>
#include <gui/objutils/tooltip.hpp>
#include <gui/objutils/snp_gui.hpp>
#include <gui/opengl/gltexturefont.hpp>
#include <gui/opengl/irender.hpp>

#include <objects/seqfeat/Variation_ref.hpp>
#include <objects/seqfeat/Variation_inst.hpp>
#include <objects/seqfeat/Delta_item.hpp>
#include <objects/seqfeat/Feat_id.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/general/User_object.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objects/seq/Seq_literal.hpp>
#include <objects/seqloc/Giimport_id.hpp>

#include <objmgr/util/sequence.hpp>
#include <objtools/alnmgr/aln_converters.hpp>
#include <objtools/snputil/snp_utils.hpp>
#include <util/sequtil/sequtil_manip.hpp>

#include <util/sequtil/sequtil_manip.hpp>
#include <corelib/ncbi_stack.hpp>
#include <math.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/// feature ruler height.
static const int kRulerHeight = 16;

/// vertical space between elements.
static const int kVertSpace = 2;

/// Restriction sites marker height (relative to the height of the bar)
static const TModelUnit kRSiteMarkerHeight = 0.5;

/// Restriction sites marker width (absolute)
static const TModelUnit kRSiteMarkerWidth = 0.4;

/// Restriction sites marker visibility threshold (ie if the marker's width is smaller then the threshold, markers are not drawn)
static const TModelUnit kRSiteMarkerThreshold = 2;

/// Shoudl match feature_ds.cpp::kMinScaleForMapping 
/// i.e. what mapping is not created then do not show ruler labels
static const float kMinScaleForRulerLabels = 16.;


// #define DEBUG_INFO_ON_SCREEN 1


//
// CVcfGlyph::CVcfGlyph()
//

CVcfGlyph::CVcfGlyph(CSeq_id& seq_id,
	                 const string& sID, 
                     size_t pos, 
                     const string& ref_allele, 
                     const string& alt_alleles) :
	 m_Object(new CVcfVariant(seq_id, sID, pos, ref_allele, alt_alleles))
{
	m_Intervals.push_back(GetRange());
}


bool CVcfGlyph::NeedTooltip(const TModelPoint& /*p*/, ITooltipFormatter& /*tt*/, string& t_title) const
{
    return true;
}


void CVcfGlyph::GetTooltip(const TModelPoint& p, ITooltipFormatter& tooltip, string& t_title) const
{
    t_title = m_Object->GetID();
    tooltip.AddRow("Variation ID:", m_Object->GetID());
    tooltip.AddRow("Alleles:", m_Object->GetAllAlleles_display());
	bool all_alt_alleles_same_len(true);
	int allele_len = m_Object->GetLongestAlleleLen(all_alt_alleles_same_len);
    if (allele_len > 0) {
        tooltip.AddRow(string(all_alt_alleles_same_len ? "A" : "Longest a") + "llele length:", NStr::NumericToString(allele_len, NStr::fWithCommas));
    }
    tooltip.AddRow("Variant Type:", m_Object->GetVariationTypeAsString());
}


TSeqRange CVcfGlyph::GetRange(void) const
{
    TSeqRange SeqRange(m_Object->GetPos(), m_Object->GetPos() + m_Object->GetLen() - 1);
    return SeqRange;
}


bool CVcfGlyph::GetCustomColor(CRgbaColor& color) const
{
    CGuiRegistry& registry(CGuiRegistry::GetInstance());
//        string sColorTheme(x_GetGlobalConfig()->GetColorTheme());
    string sColorTheme("Color");
    CRegistryReadView ColorView(CSGConfigUtils::GetColorReadView(registry, "GBPlugins.SnpTrack", "Default", sColorTheme));
    string sColorKey("Default");

    switch(m_Object->GetVariationType()) {
        case CVariation_inst::eType_snv:
            sColorKey = "SingleBase";
            break;
        case CVariation_inst::eType_mnp:
            sColorKey = "MultiBase";
            break;
        case CVariation_inst::eType_ins:
            sColorKey = "Insertion";
            break;
        case CVariation_inst::eType_del:
            sColorKey = "Deletion";
            break;
        case CVariation_inst::eType_delins:
            sColorKey = "Dips";
            break;
        default:
            break;
    }
    if(sColorKey != "Default") {
        CSGConfigUtils::GetColor(ColorView, sColorKey, color);
        return true;
    } 
    return false;
}


void CVcfGlyph::GetLabel(string& label, CLabel::ELabelType label_type) const
{
	m_Object->GetLabel(label, label_type);
}

void CVcfGlyph::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
//    string sLabel;
//    GetLabel(sLabel, CLabel::eContent);
//    LOG_POST(Trace << "<<<< CVcfGlyph::GetHTMLActiveAreas() " << this << " \"" << sLabel << " isDbVar: " << IsDbVar(GetFeature()));

    CHTMLActiveArea area;
    CSeqGlyph::x_InitHTMLActiveArea(area);
    area.m_PositiveStrand = true;

    area.m_Signature = GetSignature();
    
    // a tooltip should be generated for VCF records  to avoid an additional roundtrip and reloading of VCF blob
    string s;
    string title;
    CIRef<ITooltipFormatter> tooltip = ITooltipFormatter::CreateTooltipFormatter(ITooltipFormatter::eTooltipFormatter_CSSTable);
    GetTooltip(TModelPoint(-1, -1), *tooltip, title);
    s = tooltip->Render();
    string text = NStr::Replace(s, "\n", "<br/>");
    area.m_Descr = text;
    p_areas->push_back(area);
//        LOG_POST(Trace << ">>>> CVcfGlyph::GetHTMLActiveAreas() area with signature \"" << area.m_Signature << "\" generated for \"" << sLabel <<
//                            "\", rect: " << area.m_Bounds.Left() << ":" << area.m_Bounds.Right() << ":" << area.m_Bounds.Top() << ":" << area.m_Bounds.Bottom());
}


bool CVcfGlyph::IsClickable() const
{
    return true;
}


bool CVcfGlyph::SetSelected(bool f)
{
//    LOG_POST(Trace << "==== CVcfGlyph::SetSelected() " << this << " from " << IsSelected() << " to " << f);
    if (f != IsSelected()) {
        CSeqGlyph::SetSelected(f);
    }
    return false;
}

void CVcfGlyph::LayoutChanged()
{
    x_OnLayoutChanged();
}

void CVcfGlyph::x_DrawHighlight() const
{
	GetRenderingContext()->DrawHighlight(GetModelRect());
}


void CVcfGlyph::SetHideLabel(bool b)
{
    m_HideLabel = b;
}


bool CVcfGlyph::HasSideLabel() const
{
    return !m_HideLabel  &&  m_Config->m_LabelPos == CFeatureParams::ePos_Side;
}

/// some features may have additional info on the right (currently alleles for variations)
/// that will be shown at the same time as labels
void CVcfGlyph::GetAdditionalInfo(string& sAdditionalInfo) const
{
    sAdditionalInfo.clear();

    sAdditionalInfo = m_Object->GetRefAllele_display() + "/" + NStr::Join(m_Object->GetAltAlleles_display(), "/");
   
}

bool CVcfGlyph::x_isDrawn() const
{
    return true;
}

void CVcfGlyph::x_Draw() const
{
    _ASSERT(m_Context);
//    LOG_POST(Trace << "<<<<");

/*
// print drawing info
    {
        string sLabel;
        GetLabel(sLabel, CLabel::eContent);
        LOG_POST(Trace << "Attempting to draw " << this << " \"" << sLabel << "\" x_isDrawn: " << x_isDrawn() <<
                          ", m_Height: " << GetHeight() << ", m_Width: " << GetWidth() <<
                          ", m_Pos.m_X: " << GetLeft() << ", m_Pos.m_Y: " << GetTop());
    }
*/
    if (GetHeight() == 0  ||  m_Context->IntersectVisible(GetRange()).Empty() || !x_isDrawn()) {
//        LOG_POST(Trace << ">>>>");
        return;
    }
    // for debugging
    /*
    CScope& scope = GetMappedFeature().GetAnnot().GetScope();
    string sig = CObjFingerprint::GetFeatSignature(GetFeature(), &scope, m_sFilter);
    int subtype;
    CSeq_id_Handle id;
    TSeqPos f, t;
    CObjFingerprint::EObjectType type;
    Uint4   fingerprint;
    CObjFingerprint::ParseSignature(sig, id, f, t, type, subtype, fingerprint);
    cout << ", From: " << f
         << ", To: " << t
         << ", Subtype: " << subtype << endl;
    */
    TModelUnit base = GetTop();

    TModelUnit bar_height = x_GetBarHeight();
    x_MaybeDrawLabelAbove(base);
    base += bar_height * 0.5;   // bar center

    x_DrawVariationBar(base);

    x_MaybeDrawLabelOthers(base);
    x_DrawAdditionalInfo(base);

    if (IsSelected()) {
        m_Context->DrawSelection(GetModelRect());
    }
    // LOG_POST(Trace << ">>>>");
}

void CVcfGlyph::x_UpdateBoundingBox()
{
    _ASSERT(m_Context);
    // LOG_POST(Trace << "<<<<");

    if(x_isDrawn()) {
        TSeqRange SeqRange = GetRange();
        TModelUnit bar_height = x_GetBarHeight();
        SetHeight(bar_height);
        SetWidth(SeqRange.GetLength());
        SetLeft(SeqRange.GetFrom());

        bool showLabel = x_ShowLabel();

        // we preserve label space if it fits either on side or above
        if (showLabel) {
            IRender& gl = GetGl();

            const CGlTextureFont& font = m_Config->m_LabelFont;
            if (m_Config->m_LabelPos == CFeatureParams::ePos_Side) {
                string label;
                GetLabel(label, CLabel::eContent);
                TModelUnit text_w_px = min(gl.TextWidth(&font, label.c_str()), m_Context->GetMaxLabelWidth(font));
                TModelUnit text_w_pos = m_Context->ScreenToSeq(text_w_px + CRenderingContext::kLabelSpacePx);
                SetWidth(GetWidth() + text_w_pos);
                TModelUnit min_text_w_pos = m_Context->GetMinLabelWidthPos(font);

                if (x_LabelOnLeft()) {
                    SetLeft(GetLeft() - text_w_pos);
                    TModelRange visible_range_pos = m_Context->IntersectVisible(this);

                    if (GetLeft() < 0) {
                        if (visible_range_pos.GetLength() < min_text_w_pos) {
                            // Can't even show the minimal length of label
                            SetWidth(GetWidth() + GetLeft());
                            SetLeft(0.0);
                        }
                    }
                }
                // also reserve some space for additional info
                string sAdditionalInfo;
                GetAdditionalInfo(sAdditionalInfo);
                if (!sAdditionalInfo.empty()) {
                    TModelUnit info_width_px = min(gl.TextWidth(&font, sAdditionalInfo.c_str()), m_Context->GetMaxLabelWidth(font));
                    TModelUnit info_width_pos = m_Context->ScreenToSeq(info_width_px + CRenderingContext::kLabelSpacePx);
                    SetWidth(GetWidth() + info_width_pos);
                    if(!x_LabelOnLeft()) {
                        SetLeft(GetLeft() - info_width_pos);
                        if (GetLeft() < 0) {
                            SetWidth(GetWidth() + GetLeft());
                            SetLeft(0.0);
                        }
                    }
                }
            } else if (m_Config->m_LabelPos == CFeatureParams::ePos_Above) {
                SetHeight(GetHeight() + gl.TextHeight(&font) + kVertSpace + 1);
            }
        }
    } else {
        SetHeight(0);
        SetWidth(0);
        SetLeft(0.0);
        SetTop(0.0);
    }
    // print bounding box info
//    {
//        string sLabel;
//        GetLabel(sLabel, CLabel::eContent);
//        LOG_POST(Trace << "Bounding box for " << this << " \"" << sLabel << "\" x_isDrawn: " << x_isDrawn() <<
//                          ", m_Height: " << GetHeight() << ", m_Width: " << GetWidth() <<
//                          ", m_Pos.m_X: " << GetLeft() << ", m_Pos.m_Y: " << GetTop());
//        LOG_POST(Trace << "Call stack for " << this << ": " << CStackTrace());

//        LOG_POST(Trace << ">>>>");
//    }
}

bool CVcfGlyph::x_RedundantLabelCheck(const string& label) const
{
    // If label is redundant with the title (is contained in the title) for the
    // track, and all glyphs in the track are the same type, do not display the label.
    // This only applies though if the label for the track starts with '[' and ends with ']'
    const CSeqGlyph* parent = m_Parent;
    bool redundant = false;
    bool all_same_type = false;

    if (label.length() > 0 && label[0] == '[' && label[label.size()-1] == ']') {
        while (parent != NULL) {
            const CLayoutGroup* lg = dynamic_cast<const CLayoutGroup*>(parent);
            if (lg != NULL) {
                all_same_type = lg->AllChildrenSameType();
            }

            const CLayoutTrack* lt = dynamic_cast<const CLayoutTrack*>(parent);
            if (lt != NULL) {
                // trim blanks, brackets etc from glyph display string - e.g. [intron] to intron
                string track_title = lt->GetFullTitle();
                size_t start_idx = label.find_first_not_of(" ()[]'\" <>,.@#$&*");
                size_t stop_idx = label.find_last_not_of(" ()[]'\" <>,.@#$&*");
                size_t len = stop_idx-start_idx;

                if (len > 0 &&
                    NStr::FindNoCase(track_title, label.substr(start_idx, len)) != NPOS) {
                        redundant = true;
                }
                break;
            }
            parent = parent->GetParent();
        }
    }

    return (all_same_type && redundant);
}


bool CVcfGlyph::x_LabelOnLeft() const
{
    return true;
}

void CVcfGlyph::x_MaybeDrawLabelAbove(TModelUnit& base) const
{
    if ( !x_ShowLabel() || m_Config->m_LabelPos != CFeatureParams::ePos_Above) {
        return;
    }
    x_DrawLabelWithXPinned(base);
}

void CVcfGlyph::x_MaybeDrawLabelOthers(TModelUnit base) const
{
    if ( !x_ShowLabel()) {
        return;
    }
    if (CFeatureParams::ePos_Side != m_Config->m_LabelPos)
            return;

    x_DrawLabelWithYPinned(base);
}



void CVcfGlyph::x_DrawLabelWithXPinned(TModelUnit& base) const
{
    if (m_Config->m_LabelPos != CFeatureParams::ePos_Above) {
        return;
    }
    IRender& gl = GetGl();

    // visible part of the whole glyph (feature bar + label + additional info)
    TModelRange visible_range = m_Context->IntersectVisible(this);

    const CGlTextureFont& font = m_Config->m_LabelFont;
    TModelUnit font_height = gl.TextHeight(&font);

    if (m_Context->WillLabelFit(visible_range)) {
        // label text that will be shown
        string sLabelTextOut;

        GetLabel(sLabelTextOut, CLabel::eContent);
        // shorten the text if it is wider than visible range
        sLabelTextOut = font.Truncate(sLabelTextOut.c_str(), m_Context->SeqToScreen(visible_range.GetLength()));

        // both above and inside labels should be centered relatively to feature bar on x axis
        TModelUnit LabelX = visible_range.GetFrom() + visible_range.GetLength() * 0.5;
        TModelUnit LabelY = base;
        if (m_Config->m_LabelPos == CFeatureParams::ePos_Above) {
            // move the base to give space for the text above the feature bar
            base += font_height;
            LabelY = base;
            base += kVertSpace + 1;
        }
        if (IsSelected()) {
            gl.ColorC(m_Context->GetSelLabelColor());
        }  else if (m_Config->m_LabelPos == CFeatureParams::ePos_Inside) {
            gl.ColorC(m_Config->m_fgColor.ContrastingColor());
        } else {
            gl.ColorC(m_Config->m_LabelColor);
        }
        m_Context->TextOut(&font, sLabelTextOut.c_str(), LabelX, LabelY, true, true);
    }
}

void CVcfGlyph::x_DrawInnerLabels(TModelUnit base, vector<TModelRange> *labels_range, TSeqRange* interval) const
{
    if (m_HideLabel)
        return;
    if ((m_Config->m_LabelPos != CFeatureParams::ePos_Side) && (m_Config->m_LabelPos != CFeatureParams::ePos_Inside)) 
        return;
    TSeqRange exon_int;
    if (interval) {
        exon_int = interval->IntersectionWith(m_Context->GetVisSeqRange());
        if (exon_int.Empty())
            return;
    }

    // visible part of the whole glyph (feature bar + label + additional info)
    TModelRange visible_range = m_Context->IntersectVisible(this);

    const CGlTextureFont &font = m_Config->m_LabelFont;

    // location of feature bar, must not be overwritten by the label
    TSeqRange FeatureBarRange(GetRange());
    TModelUnit FeatureBarLeft(FeatureBarRange.GetFrom());
    TModelUnit FeatureBarRight(FeatureBarRange.GetToOpen());

    bool side_lbl{ false };
    if (CFeatureParams::ePos_Side == m_Config->m_LabelPos) {
        // how much of the space (in sequence coords) is available for the label
        TModelUnit LabelVisibleWidth{ 0 };
        if (x_LabelOnLeft()) {
            LabelVisibleWidth = (FeatureBarLeft < visible_range.GetFrom() ? 0 : FeatureBarLeft - visible_range.GetFrom());
        }
        else {
            LabelVisibleWidth = (FeatureBarRight > visible_range.GetTo() ? 0 : visible_range.GetTo() - FeatureBarRight);
        }
        side_lbl = (LabelVisibleWidth >= m_Context->GetMinLabelWidthPos());
    }

    string fl_content;
    GetLabel(fl_content, CLabel::eContent);

    CRgbaColor inner_color(m_Config->m_fgColor);
    GetCustomColor(inner_color);
    inner_color = inner_color.ContrastingColor();

    bool inside_only = m_Config->m_LabelPos == CFeatureParams::ePos_Inside;
    CSeqGlyph::x_DrawInnerLabels(base, fl_content, inner_color, font, side_lbl, inside_only, labels_range, exon_int.Empty() ? nullptr : &exon_int);
}

void CVcfGlyph::x_DrawLabelWithYPinned(TModelUnit base) const
{
    if(m_Config->m_LabelPos != CFeatureParams::ePos_Side) return;

    IRender& gl = GetGl();

    // visible part of the whole glyph (feature bar + label + additional info)
    TModelRange visible_range_pos = m_Context->IntersectVisible(this);

    const CGlTextureFont &font = m_Config->m_LabelFont;
    TModelUnit font_height = gl.TextHeight(&font);

    // location of feature bar, must not be overwritten by the label
    TSeqRange FeatureBarRange(GetRange());
    TModelUnit FeatureBarLeft(FeatureBarRange.GetFrom());
    // this is an open position i.e. +1 of the real end pos
    TModelUnit FeatureBarRight(FeatureBarRange.GetToOpen());

    // how much of the space (in sequence position coords) is available for the label
    TModelUnit AvailableLabelVisibleWidthPos(x_LabelOnLeft() ?
        (FeatureBarLeft < visible_range_pos.GetFrom() ? 0 : FeatureBarLeft - visible_range_pos.GetFrom())
        :
        (FeatureBarRight > visible_range_pos.GetToOpen() ? 0 :  visible_range_pos.GetToOpen() - FeatureBarRight));

    string fl_content;
    GetLabel(fl_content, CLabel::eContent);

    CRgbaColor inner_color(m_Config->m_fgColor);
    GetCustomColor(inner_color);
    inner_color = inner_color.ContrastingColor();


    if (AvailableLabelVisibleWidthPos < m_Context->GetMinLabelWidthPos(font)) {
        return;
    }

    // these widths are in screen pixels
    TModelUnit label_width_px = gl.TextWidth(&font, fl_content.c_str());
    TModelUnit max_width_px = min(m_Context->SeqToScreen(AvailableLabelVisibleWidthPos),
        m_Context->GetMaxLabelWidth(font));
    if (label_width_px > max_width_px) {
        label_width_px = max_width_px;
        fl_content = font.Truncate(fl_content.c_str(), label_width_px);
        if ((string::npos != fl_content.find("...")) && (fl_content.length() <= 5))
            return;
    }
    // convert back to sequence position coords
    TModelUnit label_width_pos = m_Context->ScreenToSeq(label_width_px);
    TModelUnit label_x_pos = 0.0;
    TModelUnit label_base = base;
    TModelUnit label_clear_base = base;
    TModelUnit label_clear_height = font_height;
    label_base += (font_height * 0.5);

    // draw background - larger of font height or
    // bar height (if label on same line, e.g. not label first)
    TModelUnit bar_height = m_Config->m_BarHeight;
    label_clear_base = label_base;
    if (bar_height > font_height) {
        label_clear_height = bar_height;
        label_clear_base = floor(base) + (label_clear_height * 0.5);
    }
    if (x_LabelOnLeft()) {
        label_x_pos = visible_range_pos.GetFrom();
        if (label_x_pos > GetLeft()) {
            gl.ColorC(m_Config->m_bgColor);
            m_Context->DrawBackground(
                TModelRect(label_x_pos, label_clear_base + 1, label_x_pos + label_width_pos,
                    label_clear_base - label_clear_height - 1), 0);
        }
    } else {
        label_x_pos = visible_range_pos.GetToOpen() - label_width_pos;
        if (visible_range_pos.GetTo() < GetRight()) {
            gl.ColorC(m_Config->m_bgColor);
            m_Context->DrawBackground(
                TModelRect(label_x_pos, label_clear_base + 1, visible_range_pos.GetToOpen(),
                    label_clear_base - label_clear_height - 1), 0);
        }
    }
    gl.ColorC(IsSelected() ?
        m_Context->GetSelLabelColor() : m_Config->m_LabelColor);

    m_Context->TextOut(&font, fl_content.c_str(), label_x_pos,
        label_base, false, true);
}

void CVcfGlyph::x_DrawAdditionalInfo(TModelUnit base) const
{
    if ( !x_ShowLabel() ) {
        return;
    }

    IRender& gl = GetGl();

    // visible part of the whole glyph (feature bar + label + additional info)
    TModelRange visible_range = m_Context->IntersectVisible(this);
    const CGlTextureFont& font = m_Config->m_LabelFont;
    TModelUnit font_height = gl.TextHeight(&font);

    if(m_Config->m_LabelPos == CFeatureParams::ePos_Side) {
        // location of feature bar, must not be overwritten by the additional info
        TSeqRange FeatureBarRange(GetRange());
        TModelUnit FeatureBarLeft(FeatureBarRange.GetFrom());
        TModelUnit FeatureBarRight(FeatureBarRange.GetToOpen());

        // how much of the space (in sequence coords) is available for the additional info
        // left/right location of additional info is complementary to that of the label
        TModelUnit AvailableAdditionalInfoVisibleWidthPos(!x_LabelOnLeft() ?
                        (FeatureBarLeft < visible_range.GetFrom() ? 0 : FeatureBarLeft - visible_range.GetFrom())
                        :
                        (FeatureBarRight > visible_range.GetToOpen() ? 0 :  visible_range.GetToOpen() - FeatureBarRight));

        string sAdditionalInfo;
        GetAdditionalInfo(sAdditionalInfo);
//        LOG_POST(Trace << "sAdditionalInfo: " << sAdditionalInfo);
//        LOG_POST(Trace << "AvailableAdditionalInfoVisibleWidthPos: " << AvailableAdditionalInfoVisibleWidthPos);
        // widths are in screen pixels
        TModelUnit info_width_px = gl.TextWidth(&font, sAdditionalInfo.c_str());
//        LOG_POST(Trace << "initial info_width_px: " << info_width_px);

        if(AvailableAdditionalInfoVisibleWidthPos < m_Context->GetMinLabelWidthPos()) {
            return;
        }
        TModelUnit max_width_px = min(m_Context->SeqToScreen(AvailableAdditionalInfoVisibleWidthPos),
                                   m_Context->GetMaxLabelWidth(font));
//        LOG_POST(Trace << "max_width_px: " << max_width_px);
        if (info_width_px > max_width_px) {
            info_width_px = max_width_px;
            sAdditionalInfo = font.Truncate(sAdditionalInfo.c_str(), info_width_px);
        }
//        LOG_POST(Trace << "modified sAdditionalInfo: " << sAdditionalInfo << " and info_width_px: " << info_width_px);
        // convert back to sequence coords
        TModelUnit info_width_pos = m_Context->ScreenToSeq(info_width_px);
        TModelUnit info_x_pos = 0.0;
        TModelUnit info_base = base;
        TModelUnit info_clear_base = base;
        TModelUnit info_clear_height = font_height;
        info_base += (font_height * 0.5);

//        LOG_POST(Trace << "info_width_pos: " << info_width_pos);
        // draw background - larger of font height or
        // bar height (if label on same line, e.g. not label first)
        TModelUnit bar_height = m_Config->m_BarHeight;
        info_clear_base = info_base;
        if (bar_height > font_height) {
            info_clear_height = bar_height;
            info_clear_base = floor(base) + (info_clear_height * 0.5);
        }
        if (!x_LabelOnLeft()) {
            info_x_pos = visible_range.GetFrom();
            if (info_x_pos > GetLeft()) {
//                LOG_POST(Trace << " drawing background at left: " << info_x_pos << " with offset: " << m_Context->GetOffset());
                gl.ColorC(m_Config->m_bgColor);
                m_Context->DrawBackground(
                    TModelRect(info_x_pos, info_clear_base + 1, info_x_pos + info_width_pos,
                        info_clear_base - info_clear_height - 1), 0);
            }
        } else {
            info_x_pos = visible_range.GetToOpen() - info_width_pos;
            if (visible_range.GetTo() < GetRight()) {
//                LOG_POST(Trace << " drawing background at right: " << info_x_pos << " with offset: " << m_Context->GetOffset());
                gl.ColorC(m_Config->m_bgColor);
                m_Context->DrawBackground(
                    TModelRect(info_x_pos, info_clear_base + 1, visible_range.GetToOpen(),
                        info_clear_base - info_clear_height - 1), 0);
            }
        }
        gl.ColorC(IsSelected() ?
            m_Context->GetSelLabelColor() : m_Config->m_LabelColor);

//        LOG_POST(Trace << "info_x_pos: " << info_x_pos);
        m_Context->TextOut(&font, sAdditionalInfo.c_str(), info_x_pos,
            info_base, false, true);
    }
}


void CVcfGlyph::x_DrawInsertion(TModelUnit SeqPosTriangleMidPointX,
                                 TModelUnit BoundaryYLow,
                                 TModelUnit BoundaryYHigh,
                                 TModelUnit YCenterLine) const
{
    // half the insertion triangle width in sequence coordinates
    TModelUnit SeqTriangleHalf = m_Context->ScreenToSeq(4.0);

    TModelUnit TriangleOffsetY = 1.0;
    m_Context->DrawLine(SeqPosTriangleMidPointX, BoundaryYLow, SeqPosTriangleMidPointX, BoundaryYHigh);
    m_Context->DrawTriangle(SeqPosTriangleMidPointX, YCenterLine,
        SeqPosTriangleMidPointX + SeqTriangleHalf, BoundaryYLow - TriangleOffsetY,
        SeqPosTriangleMidPointX - SeqTriangleHalf, BoundaryYLow - TriangleOffsetY);
    m_Context->DrawTriangle(SeqPosTriangleMidPointX, YCenterLine,
        SeqPosTriangleMidPointX + SeqTriangleHalf, BoundaryYHigh + TriangleOffsetY,
        SeqPosTriangleMidPointX - SeqTriangleHalf, BoundaryYHigh + TriangleOffsetY);
}

void CVcfGlyph::x_DrawDeletion(TModelUnit SeqPosTriangleMidPointX,
                                 TModelUnit BoundaryYLow,
                                 TModelUnit BoundaryYHigh) const
{
    // half the insertion triangle width in sequence coordinates
    TModelUnit SeqTriangleHalf = m_Context->ScreenToSeq(4.0);

    TModelUnit TriangleOffsetY = 1.0;
    m_Context->DrawTriangle(SeqPosTriangleMidPointX, BoundaryYHigh + TriangleOffsetY,
        SeqPosTriangleMidPointX + SeqTriangleHalf, BoundaryYLow - TriangleOffsetY,
        SeqPosTriangleMidPointX - SeqTriangleHalf, BoundaryYLow - TriangleOffsetY);
}

bool CVcfGlyph::HitTestHor(TSeqPos x, const CObject *obj)
{
    return CSeqGlyph::HitTestHor(x, obj);
}

bool CVcfGlyph::LessBySeqPos(const CSeqGlyph& obj) const
{
    // use default, compare by SeqRanges
    return CSeqGlyph::LessBySeqPos(obj);
}

void CVcfGlyph::x_DrawVariationBar(TModelUnit base) const
{
    IRender& gl = GetGl();

    TModelUnit bar_height = m_Config->m_BarHeight;
    if (m_Context->IsOverviewMode()  &&  m_HideLabel) {
        bar_height = floor(bar_height * m_Config->m_OverviewFactor);
    }
    TModelUnit YCenterLine = base;
    TModelUnit BoundaryYLow = YCenterLine - bar_height * 0.5f;
    TModelUnit BoundaryYHigh = YCenterLine + bar_height * 0.5f;

    CRgbaColor color;
    if ( !GetCustomColor(color) ) {
        color = m_Config->m_fgColor;
    }
    // a lighter version of color
    CRgbaColor color_t = color;
    color_t.SetAlpha(0.35f);

    // Draw variation bar
    TSeqRange SeqRange = GetRange();
    TSeqPos SeqPosFrom = SeqRange.GetFrom();
    TSeqPos SeqPosTo = SeqRange.GetTo();

    if (SeqRange.GetLength() > m_Context->ScreenToSeq(2.0)) {  // At least 2 pixel
        SeqPosFrom = SeqRange.GetFrom();
        SeqPosTo = SeqRange.GetTo();

        if (m_Object->GetVariationType() == CVariation_inst::eType_ins) {
            gl.ColorC(color_t);
            m_Context->DrawQuad(SeqPosFrom, BoundaryYLow, SeqPosTo + 1.0, BoundaryYHigh);
            // draw insertion
            CRgbaColor color_insert;
            GetCustomColor(CCustomFeatureColor::eInsertion, color_insert);
            gl.ColorC(color_insert);
            x_DrawInsertion((SeqPosFrom + SeqPosTo) * 0.5 + 0.5, BoundaryYLow, BoundaryYHigh, YCenterLine);
        } else if (m_Object->GetVariationType() == CVariation_inst::eType_del) {
            gl.ColorC(color_t);
            m_Context->DrawQuad(SeqPosFrom, BoundaryYLow, SeqPosTo + 1.0, BoundaryYHigh);
            // draw deletion
            x_DrawDeletion((SeqPosFrom + SeqPosTo) * 0.5 + 0.5, BoundaryYLow, BoundaryYHigh);
        } else if (m_Object->GetVariationType() == CVariation_inst::eType_delins) {
            // draw deletion
            gl.ColorC(color);
            m_Context->DrawQuad(SeqPosFrom, BoundaryYLow, SeqPosTo + 1.0, BoundaryYHigh);
            // draw insertion
            CRgbaColor color_insert;
            GetCustomColor(CCustomFeatureColor::eInsertion, color_insert);
            gl.ColorC(color_insert);
            x_DrawInsertion(SeqPosTo + 1.0, BoundaryYLow, BoundaryYHigh, YCenterLine);
        } else {
            m_Context->Draw3DQuad(SeqPosFrom, BoundaryYLow, SeqPosTo, BoundaryYHigh, color);
        }
        gl.ColorC(color_t);
        x_DrawInnerLabels(base);
    } else {
        m_Context->Draw3DQuad(SeqPosFrom, BoundaryYLow,
            SeqPosTo, BoundaryYHigh, color, true);
    }
}


TModelUnit CVcfGlyph::x_GetBarHeight() const
{
    TModelUnit height = m_Config->GetBarHeight(
        m_Context->IsOverviewMode()  &&  m_HideLabel);

    return height;
}



bool CVcfGlyph::x_ShowLabel() const
{
    bool shown = false;
    if (m_Config->m_LabelPos != CFeatureParams::ePos_NoLabel) {
        if (m_Config->m_LabelPos == CFeatureParams::ePos_Side) {
            // in MSA context, the whole range can be very small and label can be skipped altogether
            //if (m_Context->WillLabelFit(m_Context->GetVisibleRange())) {
                shown = !m_HideLabel;
            //} 
        } else { // can be either ePos_Above or ePos_Inside
            TModelRange r = m_Context->IntersectVisible(this);
            if (m_Context->WillLabelFit(r)) {
                shown = !m_HideLabel  ||  IsSelected();
            }
            if (shown) {
                string fl_content;
                GetLabel(fl_content, CLabel::eContent);
                IRender& gl = GetGl();
                TModelUnit available_width = m_Context->SeqToScreen(r.GetLength());
                TModelUnit label_width = gl.TextWidth(&(m_Config->m_LabelFont), fl_content.c_str());
                if (label_width > available_width) {
                    fl_content = m_Config->m_LabelFont.Truncate(fl_content.c_str(), available_width);
                    if ((string::npos != fl_content.find("...")) && (fl_content.length() <= 5))
                        shown = false;
                }
            }
        }

        // Check if label is redudant with track name and if so, hide:
        if (shown) {
            string fl_content;
            GetLabel(fl_content, CLabel::eContent);
            if (x_RedundantLabelCheck(fl_content))
                shown = false;
        }
    }
    return shown;
}


const CSeq_loc& CVcfGlyph::GetLocation(void) const
{
	return m_Object->GetLocation();
}


CConstRef<CObject> CVcfGlyph::GetObject(TSeqPos) const
{
	return m_Object;
}


void  CVcfGlyph::GetObjects(vector<CConstRef<CObject> >& objs) const
{
	objs.push_back(m_Object);
}


bool CVcfGlyph::HasObject(CConstRef<CObject> obj) const
{
	return m_Object == obj;
}

string CVcfGlyph::GetSignature() const
{
	string sig = CObjFingerprint::GetVcfVariantSignature(*m_Object, &m_Context->GetScope());
	return sig;
}


const CVcfGlyph::TIntervals& CVcfGlyph::GetIntervals(void) const
{
	return m_Intervals;
}




END_NCBI_SCOPE
