/*  
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladislav Evgeniev
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/vcf_histogram.hpp>
#include <gui/widgets/seq_graphic/graph_cache.hpp>
#include <util/bitset/bmserial.h>
#include <util/bitset/bmsparsevec_serial.h>
#include <util/checksum.hpp>
#include <corelib/rwstream.hpp>
#include <connect/ncbi_conn_stream.hpp>
#include <corelib/ncbiexec.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
///   CVcfHistogram

mutex CVcfHistogram::m_KeysMapMutex;
map<string, vector<string>> CVcfHistogram::m_KeysMap;

string CVcfHistogram::GetCacheKey(const string& annot_name, const size_t bin_width)
{
    CCompoundIDPool id_pool;
    auto id = id_pool.NewID(eCIC_GenericID);
    
    id.AppendString(annot_name);
    id.AppendInteger(bin_width);
        
    return id.ToString();

}

CVcfHistogram::CVcfHistogram(const string& cache_key)
{
    _ASSERT(!cache_key.empty());
    if (cache_key.empty())
        NCBI_THROW(CException, eUnknown, "Empty cache key");

    CCompoundIDPool id_pool;
    auto id = id_pool.FromString(cache_key);

    auto field_annot = id.GetFirst(eCIT_String);
    m_AnnotName = field_annot.GetString();
    if (m_AnnotName.empty()) NCBI_THROW(CException, eUnknown, "Empty annotation");

    auto field_bin_width = field_annot.GetNextNeighbor();
    m_BinWidth = field_bin_width.GetInteger();
    if (m_BinWidth == 0) NCBI_THROW(CException, eUnknown, "Empty bin width");

    CChecksum cs(CChecksum::eMD5);
    cs.AddLine(m_AnnotName);
    cs.AddLine(NStr::NumericToString(m_BinWidth));
    m_DataKey = cs.GetHexSum();
    m_DataKey += "_vcf_histogram";

    {
        lock_guard<mutex> guard(m_KeysMapMutex);
        m_KeysMap[m_AnnotName].push_back(cache_key);
    }
}


CVcfHistogram::CVcfHistogram(const CVcfHistogram& data)
    : m_AnnotName(data.m_AnnotName)
    , m_BinWidth(data.m_BinWidth)
    , m_DataKey(data.m_DataKey)
    , m_Data(data.m_Data)
{
}


void CVcfHistogram::Init(TGraphCachePool& icache_pool)
{
    lock_guard<mutex> guard(m_DataMutex);
    if (!m_Data.empty())
        return;

    CObjPoolGuard<TGraphCachePool> icache(icache_pool);
    
    // is there any data stored?
    int blob_size = 0;
    try {
        blob_size = icache->GetSize(m_DataKey, 0, NcbiEmptyString);
    } catch(const exception&) {
    }
    if (blob_size > 0) {
        try {
            CRStream is(icache->GetReadStream(m_DataKey, 0, NcbiEmptyString), 0, 0, CRWStreambuf::fOwnReader);
            int val = 0;
            vector<unsigned char> buffer(val, 0);
            buffer.resize(blob_size);
            is.read((char*)&buffer[0], buffer.size());
            if (!is) NCBI_THROW(CException, eUnknown, "Failed to read VCF histogram from cache: " + m_DataKey);
            bm::sparse_vector_deserializer<THistogramVector> deserializer;
            deserializer.deserialize(m_Data, &buffer[0]);
            return;
        } catch (exception& e) {
            ERR_POST(Error << e.what());
        }
    }
}

void CVcfHistogram::Save(TGraphCachePool& icache_pool)
{
    CObjPoolGuard<TGraphCachePool> icache(icache_pool);
    m_Data.optimize();
    bm::sparse_vector_serializer<THistogramVector>       sv_ser;
    bm::sparse_vector_serial_layout<THistogramVector>    sv_lay;
    sv_ser.enable_xor_compression();
    _ASSERT(sv_ser.is_xor_ref());
    sv_ser.serialize(m_Data, sv_lay);
    CWStream w(icache->GetWriteStream(m_DataKey, 0, NcbiEmptyString), 0, 0, CRWStreambuf::fOwnWriter);
    w.write((const char*)sv_lay.buf(), sv_lay.size());
    if (!w) NCBI_THROW(CException, eUnknown, "Failed to write VCF histogram to cache: " + m_DataKey);
    w.flush();
}

void CVcfHistogram::GetHistogram(const TSeqRange& range, CHistogramGlyph::TMap& dmap) const
{
    lock_guard<mutex> guard(m_DataMutex);
    InitHistogramGlyph(m_BinWidth, range, m_Data, dmap);
}

void CVcfHistogram::EraseCachePerAnnotation(const string& annot_name)
{
    lock_guard<mutex> guard(m_KeysMapMutex);
    auto it = m_KeysMap.find(annot_name);
    if (it == m_KeysMap.end())
        return;
    auto &histogram_cache = CGraphCache<CVcfHistogram>::GetInstance();
    for(const auto& key : it->second) {
        histogram_cache.RemoveData(key);
    }
    m_KeysMap.erase(it);
}

END_NCBI_SCOPE
