/*  $Id: vcf_track.cpp 45881 2021-01-05 03:10:02Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/seq_graphic/vcf_track.hpp>

#include <gui/widgets/seq_graphic/vcf_glyph.hpp>
#include <gui/widgets/seq_graphic/histogram_glyph.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_ds_manager.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/seq_graphic/layout_track_proxy.hpp>
#include <gui/widgets/seq_graphic/config_utils.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/obj_fingerprint.hpp>
#include <gui/objects/TextBox.hpp>
#include <objmgr/util/sequence.hpp>
#include <util/checksum.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
///   CVcfTrack
///////////////////////////////////////////////////////////////////////////////

CTrackTypeInfo CVcfTrack::m_TypeInfo("vcf_track",
                                     "VCF records");

CVcfTrack::CVcfTrack(CSGFeatureDS* ds, CRenderingContext* r_cntx)
    : CFeatureTrack(ds, r_cntx)
{
	SetLayoutPolicy(m_Layered);
}


CVcfTrack::~CVcfTrack()
{
	if (!m_DS->AllJobsFinished()) {
		m_DS->DeleteAllJobs();
	}
}


const CTrackTypeInfo& CVcfTrack::GetTypeInfo() const
{
    return m_TypeInfo;
}


void CVcfTrack::x_UpdateData()
{
    if (m_ZoomToSee) {
        string msg = "Zoom to see if data exists for this region";
        const TSeqRange& range = m_Context->GetVisSeqRange();
        TModelUnit x = (range.GetFrom() + range.GetTo() ) * 0.5;
        CSeqGlyph::TObjects objs;
        CRef<CCommentGlyph> comment(new CCommentGlyph(msg, TModelPoint(x, 0.0)));
        objs.push_back(comment);
        CRef<CCommentConfig> config(new CCommentConfig(*x_GetGlobalConfig()->GetCommentConfig()));
        config->m_ShowConnection = false;
        SetCommentGlyphs(objs, config);
        return;
    }
	CFeatureTrack::x_UpdateData();
}

void CVcfTrack::x_LoadSettings(const string& preset_style, const TKeyValuePairs &settings) 
{
    CFeatureTrack::x_LoadSettings(preset_style, settings);
    auto it = settings.find("zoom_to_see");
    m_ZoomToSee = (it != settings.end() && NStr::StringToBool(it->second));
}

void CVcfTrack::x_OnJobCompleted(CAppJobNotification& notify)
{
    m_DS->ClearJobID(notify.GetJobID());
    CRef<CObject> res_obj = notify.GetResult();
    CSGJobResult* result = dynamic_cast<CSGJobResult*>(&*res_obj);
    if (!result) {
        LOG_POST(Error << "CVcfTrack::x_OnJobCompleted() notification for job \
                does not contain results.");
        return;
    }
    CSeqGlyph::TObjects& objs(result->m_ObjectList);
    const size_t obj_size = objs.size();
    bool compact_mode = (long)obj_size > m_CompactThreshold;
    if (obj_size > 0) {
        //!! we might want a separate section later!
        CConstRef<CFeatureParams> conf(x_GetGlobalConfig()->GetFeatParams("dbvar"));
        SetObjects(objs);
        x_AdjustFeatureLayout();
        NON_CONST_ITERATE (CSeqGlyph::TObjects, iter, objs) {
            CSeqGlyph* glyph = *iter;
            glyph->SetHideLabel(compact_mode);
            glyph->SetRenderingContext(m_Context);
            CVcfGlyph* vcf_glyph = dynamic_cast<CVcfGlyph*>(glyph);
            if(vcf_glyph) {
                vcf_glyph->SetConfig(conf);
            } else {
				CHistogramGlyph* hist_glyph = dynamic_cast<CHistogramGlyph*>(glyph);
				if(hist_glyph) {
					hist_glyph->SetConfig(*x_GetGlobalConfig());
				}
			}
        }
        string msg = ", total ";
        msg += NStr::SizetToString(obj_size, NStr::fWithCommas);
        if (obj_size > 1) {
            msg += " features";
        } else {
            msg += " feature";
        }
        msg += " shown";
        SetMsg(msg);
        m_Attrs |= fNavigable;
    }
    else {
        SetObjects(objs);
        SetMsg(", no data");
    }
    x_UpdateLayout();
}


CHTMLActiveArea* CVcfTrack::InitHTMLActiveArea(TAreaVector* p_areas) const
{
    if (!GetChildren().empty()) {
        TModelUnit tb_height = x_GetTBHeight();

        // for Associated results/GAP types of tracks, generate a separate area for the track caption
        CHTMLActiveArea area;
        CSeqGlyph::x_InitHTMLActiveArea(area);

        // adjust the bounds height to that of the caption only
        TVPRect& bound(area.m_Bounds);
        bound.SetBottom(bound.Top() + (TVPUnit)tb_height + (tb_height > 0.0 ? 2 : 0));

        // user uploaded tracks are not based on NA annots
        // we don't need to generate signature for them 
        // otherwise CTrackContainer will create multiple signatures
        // and sviewer will issue seqgraphic call to get track's tooltip
        // we don't want that for user uploaded tracks
        
        area.m_Flags = CHTMLActiveArea::fNoSelection | CHTMLActiveArea::fNoPin | CHTMLActiveArea::fTrack;
        if (!(m_Attrs & fNavigable))
            area.m_Flags |= CHTMLActiveArea::fNoNavigation;
        p_areas->push_back(area);
        return &(p_areas->back());
    }
    auto area = CLayoutTrack::InitHTMLActiveArea(p_areas);
    if (area && isRmtBased()) // turn off navigation for VCFTabix
        area->m_Flags |= CHTMLActiveArea::fNoNavigation;
    return area;
}

///////////////////////////////////////////////////////////////////////////////
/// CVcfTrackFactory
///////////////////////////////////////////////////////////////////////////////
ILayoutTrackFactory::TTrackMap
CVcfTrackFactory::CreateTracks(SConstScopedObject& object,
                               ISGDataSourceContext* ds_context,
                               CRenderingContext* r_cntx,
                               const SExtraParams& params,
                               const TAnnotMetaDataList& src_annots) const
{
    TAnnotNameTitleMap annots;
    TTrackMap tracks;

    TKeyValuePairs track_settings;
    CSGConfigUtils::ParseProfileString(params.m_TrackProfile, track_settings);

    if (!params.m_Annots.empty()) {
		ITERATE(SExtraParams::TAnnots, iter, params.m_Annots) {
			if (NStr::StartsWith(*iter, "ColumnarVCF_")) {
				annots.insert(TAnnotNameTitleMap::value_type(*iter, ""));
			}
        }
    } else {

        TSeqRange range = params.m_Range;
        if (range.Empty()) {
            range = r_cntx->GetVisSeqRange();
        }


        // collect VCF tracks
        if ( !src_annots.empty() ) {
            GetMatchedAnnots(ds_context, src_annots, params, annots);
        }
    }

    // create VCF tracks
    ITERATE (TAnnotNameTitleMap, iter, annots) {
        CRef<CVcfTrack> feat_track;
        CIRef<ISGDataSource> ds = ds_context->GetDS(
            typeid(CVcfFeatureDSType).name(), object);
        CVcfFeatureDS* vcf_ds = dynamic_cast<CVcfFeatureDS*>(ds.GetPointer());
        feat_track.Reset(new CVcfTrack(vcf_ds, r_cntx));
        feat_track->SetAnnot(iter->first);
        if ( !iter->second.empty() ) {
            feat_track->SetTitle(iter->second);
        }
        tracks[iter->first] = feat_track.GetPointer();
    }

    return tracks;
}

void CVcfTrackFactory::GetMatchedAnnots(
    const TAnnotMetaDataList& src_annots,
    const ILayoutTrackFactory::SExtraParams& params,
    TAnnotNameTitleMap& out_annots) const
{
	//!!
	// currently VCF annots are captured by the fact that they have special "ColumnarVCF_" prefix
	// this prefix may change
	for (auto& i : src_annots) {
		if(NStr::StartsWith(i.first, "ColumnarVCF_")) {
			out_annots.insert(TAnnotNameTitleMap::value_type(
				i.first, i.second->m_Title));
		}
	}
}


void CVcfTrackFactory::GetMatchedAnnots(
    ISGDataSourceContext* ds_context,
    const TAnnotMetaDataList& src_annots,
    const ILayoutTrackFactory::SExtraParams& params,
    TAnnotNameTitleMap& out_annots) const
{
    INonAsnDataRegistry *registry = dynamic_cast<INonAsnDataRegistry*>(ds_context);
    if (nullptr == registry) {
        return GetMatchedAnnots(src_annots, params, out_annots);
    }
    for (auto& i : src_annots) {
        if (!registry->HasNonAsnData(i.first))
            continue;

        out_annots.insert(TAnnotNameTitleMap::value_type(i.first, i.second->m_Title));
    }
}


string CVcfTrackFactory::GetExtensionIdentifier() const
{
    return CVcfTrack::m_TypeInfo.GetId();
}


string CVcfTrackFactory::GetExtensionLabel() const
{
    return CVcfTrack::m_TypeInfo.GetDescr();
}


CRef<CTrackConfigSet>
CVcfTrackFactory::GetSettings(const string& /*profile*/,
                              const TKeyValuePairs& settings,
                              const CTempTrackProxy* track_proxy) const
{
	//!! totally rework this, probably similar to CSnpTrackFactory::GetSettings()
    CRef<CTrackConfigSet> config_set(new CTrackConfigSet);
    CRef<CTrackConfig> config(new CTrackConfig);
    config_set->Set().push_back(config);
    config->SetHelp() = GetThisTypeInfo().GetDescr();
    config->SetLegend_text("anchor_5");

    string rendering_str = CFeatureTrack::LinkedFeatDisplayValueToStr(ELinkedFeatDisplay::eLFD_Expandable);
    string hist_t = "0";

    ITERATE (TKeyValuePairs, iter, settings) {
        try {
            if (NStr::EqualNocase(iter->first, "Rendering")) {
                rendering_str = iter->second;
                if (rendering_str == "Histogram") // support legacy value
                    rendering_str = "Heatmap";
            } else if (NStr::EqualNocase(iter->first, "HistThreshold")) {
                int val = NStr::StringToInt(iter->second);
                if (val > 0) hist_t = iter->second;
            }
        } catch (CException& e) {
            LOG_POST(Warning << "CVcfTrackFactory::x_LoadSettings() "
                << e.GetMsg());
        }
    }

    string sort_by;
    if (track_proxy)  {
        sort_by = track_proxy->GetSortBy();
    }

    CRef<CChoice> choice = CTrackConfigUtils::CreateChoice(
        "Rendering", "Rendering options",
        rendering_str,
        "Rendering options for linked features.");

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CFeatureTrack::LinkedFeatDisplayValueToStr(ELinkedFeatDisplay::eLFD_Default),
        "Show all",
        "All features, both parent and all child features",
        "The  bar on top represents the variant region. Each bar below represents \
        the supporting level variants used to define this region.<br>Red: A Deletion or Loss<br> \
        Green: A Gain or Insertion<br>Light Gray: An Inversion"));

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CFeatureTrack::LinkedFeatDisplayValueToStr(ELinkedFeatDisplay::eLFD_Expandable),
        "Show parent, Expand children upon a click",
        "Show parent on one line and expand all children upon a mouse click.",
        "The bar on top represents the variant region. The bars below appear upon a mouse click and each bar represents \
        the supporting level variants used to define this region.<br>Red: A Deletion or Loss<br> \
        Green: A Gain or Insertion<br>Light Gray: An Inversion"));

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CFeatureTrack::LinkedFeatDisplayValueToStr(ELinkedFeatDisplay::eLFD_Packed),
        "Show parent, Merge children",
        "Show parent on one line and all children merged below it.",
        "The  bar on top represents the variant region. The bar below represents \
        all of the sample level variants merged onto a single line.<br>Red: A Deletion or Loss<br> \
        Green: A Gain or Insertion<br>Light Gray: An Inversion"));

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CFeatureTrack::LinkedFeatDisplayValueToStr(ELinkedFeatDisplay::eLFD_ParentHidden),
        "Show children, not parent",
        "Show only the supporting variants (chidlren) not the variant region (parent)",
        "The bar represent supporting variants in this study.<br>Red: A Deletion or Loss<br>\
        Green: A Gain or Insertion<br>Light Gray: An Inversion"));

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CFeatureTrack::LinkedFeatDisplayValueToStr(ELinkedFeatDisplay::eLFD_Hidden),
        "Show parent, not children ",
        "Show the variant region (parent) not the supporting variants (children)",
        "The bar represents the variant region."));

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        "SingleLine",
        "Display on a single line",
        "Track is displayed, with all features collapsed on a single line.",
        "All the features are packed into one line."));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        "Heatmap",
        "Heatmap",
        "Track is displayed, with all features displayed as a heatmap.",
        "All the features are packed and shown as a heatmap."));

    config->SetChoice_list().push_back(choice);


    CRef<CTextBox> txt_box(new CTextBox);
    txt_box->SetName("HistThreshold");
    txt_box->SetValue(hist_t);
    txt_box->SetDisplay_name("Heatmap Threshold");
    txt_box->SetHelp("Display a heatmap when there are more features in this track than the value specificed."
                     " 'None' or '0' means threshold is disabled, and the display is up to 'Rendering Options'");
    config->SetText_boxes().push_back(txt_box);

    return config_set;
}


END_NCBI_SCOPE
